#------------------------------------------------------------------------------
# snackAmp - Catalog package
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/catalog.tcl,v 1.61 2004/02/12 00:48:38 wilkason Exp $
}
namespace eval catalog {
   variable lastList [list]
}
#------------------------------------------------------------------------------
# Create a csv file to contain the listing
#------------------------------------------------------------------------------
proc catalog::documentCache {args} {
   global TA snackAmpSettings
   ##
   # Restore the cached files from each media Dir
   #
   set catFiles [list]
   foreach {Path} $TA(mediaDirs) {
      catch {unset dirListing}
      if {[catch {set fid [open [file join $Path $TA(cacheFile)] r]}]} {
         return
      }
      fconfigure $fid -buffersize 100000
      catch {unset TaFiles}
      set TaFiles [gets $fid]
      close $fid
      ##
      # Cached files are store relative to this dir so join with mediaDir to get abs file name
      #
      foreach {File} $TaFiles {
         lappend dirListing([file dirname $File]) [file tail $File]
      }
      # Update latest cache time to current file
      if {[catch {set fid [open [file join $Path listing.csv] w]}]} {
         return -code error
      }
      lappend catFiles [file join $Path listing.csv]
      puts $fid [subst {"Folder","Title"}]
      foreach {Dir} [lsort [array names dirListing]] {
         Update
         set Songs $dirListing($Dir)
         foreach {Song} [lsort $Songs] {
            puts $fid [subst {"$Dir","$Song"}]
         }
      }
      close $fid
   }
   tk_messageBox -icon info -type ok -message "csv list of files created\n[join $catFiles \n]"
}
#------------------------------------------------------------------------------
# Function   : catalog::pushList/popList
# Description: Save/return the current master list order, called after
#              the master list is created.
# Author     : Tom Wilkason
# Date       : 12/18/2002
#------------------------------------------------------------------------------
proc catalog::pushList {lst} {
   #puts [me];update
   variable lastList
   set lastList $lst  ;# This uses about 4Mbytes for 28K tracks
   ##
   # What we really should do is save the playlist prior to filter
   # so we can restore it properly after filtering is over.
   #
   return 1
}
proc catalog::popList {args} {
   #puts [me];update
   # If the master list is dirty, regenerate
   # the shuffled/sorted so we don't miss
   # any changes
   variable lastList
   if {$db::isSorted==0} {
      db::makeSorted
      return [filtering::createMasterList]
   } elseif {[llength $lastList]} {
      return $lastList
   } else {
      return [filtering::createMasterList]
   }
}
##
# Return a list of file matching the APL specs (in current APL sort order).
# In the current sorted order
# TBD: Can this be sped up? 'file dirname' is now slow
proc catalog::aplFiles {{max 0}} {
   Trace
   ##
   # Create the folder tree entries if needed
   #
   if {[aplFolder::entries]==0} {
      aplFolder::createTeFolders 0 1
   }
   set insertList [list]
   set len 0
   foreach {File} [catalog::popList] {
      if {[aplFolder::isSet [file dirname $File]]} {
         lappend insertList $File
         if {$max > 0 && $len >= $max} {
            break
         }
         incr len
      }
   }
   return $insertList
}
##
# Apply the current auto playlist
#
proc catalog::applyAPL {args} {
   Trace
   set max [expr {$::snackAmpSettings(maxAPLApply)-1}]
   if {$max < 0} {
      set max 1000000
   }
   # Get a list of all APL tracks, sorted appropriatly
   set insertList [catalog::aplFiles $max]
   LB::deleteAll
   if {[llength $insertList] > 0} {
      LB::insertList 0 $insertList
      # Show number of files Parent listbox
      LB::tryRestorePlaying
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::saveAs
# Description: Save the current file as some format
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::saveAs {args} {
   set file [snack::getSaveFile]
   if {[string length $file]==0} {
      return
   }
   # Stop (should work even if playing)
   # Clone current sound
   set snd [snack::sound -file [soundControl::fileName]]
   # Get extension
   set ext [string tolower [string range [file extension $file] 1 end]]
   ##
   # Status Window
   #
   # Save the file
   if {[catch {
      switch -- $ext {
         "ogg" {
            #W: Prompt for -nominalbitrate, -maxbitrate, -minbitrate or use existing
            switch -- [$snd config -fileformat] {
               MP3 {set rate [$snd config -bitrate]}
               OGG {set rate [$snd config -nominalbitrate]}
               WAV {set rate [expr {[$snd config -rate]}]}
               default {set rate 0}
            }
            if {$rate>0} {
               waitWindow::show "Saving $file ... 0% complete"
               $snd write $file -fileformat $ext -nominalbitrate $rate -progress catalog::waiter
            } else {
               waitWindow::show "Saving ... 0% complete"
               $snd write $file -fileformat $ext -progress catalog::waiter
            }
         }
         default {
            $snd write $file -fileformat $ext -progress catalog::waiter
         }
      }
   } result]} {
      tk_messageBox -icon warning -type ok -title "Save Error" -message $result
   }
   # Make sure it is in the master list
   db::addToDbIfNeeded $file
   $snd destroy
   waitWindow::close
}
#------------------------------------------------------------------------------
# Function   : catalog::waiter
# Description: Update global used for wait box
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::waiter {cmd percent} {
   set percent [expr {round($percent*1000)/10.0}]
   waitWindow::show  "Saving ... $percent% complete"
   update   ;# allow music to play
}

#------------------------------------------------------------------------------
# Return true if this files has an extension that we car about
#------------------------------------------------------------------------------
proc catalog::isTrackFile {file} {
   global taExtensions
   return [info exists taExtensions([file extension $file])]
}
#------------------------------------------------------------------------------
# Function   : catalog::MonitorScanFolders
# Description: Monitor the base folders for incoming data, useful
#              for napaster adding files.
# Author     : Tom Wilkason
# Date       : 12/30/2000
#------------------------------------------------------------------------------
proc catalog::MonitorScanFolders {args} {
   global snackAmpSettings TA
   if {$snackAmpSettings(MonitorScanFolders)} {
      for {set i 0} {$i < $snackAmpSettings(ScanFolderDepth)} {incr i} {lappend patt *}
      set pattern [eval file join $patt]
      #saLog "Checking for new files with $pattern"
      foreach {Folder} $snackAmpSettings(IncomingDirs) {
         # First time will be empty
         foreach {file} [glob  -nocomplain -directory $Folder */*] {
            if {[catalog::isTrackFile $file]} {
               db::cacheFileName $file 0
               Update
            }
         }
      }
   }
   after cancel catalog::MonitorScanFolders
   after 60000 catalog::MonitorScanFolders
}

#------------------------------------------------------------------------------
# Function   : catalog::fixMediaPath
# Description: Make sure the directories in the media path don't have trailing slashes
#              needed by tree matching functions
# Author     : Tom Wilkason
# Date       : 1/11/2001
# TBD: we can revalidate the path here
#------------------------------------------------------------------------------
proc catalog::fixMediaPath {args} {
   global TA
   cleanPathList ::snackAmpSettings(IncomingDirs) 1
   cleanPathList ::snackAmpSettings(mediaDirs) 0 ;#1
   cleanPathList ::snackAmpSettings(RemovableFolders) 0
   foreach {folder} $::snackAmpSettings(mediaDirs) {
      if {![info exists TA(Dirty,$folder)]} {
         set TA(Dirty,$folder) 0
      }
   }
}
#------------------------------------------------------------------------------
# Function to routinely monitor the Cache files and read them in if they change
#------------------------------------------------------------------------------
# proc catalog::MonitorCacheFolders {args} {
#    global snackAmpSettings TA
#    if {$snackAmpSettings(MonitorCacheFolders)} {
#       foreach {Folder} $TA(mediaDirs) {
#          if {[catch {
#             file stat [file join $Folder $TA(cacheFile)] Stat
#             if {![info exists TA(CheckedCache,$Folder)]} {
#                set TA(CheckedCache,$Folder) $Stat(mtime)
#             } else {
#                if {$Stat(mtime) > $TA(CheckedCache,$Folder)} {
#                   catalog::validateMediaPaths
#                   saLog "[clock format [clock seconds]]: Read Cache $Stat(mtime) vs. $TA(CheckedCache,$Folder)"
#                   break
#                }
#             }
#          } result]} {
#             saLog "MonitorCacheFolders: $result"
#          }
#       }
#    }
#    # Check in two minutes
#    after cancel catalog::MonitorCacheFolders
#    after 120000 catalog::MonitorCacheFolders
# }
# ----------------------------------------------------------------------------
# Catalog track details and put into database
# ----------------------------------------------------------------------------
proc catalog::Catalog {path} {
   set totalTime 0
   set totalSize 0
   db::tableTreeData $path {} 0 20 "" totalTime totalSize tableData _Index2File
   waitWindow::close
}
#----------------------------------------------------------------------------
# Rescan all the media dirs after asking for permission.
#----------------------------------------------------------------------------
proc catalog::IndexAskFirst {args} {
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to index the following folders?\n$::snackAmpSettings(mediaDirs)"]=="yes"} {
      catalog::Index
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::Index
# Description: Reread the directories for files, include subdirectories
#              For each folder, add a menu entry
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc catalog::Index {{Folder ""}} {
   global TA snackAmpSettings
   Trace
   Wait 1
   set here [pwd]
   # If a folder is specifed, catalog all media dirs, otherwise just
   # catalog the specified folder retaining information on other folders.
   if {$Folder == ""} {
      waitWindow::show "Indexing Tracks ... please wait"
      update idletasks
      set saveNames [aplFolder::savePattern *]  ;# want folder also
      aplFolder::stripBranch *
      # Get a list of all the files intoList
      # Only scan the known mediaDirs
      foreach dir $TA(mediaDirs) {
         # Flag all as unmounted till recached
         db::unmountBranch [file join $dir *]
         indexRecurse $dir
         db::removeUnmountedFiles [file join $dir *]
      }
   } else {
      # Set master list to all but those files in this branch
      waitWindow::show "Indexing Tracks ... please wait"
      update
      # Flag these files as unmounted
      # Then clear folders
      set saveNames [aplFolder::savePattern $Folder*]  ;# want folder also
      aplFolder::stripBranch [file join $Folder *]
      db::unmountBranch [file join $Folder *]
      # Rescan selected folder (restore any apl folder settings)
      indexRecurse $Folder
      db::removeUnmountedFiles [file join $Folder *]
      # Restore prior settings
      aplFolder::restoreNames $saveNames
   }
   db::dirtyList
   ##
   # Build the folder list for the menu
   #
   waitWindow::show "Cleaning Up ... Almost Done"
   update
   # Save the files
   db::Commit
   # Reread cache to set teFolders and sort the music
   # (only if all media dirs were scanned)
   if {$Folder == ""} {
      waitWindow::show "Verifying/Reading Catalogs"
      ##
      # Update the APL settings
      # (should really restore the old memory settings)
      # Restore prior settings, note it indexing
      # is done prior to first time APL use, then
      # all apl values are 0. Need to fiX
      aplFolder::restoreNames $saveNames
      #readAPLFile $snackAmpSettings(lastFolderFile)
   }
   if {[APLExists]} {
      APL::buildAPL $::TREE
      APL::treeBind
      Tree::open $::TREE / 1
   }
   waitWindow::close
   Wait 0
}
#------------------------------------------------------------------------------
# Escape characters in file/folder names that cause problems
# with globbing and regexp.
#------------------------------------------------------------------------------
proc catalog::globEscape {dir} {
   return [string map {( \\( ) \\) \{ \\{ \} \\} \] \\] \[ \\[} $dir]
}
#------------------------------------------------------------------------------
# Recursively scan a directory and build a list of all files
#------------------------------------------------------------------------------
proc catalog::indexRecurse {dir} {
   global TA LB snackAmpSettings
   set dirList [list]
   # Show banner every once in a while
   if {[clock clicks -milliseconds]%10 == 0} {
      waitWindow::show "Indexing $dir"
   }
   update       ;# Allow music to play and update display
   set hasSome 0
   # Allow {} and [] in names (glob uses them as special chars)
   set gdir [globEscape [file join $dir *]]
   if {$snackAmpSettings(fastscan)} {
   # This is the fastest recursive globbing method I can come up with
      foreach file [glob -nocomplain $gdir] {
         # This is problematic if folder ending in .mp3
         if {[catalog::isTrackFile $file]} {
            # Cache if needed
            db::cacheFileName $file
            incr hasSome
         } elseif {[file isdirectory $file]} {
            lappend dirList $file
            incr hasSome
         }
      }
   } else {
      # This is a bit slower due to additional directory checks
      # but will see ALL directories
      foreach file [glob -nocomplain $gdir] {
         if {[file isdirectory $file]} {
            lappend dirList $file
            incr hasSome
         } elseif {[catalog::isTrackFile $file]} {
            db::cacheFileName $file
            incr hasSome
         }
      }
   }
   # Only set the folder (as existing) for the tree if
   # it was non-empty
   if {$hasSome} {
      aplFolder::Set $dir 0      ;# Add this dir, if new will be turned of
   }
   # process each directory recursively
   foreach {folder} $dirList {
      indexRecurse $folder
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::FlagDirtyCache
# Description: Flag one of the media dir caches as dirty, so we can know to
#              write it later.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::FlagDirtyCache {{File {}}} {
   global TA snackAmpSettings
   ##
   # Save the settings
   #
   fixMediaPath
   ##
   # Cache the data in each mediaDir for shared use
   # If a file was specifed, only do the folder that
   # it is contained in
   set Folders [list]
   if {$File == {}} {
      set Folders $TA(mediaDirs)
   } else {
      foreach Path $TA(mediaDirs) {
         if {[string match "$Path*" $File] } {
            set Folders [list $Path]
            break
         }
      }
   }
   # Increment dirty counter
   foreach {folder} $Folders {
      incr TA(Dirty,$folder)
   }
}
#------------------------------------------------------------------------------
# Re-read all of the catalog files
#------------------------------------------------------------------------------
# proc catalog::ReReadCatalogs {args} {
#    catalog::validateMediaPaths
#    if {[APLExists]} {
#       APL::buildAPL $::TREE
#       APL::treeBind
#       Tree::open $::TREE / 1
#    }
# }
#----------------------------------------------------------------------------
# Verify all the mediadirs and set the mounted flag appropriatly
#----------------------------------------------------------------------------
proc catalog::validateMediaPaths {args} {
   Trace
   variable dbObj
   variable DB
   global TA snackAmpSettings
   set TA(mediaDirs) [list]
   foreach {Path} $snackAmpSettings(mediaDirs) {
      if {[file isdirectory $Path]} {
         lappend TA(mediaDirs) $Path
      } else {
         saLog "Media folder '$Path' not found (is it mounted?)"
      }
   }
   return "" ;# always succeed
}
#------------------------------------------------------------------------------
# Function   : catalog::manageMediaFolders
# Description: Bring up dialog then update folder caches if needed
# Author     : Tom Wilkason
# Date       : 6/5/2002
#------------------------------------------------------------------------------
proc catalog::manageMediaFolders {args} {
   package require saSettings
   global snackAmpSettings
   set oldMedia $::TA(mediaDirs)
   folderManager::launch mediaDirs
   # Blocked here until GUI closes
   catalog::fixMediaPath  ;# Remove braces if added pulling data off listbox
   catalog::validateMediaPaths
   if {$oldMedia != $::TA(mediaDirs)} {
      $::TA(treeButton) configure -relief flat
      destroy .wFC
   }
}
#------------------------------------------------------------------------------
# Function   : playlistStartup
# Description: Populate the playlist with the appropriate files
# Author     : Tom Wilkason
# Date       : 6/5/2002
#------------------------------------------------------------------------------
proc catalog::playlistStartup {args} {
   Trace
   global snackAmpSettings TA
   ##
   # Rescan if set true or the cache read failed
   #
   if {$snackAmpSettings(AutoRescan) || $TA(rescan)} {
      catalog::Index
   }
   ;# Validate paths
   catalog::validateMediaPaths
   # Read cache files, now restore playlist if required
   # otherwise just toss up music based on default auto-playlist file
   if {$snackAmpSettings(restorePlayList)} {
      waitWindow::show "Restoring Playlist"
      update
      if {[catch {catalog::ReadPlaylistCache} result]} {
         waitWindow::show "Reading [file tail $snackAmpSettings(lastFolderFile)]"
         update
         readAPLFile $snackAmpSettings(lastFolderFile) 1
         saLog "No play cache file found\n$result"
      }
   } else {
      waitWindow::show "Reading [file tail $snackAmpSettings(lastFolderFile)]"
      readAPLFile $snackAmpSettings(lastFolderFile) 1
   }
}

#------------------------------------------------------------------------------
#            R E M O V A B L E   M E D I A   S U P P O R T
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Function   : catalog::TagMedia
# Description: Return a hopefully unique tag for some removable media, note
#              if the media changes then the tags won't match.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::TagMedia {vol} {
   if {[catch {set first [lindex [lsort [glob -nocomplain [file join $vol *]]] 0]} result]} {
      return -code error $result
   }
   return [file tail "$first-[file mtime $first]"]
}
#------------------------------------------------------------------------------
# Function   : catalog::CatalogMedia
# Description: Reread the directories for files, include subdirectories
#              For each folder, add a menu entry
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc catalog::CatalogMedia {{vol ""}} {
   global TA snackAmpSettings
   if {$snackAmpSettings(UseRemovableFolders) == 0} {
      set ans [tk_messageBox -icon question -type yesno -title "Warning" \
         -message "You have Removable Media disabled, do you want to enable it?"]
      if {$ans == "yes"} {
         set snackAmpSettings(UseRemovableFolders) 1
         catalog::validateMediaPaths
      }
   }
   if {[llength $snackAmpSettings(RemovableFolders)] == 0} {
      set ans [tk_messageBox -icon question -type yesno -title "Warning" \
         -message "You need to setup your removable media first,\ndo you want to set them up now?"]
      if {$ans=="yes"} {
         package require saSettings
         after idle folderManager::launch RemovableFolders
         return
      } else {
         return
      }
   }
   if {$vol == ""} {
      set ::useFolder [lindex $snackAmpSettings(RemovableFolders) 0]
   } else {
      set ::useFolder $vol
   }
   toplevel .p  -class SnackAmp
   wm transient .p .saTop
   wm title .p "Removable Media Selector"
   pack [message .p.l -width 300 \
      -text "Select Removable Media folder, Enter Label for it, Insert Media into drive, then press OK to create a catalog for it"] \
      -side top -expand true -fill both
   listOptionMenu .p.vol ::useFolder $snackAmpSettings(RemovableFolders)
   .p.vol configure -width 20
   pack .p.vol -side top
   pack [entry .p.en -font entry -textvariable ::MediaLabel -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)] \
      -side top -fill x -expand true
   settings::regColor .p.en -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
   pack [frame .p.b -height 2 -borderwidth 2 -relief sunken] \
      -fill x -expand true -side top
   pack [frame .p.bb] -side top
   pack [button .p.bb.ok -width 8 -borderwidth 1  -relief raised -text OK -command {
      destroy .p
      set ::TA(continue) 1
   }] -side left -padx 6
   pack [button .p.bb.cancel -width 8 -borderwidth 1 -relief raised -text Cancel -command {
      destroy .p
      set ::TA(continue) 0
   }] -side left -padx 6
   focus .p.en
   bind .p <Return> [list .p.bb.ok invoke]
   bind .p <Escape> [list .p.bb.cancel invoke]
   centerWindow .p
   myIcon .p
   # Wait for user to close window
   grab .p
   tkwait visibility .p
   raise .p
   set ::TA(continue) 0
#  ----------------------------------------------------------------------------
#  Blocking
#  ----------------------------------------------------------------------------
   vwait ::TA(continue)
   # Cancel key hit
   if {$::TA(continue) == 0} {
      return
   }
   # No label
   set ::MediaLabel [string trim $::MediaLabel]
   if {[string length $::MediaLabel]==0} {
      set result [tk_messageBox -icon question -type yesno -title "No Label" \
         -message "You did not specify a catalog label, do you want to start over?"]
      if {$result == "yes"} {
         after 100 catalog::CatalogMedia $::useFolder
      } else {
         return
      }
   }
   ##
   # Now catalog the media
   #
   set ::mediaList [list]
   # save folder settings
   set Tag [file join $::HOME $::MediaLabel.rmc]
   if {[file exists $Tag]} {
      set result [tk_messageBox -icon question -type yesno -title "File Exists" \
         -message "Select Yes to overwrite file $Tag. No to to try again"]
      if {$result == "no"} {
         after 100 catalog::CatalogMedia
         Wait 0
         return
      }

   }
   # Prompt user for Name (TBD)
   if {[catch {catalog::TagMedia $::useFolder} result]} {
      tk_messageBox -icon warning -type ok -title "Tag Problem" \
         -message "Could not create Tag for $::useFolder\nIt may be blank\n$result"
      return
   }
   Wait 1
   waitWindow::show "Cataloging $::useFolder ...Please Wait"
   update
   # useFolder has to end in / for this to work (they should though)
   CatalogMediaRecurse $::useFolder [llength [file split $::useFolder]]
   if {[catch {set fid [open $Tag w]} result]} {
      waitWindow::close
      tk_messageBox -icon warning -type ok -title "Access Problem" \
         -message "Could not gain write access to $Tag"
      Wait 0
      waitWindow::close
      return
   }
   fconfigure $fid -buffersize 100000

   puts $fid [catalog::TagMedia $::useFolder]
   puts $fid $::mediaList
   close $fid
   unset ::mediaList
   aplFolder::createTeFolders 0 1
   catalog::validateMediaPaths
   Wait 0
   waitWindow::close
   update
   catch {destroy .wFC}
   tk_messageBox -icon info -type ok -title "Removable Media Catalog" \
      -message "Catalog of '$::MediaLabel' Complete"

}
#------------------------------------------------------------------------------
# Function   : CatalogMediaRecurse
# Description: Recursively scan a directory and build a list of all files
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::CatalogMediaRecurse {dir strip} {
   global TA LB mediaList
   set dirList [list]
   Update       ;# Allow music to play and update display
   foreach {file} [glob -nocomplain [globEscape [file join $dir *]]] {
      if {[file isdirectory $file]} {
         lappend dirList $file
      } elseif {[catalog::isTrackFile $file]} {
         set appendFile [eval file join [lrange [file split $file] $strip end]]
         lappend mediaList $appendFile
      }
   }
   # process each directory recursively
   foreach {file} $dirList {
      CatalogMediaRecurse [file join $dir $file] $strip
   }
}
#------------------------------------------------------------------------------
# Function   : readMediaCatalogs
# Description: Read the disk based media catalogs for the previously indexed
#              media folders. Build up proper name
# Author     : Tom Wilkason
# Date       : 3/5/2001
#------------------------------------------------------------------------------
proc catalog::readMediaCatalogs {args} {
   set catalogList [list]
   foreach {File} [glob -nocomplain [globEscape [file join $::HOME *.rmc]]] {
      if {[catch {set fid [open $File r]} result]} {
         waitWindow::close
         tk_messageBox -icon warning -type ok -title "$File Open Error" -message $result
         continue
      }
      # Save the tag for this description
      set mediaTag [gets $fid]
      set Entries [gets $fid]
      close $fid
      set Description [file rootname [file tail $File]]
      set ::MountTagToDescription($mediaTag) $Description
      foreach {Item} $Entries {
         lappend catalogList [file join $::RMT $Description $Item]
      }
   }
   return [lsort $catalogList]
}
#------------------------------------------------------------------------------
# Function   : findIfExists
# Description: Return the file name of a song if it really exists,
#              attempt to translate the name into a removeable media mount
#              even possibly asking the user to mount the volume.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
#proc catalog::findIfExists {fullName} {
#    if {![catch {file attributes $fullName -longname} fullName] || ([set fullName [TranslateMount $fullName]]!={})} {
#       return $fullName
#    } else {
#       return {}
#    }
#   return [TranslateMount $fullName]

#   if {[file isfile $fullName] || ([set fullName [TranslateMount $fullName]]!={})} {
#      return $fullName
#   } else {
#      return {}
#   }
#}
#------------------------------------------------------------------------------
# Function   : isARemMedia
# Description: Return true if the file is on a removable media
#              find if it is in a mounted folder. If it is not, then prompt the
#              user to mount it.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::isARemMedia {File} {
   return [string match "$::RMT*" $File]
}
#------------------------------------------------------------------------------
# Function   : TranslateMount
# Description: During a play, determine the true filename if it doesn't exist
#              find if it is in a mounted folder. If it is not, then prompt the
#              user to mount it.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::TranslateMount {File} {
   # Get lower two parts of name
   if {[string match [file join $::RMT *] $File]} {
      set parts [file split $File]
      # Rip off the RMC and Description parts
      # so we can replace them from the
      # mount lookups
      set base [eval file join [lrange $parts 0 2]]
      set passes 0
      while {1} {
         if {[info exists ::MountedMedia($base)]} {
            set result [eval file join [list $::MountedMedia($base)] [lrange $parts 3 end]]
            if {[file exists $result]} {
               # All is well
               return $result
            } else {
               # Prompt for mount if enabled, otherwise
               # try to silently mount once.
               # This shouldn't happen unless the the media
               # was swapped from the CD.
               if {$::snackAmpSettings(PromptRemovableFolders)} {
                  MountMedia "'[lindex $parts 2]'" 0
               } else {
                  if {$passes==0} {
                     MountMedia "'[lindex $parts 2]'" 1
                     incr passes
                  } else {
                     return {}
                  }
               }
            }
         } else {
            # Prompt for mount
            # This will happen if not yet mounted.
            if {$::snackAmpSettings(PromptRemovableFolders)} {
               MountMedia "'[lindex $parts 2]'" 0
            } else {
               if {$passes==0} {
                  MountMedia "'[lindex $parts 2]'" 1
                  incr passes
               } else {
                  return {}
               }
            }
         }
      }
   } else {
      if {[file exists $File]} {
         return $File
      } else {
         return {}
      }
   }
}
#------------------------------------------------------------------------------
# Function   : MountMedia
# Description: Go through all of the mountable volumes and determine
#              what is each of them.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::MountMedia {Message {quiet 0}} {
   global snackAmpSettings
   if {! $quiet} {
      if {$::snackAmpSettings(UseRemovableFolders) == 0} {
         set ans [tk_messageBox -icon question -type yesno -title "Warning" \
            -message "You have Removable Media disabled, do you want to enable it?"]
         if {$ans == "yes"} {
            set ::snackAmpSettings(UseRemovableFolders) 1
            catalog::validateMediaPaths
         } else {
            return
         }
      }
      if {[llength $::snackAmpSettings(RemovableFolders)] == 0} {
         set ans [tk_messageBox -icon question -type yesno -title "Warning" \
            -message "You need to setup your removable media first,\ndo you want to set them up now?"]
         if {$ans=="yes"} {
            package require saSettings
            after idle folderManager::launch RemovableFolders
            return
         } else {
            return
         }
      }
      set result [tk_messageBox -icon question -type okcancel \
         -message "Please insert $Message media into one or more of\n$::snackAmpSettings(RemovableFolders)" ]
      if {$result == "cancel" && $::snackAmpSettings(PromptRemovableFolders)} {
         set ans [tk_messageBox -icon question -type yesno \
            -message "To prevent furthur prompts can I turn off\n'Allow Removable Media Prompts'" ]
         if {$ans=="yes"} {
            set ::snackAmpSettings(PromptRemovableFolders) 0
         }
         return
      }
   }
   set Mounted {}
   set Empty {}
   catch {unset ::MountedMedia}
   foreach {vol} $::snackAmpSettings(RemovableFolders) {
      ##
      # Foreach media that has data in it.
      #
      if {![catch {set Tag [catalog::TagMedia $vol]} result]} {
         # Save a mapping from the tag to the vol name
         # Skip if find media that hasn't been indexed
         if {[catch {set ::MountedMedia([file join $::RMT $::MountTagToDescription($Tag)]) $vol} result]} {
            if {! $quiet} {
               saLog "Unknown Media Found:$Tag not in '[array names ::MountTagToDescription]'"
               set result [tk_messageBox -icon question -type yesno \
                  -message "Unknown Media Found:$Tag in $vol\ndo you want to catalog it?'" ]
               if {$result=="yes"} {
                  set ::useFolder $vol
                  after idle catalog::CatalogMedia
                  return
               }
               # Prompt user if they want to catalog it.
            }
         } else {
            append Mounted "'$::MountTagToDescription($Tag)' mounted on '$vol'\n"
         }
         # Lookup in Play is if file doesn't exist, then
         # Look for it in MountedMedia, if exists then map
         # to the vol name. Then correctCase if it exists again.
      } else {
         append Empty "\nMedia '$vol' is empty"
      }
   }
   if {! $quiet} {
      if {$Mounted != {}} {
         tk_messageBox -type ok -icon info \
         -message "The following volumes were mounted\n$Mounted$Empty"

      } else {
         tk_messageBox -type ok -icon warning \
         -message "No media was found in one of \n$::snackAmpSettings(RemovableFolders)$Empty"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::RenameFile
# Description: Try to rename/move a file, handle sound objects if they are playing
#              resume play if just renamed a playing file
# Author     : Tom Wilkason
# Date       : 3/22/2002
#------------------------------------------------------------------------------
proc catalog::RenameFile {oldName newName} {
   Trace
   set wasPlaying 0
   ##
   # Can't rename files on mounted media
   #
   if {[isARemMedia $oldName] || [isARemMedia $newName] || $::snackAmpSettings(dontMoveFiles)} {
      return -code error "\nYou don't have permission to rename or move $oldName"
   }
   ##
   # If the current song playing is the one to be renamed
   # we must unlink it from snack then later relink it
   #W: Need to add code to prevent writing to RMC files
   set justCaseChange [expr {[string equal -nocase $oldName $newName] && ($::tcl_platform(platform) == "windows")}]
   if {[file exists $newName] && !$justCaseChange} {
      if {[file isdirectory $newName]} {
         set res [tk_messageBox -icon question -type yesno -title "Target Exists" \
            -message "Warning: Folder $newName exists\nis it OK to overwrite any duplicate files?"]
      } else {
         soundControl::trackInfo oldInfo $oldName
         soundControl::trackInfo newInfo $newName
         # oldName is really latest file, newName will be existing file
         set textInfo "
     \tNew ->\tExisting
Len: \t$oldInfo(duration)\t$newInfo(duration)
Rate:\t$oldInfo(Rate)\t$newInfo(Rate)
Size:\t$oldInfo(size)\t$newInfo(size)
"
         if {($oldInfo(size) == $newInfo(size))} {
            set res [tk_messageBox -icon question -type yesno -title "File Exists" \
               -message "Warning: $newName exists\nand appears to be identical, do you still want to overwrite it?"]
         } else {
            set res [tk_messageBox -icon question -type yesno -title "File Exists" \
               -message "Warning: $newName\nexists, do you want to overwrite it?\n$textInfo"]
         }
      }
      if {$res != "yes"} {
         return -code error "$newName File Exists"
      }
   }
   set currentSound [soundControl::fileName]
   if {[string equal $currentSound $oldName]} {
      # This is a trick to see if the song is actually playing
      set wasPlaying [soundControl::isPlaying]
      set location [soundControl::samplePosition]
      soundControl::stop
      soundControl::reset
      set resumePlay 1
   } else {
      set resumePlay 0
   }
   ##
   # Rename the file, if on windows then allow of a case insensitive name
   # Note: -forces allows overwritting of files but supresses errors if
   # the overwrite failed. This messes up upper error handling.
   if {$justCaseChange} {
      if {[catch {file rename $oldName $oldName.tmp} result]} {
         return -code error $result
      } else {
         if {[catch {file rename $oldName.tmp $newName} result]} {
            return -code error $result
         }
         FlagDirtyCache $newName
         FlagDirtyCache $oldName
      }
   } else {
      # User may have inserted path separators to move into new folder
      set dirPart [file dirname $newName]
      if {![file exists $dirPart]} {
         if {[catch {file mkdir $dirPart} result]} {
            return -code error $result
         }
      }
      if {[file isdirectory $oldName] && [file isdirectory $newName]} {
#         saLog "Moving folders '$oldName' to existing folder '$newName'"
         set srcs [glob -nocomplain -directory $oldName *]
         set errors 0
         set Result ""
         foreach {file} $srcs {
            if {[catch {file rename -force -- $file $newName} result]} {
               append Result "\n$result"
               incr errors
            }
         }
         if {$errors > 0} {
            return -code error " ($errors) Move Errors $Result"
         }
         # Old folder will be left, can remove it if empty (if all went well)
         if {[file exists $oldName] && [llength [glob -nocomplain -directory $oldName *]]==0} {
            catch {file delete $oldName}
         }
       } else {
         # The target folder is new
         if {[catch {file rename -force -- $oldName $newName} result]} {
            return -code error $result
         }
      }
      FlagDirtyCache $newName
      FlagDirtyCache $oldName
   }
   ##
   # Resume playing if active song was renamed
   #
   if {$resumePlay} {
      soundControl::fileName $newName
      if {$wasPlaying} {
         # Resume play after calling routines finish
         after 1000 Play current $location
      }
   }
}
##
# Properties Dialog
#
proc catalog::properties {tracks} {
   set long [list]
   set short [list]
   foreach {track} $tracks {
      if {[file exists $track]} {
         lappend short [file tail $track]
         lappend long $track
      }
   }
   if {[llength $long] <= 0} {
      return
   }
   set i 0
   while {[info command .prop$i] != {}} {incr i}
   set top .prop$i
   toplevel $top
   wm title $top "Track Properties"
   if {[llength $long]==1} {
      set hull [labeledFrame $top.data -text [lindex $short 0]]
      set text [trackTextData [lindex $long 0] Details]
      label $hull.info -text $text -justify left -anchor nw
      pack $hull.info -expand true -fill both
      pack $top.data -expand true -fill both
   } else {
      set tof [tabbedOptionFrame::Create $top.tf -Width 40 -options $short -title "Track Properties"]
      pack $tof -expand true -fill both
      set i 0
      foreach {track} $long {
         set text [trackTextData $track Details]
         set hull [tabbedOptionFrame::getHull $tof $i]
         label $hull.info -text $text -justify left -anchor nw
         pack $hull.info -expand true -fill both
         pack $hull -expand true -fill both
         incr i
      }
   }
   update idletasks
   myIcon $top
}
#------------------------------------------------------------------------------
# Function   : catalog::trackTextData
# Description: Return the details of a song in both an array and a text
#              string suitable for balloon help.
# Author     : Tom Wilkason
# Date       : 11/20/2001
#------------------------------------------------------------------------------
proc catalog::trackTextData {fullName _Details} {
   upvar $_Details Details
   set Details(ext) ""
   ##
   # Note: Grab this from a database instead of here if it exists
   #
   set hasTag [db::lookup $fullName Details]
   if {[catch {
      set Details(duration) [clock format [expr {round($Details(Duration))}] -format "%M:%S"]
      set Details(ext) [string trimleft [string toupper [file extension $fullName]] .]
      if {$::snackAmpSettings(autoLevel) && [string is double $Details(sampRate)] && $Details(sampRate)>0 && $Details(Stop)} {
         append endData "Start\t: [format %.1f [expr {$Details(Start)/(1000*$Details(sampRate))}]] sec"
         append endData "\nStop\t: [format %.1f [expr { $Details(Duration)- ($Details(Stop)/(1000*$Details(sampRate))) }]] sec before end\n"
      }
      append endData "Rating\t: "  $Details(Rating)
      append endData "\nVotes\t: $Details(Votes)"
      if {$Details(Played) > 0} {
         append endData "\nPlayed\t: "  $Details(Played) " Times"
      } else {
         append endData "\nPlayed\t: Not yet Played"
      }
      if {$Details(LastPlay) > 0} {
         append endData ", Last Played: "  [clock format $Details(LastPlay) -format $::snackAmpSettings(timeString)]
      }
   } result]} {
      return "No Data for $fullName available"
   }
   switch -- $Details(ext) {
      MP3 {append Details(Rate) " kbps"}
      OGG {append Details(Rate) " kbps"}
      WAV {append Details(Rate) " kHz"}
      default {set Details(Rate) "Unknown"}
   }
   append Details(sampRate) " Khz"
   ##
   # File info
   #
   if {[catch {file stat $fullName Stat} result]} {
      return Unknown
   }
   set Details(fSize) "[commify $Stat(size)] Bytes"
   set Details(fDate) [clock format $Stat(mtime) -format $::snackAmpSettings(timeString)]
   ##
   # Balloon help
   #
   set Info         "File\t: [file tail $fullName]\nFolder\t: [file dirname $fullName]\n"
   if {$hasTag} {
      append Info [id3Tag::formatID3Data Details]
      set Details(ID3) $Info
   } else {
      set Details(ID3) {}
   }
   id3Tag::appendIfNotBlank Info "Size\t:"     $Details(fSize)
   id3Tag::appendIfNotBlank Info "Date\t:"     $Details(fDate)
   id3Tag::appendIfNotBlank Info "Duration\t:" $Details(duration)
   id3Tag::appendIfNotBlank Info "Bitrate\t:"  $Details(Rate)
   id3Tag::appendIfNotBlank Info "Freq\t:"     $Details(sampRate)
   if {$::snackAmpSettings(autoLevel)} {
      id3Tag::appendIfNotBlank Info "Gain\t:"  $Details(Gain)
   } else {
      set Details(levelFactor) "Not Set"
   }
   append Info $endData
   return [string trim $Info]
}
#------------------------------------------------------------------------------
# Function   : trackFullData
# Description: Return the details of a song in an array for indexing
#              music files.
# Author     : Tom Wilkason
# Date       : 11/20/2001
#------------------------------------------------------------------------------
proc catalog::trackFullData {fullName _Details} {
   variable TableCols
   upvar $_Details Details
   ;# Blank out desired columns
   foreach {index} $::TA(cols) {
      set Details($index) {}
   }
   ##
   # Retrieve the track sound info
   #
   soundControl::trackInfo Details $fullName
   if {[catch {gainControl::get $fullName} gain]} {
      set Details(Gain) "--"
   } else {
      set Details(Gain) [gainControl::levelText $gain]
   }
   ##
   # Get additional file and ID3 info
   #
   set Details(File)   [file tail $fullName]
   ##
   # ID3 Data, if no tag then set Title = File root name
   #
   set result [id3Tag::id3Label $fullName Details]
   if {$result==0} {
      set Details(Title) [file rootname $Details(File)]
   }
   return $result
}
#------------------------------------------------------------------------------
# Function   : DeleteFile
# Description: Delete a file, if it is playing then handle it
# Author     : Tom Wilkason
# Date       : 3/24/2002
#------------------------------------------------------------------------------
proc catalog::DeleteFile {fileName} {
   set currentSound [soundControl::fileName]
   set goNext 0
   if {[string equal $currentSound $fileName]} {
      # This is a trick to see if the song is actually playing
      if {[soundControl::isPlaying]} {
         soundControl::stop
         set goNext 1
      }
      soundControl::reset
   }
   # Now try to delete it
   if {[catch {
      file delete -force $fileName
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -message "Error removing $fileName\n$result"
   } else {
     catalog::removeFile $fileName
     FlagDirtyCache $fileName
   }
   return $goNext
}

#------------------------------------------------------------------------------
# Function   :Remove
# Description: Remove a file from the master and playlists list if it exists
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::removeFile {File} {
   global PLB
   # Remove from the master list
   set index [db::whereInDb $File]
   if {$index >= 0} {
      ##
      # Remove from the master list
      #
      db::removeFile $index
      ##
      # Remove from the Playlist
      #
      LB::removeFile $File
      ##
      # Remove from the APL File List (in case it is there)
      #
      catch {aplLB::RemoveFile $File}
   }
}
#------------------------------------------------------------------------------
# Function   : renameInPlaceDialog
# Description: Bring up a in place editing window on a listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
#------------------------------------------------------------------------------
proc catalog::renameInPlaceDialog {W table {oldName "xxx"}} {
   if {![winfo ismapped $W]} {
      return ""
   }
   ##
   # Need to compute location for dialog
   #

   set active [$W index active]
   set bba [$W bbox $active]
   if {[llength $bba]<4} {
      return ""
   }
   set w  .rndialog
   destroy $w
   toplevel $w  -class SnackAmp
   foreach {x y - -} $bba {break}
   # If table, adjust for heading
   if {$table} {
      set y [expr {$y - [winfo y $W]}]
   }
   # Left Hand alignment
   set rx [winfo rootx $W]
   set ry [winfo rooty $W]
   set X [expr {$rx+$x}]
   set Y [expr {$ry+$y}]
   # Make it look like the text underneath
   set font [$W cget -font]
   set bg [$W cget -background]
   set fg [$W cget -foreground]
   set ::TA(rename) $oldName
   # Set size available width and linespacing+2
   set wid [expr {[string length $oldName]+5}]
   set Wwid [winfo width $W]
   set Wheight [expr {[font metrics $font -linespace]+2}]
   pack [entry $w.rn -bd 2 -relief groove -font $font -insertbackground $fg \
      -width $wid -bg $bg -fg $fg -textvariable ::TA(rename)] -expand true -fill x
   #$w.rn selection range 0 end
   wm withdraw $w
   wm overrideredirect $w 1
   if {[string length [focus]]} {
      wm transient $w [winfo toplevel [focus]]
   }
   focus -force $w.rn
   wm geometry $w "${Wwid}x$Wheight+$X+$Y"
   bind $w <Escape> {
      destroy .rndialog
      set ::TA(rename) ""
   }
   bind $w <FocusOut> {
      destroy .rndialog
      set ::TA(rename) ""
   }
   bind $w <Return> [list destroy .rndialog]
   wm deiconify $w
   # Force on top if possible
   update idletasks
   if {$ontop::enabled && [ontop::state $W]} {
      ontop::control $w 1
   }
   tkwait window $w
   # Waiting for window to go away
   return $::TA(rename)
}

#------------------------------------------------------------------------------
# Function   : catalog::RenameEntry
# Description: Replace a file in the master list with another one only if it exists
#              Assume file itself has already been renamed.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::RenameEntry {oldFile newFile} {
   Trace
   global PLB
   # If it doesn exist, such as incoming data
   # then go ahead and add it to master list
   # Note: At this point, the actual file is renamed to newFile, so oldFile no longer exists
   set index [db::addToDbIfNeeded $oldFile]  ;# Bug: oldFile is no longer
   if {$index < 0} {
      set index [db::whereInDb $oldFile]
   }
   if {$index >= 0} {
      set tail [file tail $oldFile]
      ##
      # Replace equalization if tail of file was renamed
      #
      gainControl::move $oldFile $newFile
      ##
      # Replace full name in master list
      #
      db::renameFile $index $newFile
      catalog::FlagDirtyCache $newFile
      ##
      # Update Main Play List
      #
      LB::changePlaylistFiles [list $oldFile $newFile]
      ##
      # Update APL oldFile List (if visible in list)
      #
      if {[APLExists]} {
         ##
         # Schedule a rescan to be done at the next event, if one is pending
         # then don't schedule another. Not very clean but it works.
         #W: Move this to the trackList namespace as a function
         # trackList::changeTable $old $new
         set newTail [file tail $newFile]
         set newData [list]
         set changed 0
         if {[info exists trackList::TABLEDATA]} {
            foreach row $trackList::TABLEDATA {
               set cc 0
               set key [lindex $row 0]  ;# Used to lookup full file name, key is hidden column
               foreach col $row {
                  # If col is old file name, replace it with the new one
                  if {[string equal $col $tail]} {
                     lset row $cc $newTail
                     # Replace name entry for this row key
                     #set trackList::Index2File($key) $newFile
                     incr changed
                  }
                  incr cc
               }
               lappend newData $row
            }
            # If one or more files renamed, then re-insert the data by set the TABLEDATA
            # lined variable to the new table data.
            if {$changed} {
               set trackList::TABLEDATA $newData
               # Schedule a rescan when all is said and done, don't re-populate
               # since we did in place above, just update the cache file.
               trackList::populateHighighted 0 0
            }
         }
      }
      db::dirtyList
   } else {
      saLog "Could not find $oldFile in list to replace"
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::moveFolder
# Description: Move a folder (and its files) to some other place
#              and adjust the APL tree accordingly.
#              folder is existing full name
#              target is new full name
#              only called by APL code
# Author     : Tom Wilkason
# Date       : 3/3/2002
#------------------------------------------------------------------------------
proc catalog::moveFolder {folder target} {
   Trace
   global LB
   ##
   # Don't allow moving of removeable media folder
   #
   if {[isARemMedia $folder] || [isARemMedia $target]} {
      return
   }
   # Append / to folder so we don't match
   # Other folders with same base name
   # eg. Classic vs. Classic Rock
   set folder [string trimright $folder /]
   set target [string trimright $target /]
   if {[string equal $folder $target]} {
      return
   }
   APL::busyCursor 1
   update idletasks
   if {[catch {catalog::RenameFile $folder $target} result]} {
      debug $result
      APL::busyCursor 0
      return -code error $result
   }
   # Remove old folder name
   APL::delItem $folder
   ##
   # Replace teFolder entries with new name
   #
   set subLen [expr {[string length $folder]-1}]
   set i -1
   set pattern [file join $folder *]
   foreach path [aplFolder::foldersMatchingPattern * 1] {
      incr i
      if {[string match $pattern $path] || [string equal $folder $path]} {
         set newName [string replace $path 0 $subLen $target]
         aplFolder::rename $path $newName
         APL::newItem $newName
      }
   }
   ##
   # Replace each matching entry inList and update playlist and APL file list
   # if affected songs were shown in them.
   #
   set i -1
   # Append to folder for pattern matching
   set i 0
   #W: Function to return indicies of match
   set changed [list]
   foreach {file} [db::filesMatchingPattern $pattern] {
      set newName [string replace $file 0 $subLen $target]
      incr i
      db::renameFile $file $newName
      #W: This could be improved to pass in a paired list so we
      # don't have to call this function for each file.
      lappend changed $file $newName
   }
   LB::changePlaylistFiles $changed
   if {$i > 0} {
      db::dirtyList
   }
   APL::busyCursor 0
}
#------------------------------------------------------------------------------
# Function   : catalog::moveClipboardFiles
# Description: Move the files in the clipboard to some new folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::moveClipboardFiles {target} {
   Trace
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return ""
   }
   clipboard clear
   set files [list]
   # Ignore clibboard entries that aren't files
   # Ignore clibboard entries are duplicates
   # Ignore errors due to funky clipboard data
   catch {
      foreach {file} [lflip $tryFiles] {
         if {[lsearch -exact $files $file] < 0 && [file exists $file]} {
            lappend files $file
         }
      }
   }
   set moved 0
   soundControl::latency long
   if {[llength $files] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to move files or folders"
         return 0
      }
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to move:\n[join $files \n]\n to\n$target"]
      if {$result == "yes"} {
         set errors [list]
         ##
         # For each file to be moved, move it
         #
         foreach file $files {
            set tail [file tail $file]
            set newName [file join $target $tail]
            waitWindow::show "Moving $tail ... Please Wait"
            update
            ##
            # If source is a folder, move it otherwise
            # move the file
            #
            if {[file isdirectory $file]} {
               if {[catch {
                  if {![string equal $file $target]} {
                     catalog::moveFolder $file $newName
                  }
               } result] \
               } then {
                  debug $result
                  lappend errors $file
               }
            } else {
               if {![string equal $file $newName]} {
                  if {[catch {catalog::RenameFile $file $newName} result]} then {
                     debug $result
                     lappend errors $file
                  } else {
                     #W: This can be slow on big moves, you have a full lookup
                     # for each file moved.
                     catalog::RenameEntry $file $newName
                     incr moved
                  }
               }
            }
         }
         waitWindow::close
         # Flag these two areas as dirty
         catalog::FlagDirtyCache $target
         catalog::FlagDirtyCache $newName
         # Handle any errors
         if {[llength $errors] > 0} {
            tk_messageBox -type ok -icon warning -message "The following files were not moved\n[join $errors \n]"
         }
      }
   }
   if {$moved > 0} {
      db::dirtyList
   }
   soundControl::latency normal
   return $moved
}

#------------------------------------------------------------------------------
# Function   : renameFolder
# Description: Prompt then rename a subfolder under some level
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::renameFolder {folder} {
   set tail [file tail $folder]
   set base [file dirname $folder]
   set newName $tail
   set res 0
   if {[getEntryDialog $newName newName "Rename Folder"]=="ok" && ![string equal $tail $newName]} {
      if {[catch {
         set newFullName [file join $base $newName]
         moveFolder $folder $newFullName
         if {[APLExists]} {
            APL::treeSelect $newFullName
            APL::viewHighlighted $::TREE
         }
         catalog::FlagDirtyCache $newFullName
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not rename folder $newName\n$result"
         set res 0
      }
   }
   return $res
}

#------------------------------------------------------------------------------
# Function   : removeFolder
# Description: Remove a folder and all of its children files
#              Only used by APL tree
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::removeFolder {folder} {
   set i 0
   if {$::snackAmpSettings(dontMoveFiles)} {
      tk_messageBox -icon warning -type ok \
         -message "You don't have permission to remove folders"
      return
   }

   set pattern [file join $folder *]
   # If a removeable file in name then check if at base level and if so
   # verify before removing the catalog file from disk
   set contents [glob -nocomplain $pattern]
   set i [llength $contents]  ;# don't just check music tracks
   set parts [file split $folder]
   if {[isARemMedia $folder] && [llength $parts] == 3} {
      set message "Are you sure you want to remove the removable catalog $folder"
      set catalog 1
      set icon question
   } else {
      set message "Are you sure you want to remove $folder and its decendents, it contains $i files\n\nAll files and folders will be permanently deleted!"
      set catalog 0
      set icon warning
   }
   if {$i == 0 || [tk_messageBox -icon $icon -title "Please Respond" -type yesno -message $message] == "yes"} {
      ##
      # Remove the files & icon
      #
      APL::delItem $folder
      ##
      # Handle the files
      #
      foreach {file} $contents {
         if {[catch {
            DeleteFile $file
         } result]} then {
            debug $result
         }
      }
      catalog::FlagDirtyCache $folder
      ##
      # Handle the folders, work from the longest folder back
      # to maximize the deleted folders
      foreach path [lsort -decreasing [aplFolder::foldersMatchingPattern $folder*]] {
         if {[string match $pattern $path] || [string equal $folder $path]} {
            if {[catch {
               file delete -force $path
               aplFolder::Unset $path
            } result]} then {
               debug $result
            }
         }
      }
      APL::treeSelect [file dirname $folder]
      aplLB::UpdateList [file dirname $folder]
      FlagDirtyCache $folder
      gui::updateFileDisplay
      if {$catalog} {
         # Remove the RM catalog
         set fname [file join $::HOME [lindex $parts 2].rmc]
         file delete $fname
      }
   }
   return
}
#------------------------------------------------------------------------------
# Function   : catalog::SavePlaylistCache
# Description: Cache the current playlist for later restoration
# Author     : Tom Wilkason
# TBD: Save/restore extended info if it is faster than a lookup
#------------------------------------------------------------------------------
proc catalog::SavePlaylistCache {args} {
   global TA
   set sList [LB::keyOrNamesAll 1]
   set loc [LB::idxPlaying]
   if {$loc < 0} {
      set loc [LB::idxActive]
   }
   append sList "\n$loc"
   append sList "\n[soundControl::samplePosition]"
   append sList "\n$soundControl::playState"
   if {[catch {safeWrite $TA(playCacheFile) $sList} result]} {
      return -code error "Save Error: Unable to open '$TA(playCacheFile)' for writing\n"
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::ReadPlaylistCache
# Description: Restore the playlist from a cached file (during initialization)
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::ReadPlaylistCache {args} {
   global TA LB
   if {[catch {open $TA(playCacheFile) r} InFile]} {
      return -code error "Read Error: Unable to open '$TA(playCacheFile)' for reading\n"
   }
   fconfigure $InFile -buffersize 300000
   set sList  [gets $InFile]
   set cIndex [gets $InFile]
   # Restore position within the file
   if {[eof $InFile]} {
      set cPosition -1
   } else {
      set cPosition [gets $InFile]
   }
   if {[eof $InFile]} {
      set playState "Stopped"
   } else {
      set playState [gets $InFile]
   }
   close $InFile
   # Replace current playlist very fast
   LB::deleteAll
   if {[llength $sList] > 0} {
      LB::insertList end $sList
      gui::updateFileDisplay
      if {$::snackAmpSettings(resumePlay)} {
         if {$playState=="Playing"} {
            after idle catalog::resumePlay $cIndex $cPosition
         } elseif {$playState=="Paused"} {
            after idle catalog::resumePlay $cIndex $cPosition 1
         }
      } else {
         macros::Highlight $cIndex
      }
      saLog "Restored playlist cache '$TA(playCacheFile)' of [llength $sList] files"
   }
}
##
# Function to resume play
#
proc catalog::resumePlay {index offset {pause 0}} {
   macros::Highlight $index
   Play active $offset $pause
}
#------------------------------------------------------------------------------
# Return a list of all auto-playlist files
#------------------------------------------------------------------------------
proc catalog::getAutoPlaylists {args} {
   return [lsort -dictionary [glob -nocomplain -directory $::HOME *.tafs]]
}
#------------------------------------------------------------------------------
# Return a list of all playlist files
#------------------------------------------------------------------------------
proc catalog::getPlaylists {args} {
   return [lsort -dictionary [glob -nocomplain -directory $::HOME *.taf *.m3u]]
}

#----------------------------------------------------------------------------
# Utility routines
#----------------------------------------------------------------------------
proc tester {{num 5}} {
   set TRACKS [db::sortedList]
   puts [time {
      foreach {file} $TRACKS {
         set dir [file dirname $file]
      }
   } $num]
   puts [time {
      foreach {file} $TRACKS {
         set dir [file root $file]
      }
   } $num]
   puts [time {
      foreach {file} $TRACKS {
         set dir [file tail $file]
      }
   } $num]
}
#------------------------------------------------------------------------------
# Function   : catalog::ShowDups
# Description: Command line routine to show potential duplicate files
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc catalog::ShowDups {{pattern *}} {
   foreach {file} [db::filesMatchingPattern $pattern] {
      set dir [file dirname $file]
      set tail "$dir/[lrange [split $file -] end-1 end]"
      set fileTail [lrange [file split $file] end-1 end]
      if {![info exists tailHash($tail)]} {
         set tailHash($tail) $file
      } else {
         update
         saLog "Duplicate $tail\n   '$file'\n   '$tailHash($tail)'\n"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::DupFiles
# Description: Command line routine to show potential duplicate files
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc catalog::DupFiles {{pattern *}} {
   set patt [file join $::RMT *]
   foreach {file} [db::sortedList] {
      if {[string match $pattern $file] && ![string match $patt $file]} {
         set tail [lindex [file split $file] end]
         if {![info exists tailHash($tail)]} {
            set tailHash($tail) $file
         } else {
            update
            saLog "Duplicate $tail\n   '$file - [file size $file]'\n   '$tailHash($tail) - [file size $tailHash($tail)]'\n"
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::DupSizes
# Description: Command line routine to show potential duplicate files
#              based on like file sizes. Good for finding renamed files.
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc catalog::DupSizes {{pattern *} {threshold 2}} {
   foreach {file} [db::filesMatchingPattern $pattern] {
      if {![catch {set tail [file size $file]} result]} then {
         lappend tailHash($tail) $file
      }
   }

   foreach entry [array names tailHash] {
      set files  $tailHash($entry)
      if {[llength $files] > 1} {
         set last [expr {[llength $files]-1}]
         set matches 0
         # Pass through each set of duplicate files
         for {set i 0} {$i < $last} {incr i} {
            set base [split [string map {" " ""} [string tolower [file tail [lindex $files $i]]]] -]
            set next [split [string map {" " ""} [string tolower [file tail [lindex $files [incr i]]]]] -]
            foreach part $base {
               if {[lsearch -exact $next $part] > -1} {
                  incr matches
               }
            }
         }
         if {$matches > $threshold} {
            puts "Duplicates found for size $entry ($matches)"
            foreach file $files {
               puts "   $file"
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Read the cached files into memory
# Set up the default teFolders
# Returns a list of paths not found
#------------------------------------------------------------------------------
proc catalog::convertCacheFiles {args} {
   global TA TFList snackAmpSettings
   ##
   # Restore the cached files from each media Dir
   # Support either new type (a list) or old type (Array)
   set result [list]
   set TFList [list]
   # Move list into a hash table for faster access
   # and to allow replacing parts by single path
   catalog::validateMediaPaths
   foreach {Path} [lsort $TA(mediaDirs)] {
      waitWindow::show "Importing Directory $Path - Please Wait"
      update
      # Read either v1 or v2 cache files
      if {![file exists $Path]} {
         waitWindow::close
         tk_messageBox -icon warning -type ok -title "Warning" \
            -message "Media Folder '$Path' does not exist,\nit may be named improperly or on \nan unavailable network computer."

      } else {
         if {[catch {set fid [open [file join $Path $TA(cacheFile)] r]}]} {
            lappend result $Path
         } else {
            # V2 files
            fconfigure $fid -buffersize 100000
            gets $fid tLlist
            close $fid
            saLog "Converting $Path (v2)"
            # add the new ones for this path back
            set i 0
            set total [llength $tLlist]
            set jump [expr {$total/100}]
            foreach {File} $tLlist {
               set file [file join $Path $File]
               if {$i%$jump==0} {
                  waitWindow::show "[expr {($i*100)/$total}]% Complete Importing $Path ... Please Wait"
                  update
               }
               db::cacheFileName $file
               incr i
            }
            db::commit
         }
      }
   }
   return
}
#------------------------------------------------------------------------------
# Function   : fixplus::correctCase
# Description: Replace + in file names (due to http download), also fix other
#              case problems and clean up spacing.
# Author     : Tom Wilkason
# Date       : 1/8/2001
#------------------------------------------------------------------------------
namespace eval fixplus  {}
proc fixplus::correctCase {folder} {
   set changed 0
   # Use a folder or the selected names
   if {[file isdirectory $folder]} {
      set Files [glob -nocomplain -directory $folder -types {f l} *]
   } else {
      set Files $folder
   }
   foreach File $Files {
      set folder [file dirname $File]
      set File   [file tail $File]
      ##
      # Clean up the file name
      #
      set ext [file extension $File]
      set root [file rootname $File]   ;# base part of name
      if {[catalog::isTrackFile $File]} {
         # Clean out funny stuff
         set newRoot [cleanName $root]
         set ofn [file join $folder $File]
         set fnf [file join $folder "$newRoot$ext"] ;#full new file
         ##
         # Have new file name
         #
         if {![string equal $newRoot $root]} {
            ##
            # Try to rename the file here
            #
            if {[catch {catalog::RenameFile $ofn $fnf} result]} {
               saLog "correctCase Error:$result"
            } else {
               catalog::RenameEntry $ofn $fnf
               incr changed
            }
         }
      }
   }
   if {$changed > 0} {
      db::dirtyList
   }
   return $changed
}
#------------------------------------------------------------------------------
# Function   : cleanName
# Description: Clean up a file name to use - as separator
# Author     : Tom Wilkason
# Date       : 3/19/2002
#------------------------------------------------------------------------------
proc fixplus::cleanName {root} {
   # sp is user supplied split char, "-" is recommended
   set sp [string trim $::snackAmpSettings(splitChar)]
   set root [string map {[ ( ] )} $root]
   regsub -all -- {[+_ ]+} $root { } root
   regsub -all -- {( *- *)|( *\+ *)} $root " $sp " root
   set root [string trim $root]
   # Fix first part if a number
   set reExp [subst -nocommand {^([0-9]+)[$sp ]*(.*?)}]
   if {[regexp -- $reExp $root -> num rest]} {
      set root "$num $sp $rest"
   }
   # Cap each token
   set root [stringTitle $root]
   return $root
}
#  ----------------------------------------------------------------------------
#  aplFolder namespace: Wrap access to the auto-playlist folder tree settings
#  ----------------------------------------------------------------------------

namespace eval aplFolder {
}
#------------------------------------------------------------------------------
# Function   : foldersMatchingPattern
# Description: Return a list of folders in or under a path
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::foldersMatchingPattern {{pattern *} {sorted 0}} {
   if {$sorted} {
      return [lsort [array names ::teFolders $pattern]]
   } else {
      return [array names ::teFolders $pattern]
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::getFolders
# Description: Return a list of some folders and subfolders of subfolders 'subfolders'
# Author     : Tom Wilkason
# Date       : 2/16/2003
#------------------------------------------------------------------------------
proc aplFolder::getFolders {selectedFolder {subfolders 0}} {
   lappend folders $selectedFolder
   # Find any remaining subfolders
   if {$subfolders>0} {
      set baseDepth [llength [file split $selectedFolder]]  ;#0 is just this folder
      foreach {index} [aplFolder::foldersMatchingPattern $selectedFolder/*] {
         set thisDepth [llength [file split $index]]
         if {$thisDepth <= ($baseDepth+$subfolders)} {
            lappend folders $index
         }
      }
   }
   # It would be nice to have -dictionary, but the sorted search will fail
   return [lsort $folders]
}
#------------------------------------------------------------------------------
# Function   : folderArray
# Description: Return an array (index element) of folders in or under a path
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::folderArray {{pattern *}} {
   return [array get ::teFolders $pattern]
}
#------------------------------------------------------------------------------
# Function   : isSet
# Description: Return true if the folder item is set (and exists)
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::isSet {dirName {value 1}} {
   if {[info exists ::teFolders($dirName)] && ($::teFolders($dirName) == $value)} {
      return 1
   } else {
      return 0
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::Exists
# Description: Veriy the entry or the actual folder exists
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Exists {Path {check 0}} {
   if {$check} {
      return [expr {[info exists ::teFolders($Path)] || [file isdirectory $Path]}]
   } else {
      return [info exists ::teFolders($Path)]
   }
}
#------------------------------------------------------------------------------
# Function   : Unset
# Description: Remove one or moer entries
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Unset {key} {
   unset ::teFolders($key)
}
#------------------------------------------------------------------------------
# Function   : stripBranch
# Description: Remove one or moer entries
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::stripBranch {pattern} {
   array unset ::teFolders $pattern
}
#------------------------------------------------------------------------------
# Function   : SetSome
# Description: Set some values to a value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::SetSome {pattern value} {
   foreach {key} [array names ::teFolders $pattern] {
      set ::teFolders($key) $value
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::Set
# Description: Set an entry to some value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Set {key value} {
   set ::teFolders($key) $value
}
#------------------------------------------------------------------------------
# Function   : aplFolder::SetIfBlank
# Description: Set an entry to some value only if it doesn't already exist
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::SetIfBlank {key value} {
   if {![info exists ::teFolders($key)]} {
      set ::teFolders($key) $value
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::SetIfExists
# Description: Set an entry to some value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::SetIfExists {key value} {
   if {[info exists ::teFolders($key)]} {
      set ::teFolders($key) $value
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::Get
# Description: Set an entry to some value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Get {key} {
   return $::teFolders($key)
}
#------------------------------------------------------------------------------
# Function   : aplFolder::rename
# Description: Rename an entry retaining its old value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::rename {path newName} {
   set value $::teFolders($path)
   unset ::teFolders($path)
   set ::teFolders($newName) $value
}
#----------------------------------------------------------------------------
# Return the number of tree entries
#----------------------------------------------------------------------------
proc aplFolder::entries {args} {
   return [array size ::teFolders]
}
proc aplFolder::savePattern {pattern} {
   return [array get ::teFolders $pattern]
}
proc aplFolder::restoreNames {saveNames} {
   foreach {folder value} $saveNames {
      if {[aplFolder::Exists $folder]} {
         aplFolder::SetIfExists $folder $value
      }
   }
}
#------------------------------------------------------------------------------
# Function   : db::createTeFolders
# Description: Create all the teEntries and set them to 0 for the
#              database.
#              if a parent doesn't exist, create one and set it to 0
# Author     : Tom Wilkason
# Date       : 8/8/2003
#
#------------------------------------------------------------------------------
# bug this is 10x slower when using metakit, looks like file dirname is much slower
# not much faster than using db::sortedList
proc aplFolder::createTeFolders {{how 0} {fixMount 0}} {
   Trace
   global teFolders
   set mediaPatterns [list]
   foreach {dir} $::TA(mediaDirs) {
      lappend mediaPatterns [file join $dir *]
   }

   set hits 0
   # Looping is very fast, the [file dirname] is 3x slower
   # move to database
   $db::dbObj loop row {
      foreach {folder Mounted} [$db::dbObj get $row Folder Mounted] {break}
      ##
      # Correct any mounted media
      #
      if {$fixMount} {
         set imd [isInMediaDir $folder $mediaPatterns]
         if {$imd!=$Mounted} {
            $db::dbObj set $row Mounted $imd
            set Mounted $imd
         }
      }
      if {$Mounted==1} {
         while {![info exists teFolders($folder)]} {
            incr hits
            set teFolders($folder) $how
            # Backup one
            set root [file dirname $folder]
            if {[string equal $folder $root]} {
               break
            } else {
               set folder $root
            }
         }
      }
   }
   return $hits
}
##
# Check to see if a track is in a valid media folder
#
proc aplFolder::isInMediaDir {file mediaPatterns} {
   foreach {patt} $mediaPatterns {
      if {[string match $patt $file]} {
         return 1
      }
   }
   return 0
}
#------------------------------------------------------------------------------
# Function   : setTeDefault (recursive)
# Description: Fixup a branch of teFolders that may be undetermined
#              Work backwards and set the childrens value to that
#              of the parent all the way to the root. Default to 0
# Author     : Tom Wilkason
# Date       : 3/24/2001
#------------------------------------------------------------------------------
proc aplFolder::setTeDefault {Index} {
   Trace
   global teFolders
   switch -- $teFolders($Index) {
      -1 {
         # Folder wasn't in new APL file,
         # Set it to it's parents value if possible
         set parent [file dirname $Index]
         if {$parent != $Index} {
            if {[info exists teFolders($parent)]} {
               set teFolders($Index) [aplFolder::setTeDefault $parent]
            } else {
              set teFolders($Index) 0
            }
         } else {
            set teFolders($Index) 0
         }
      }
      1 {
         # Folder is set on, if the parent was set off change
         # the parent to a partial state (we know it can't be off)
         set parent [file dirname $Index]
         if {($parent != $Index) && [aplFolder::isSet $parent 0]} {
            set teFolders($parent) 2
         }
      }
      2 {
         # If it was set partial before, just turn it off
         set teFolders($Index) 0
      }
      default {}
   }
   return $teFolders($Index)
}
##
# Verify each tefolder entry (very slow)
#
proc aplFolder::check {args} {
   foreach {folder} [array names ::teFolders] {
      if {![file isdirectory $folder]} {
         unset ::teFolders($folder)
      }
   }
}
##
# Create an html page and launch the browser on it for the
# files on the clipboard
#
proc html {args} {
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return ""
   }
   clipboard clear
   set files [list]
   # Ignore clibboard entries that aren't files
   # Ignore clibboard entries are duplicates
   # Ignore errors due to funky clipboard data
   set html ""
   catch {
      foreach {file} [lflip $tryFiles] {
         if {[db::lookup $file Data 1]>=3} {
            append html "<TR><TD>$Data(Artist)</TD>$Data(Title)<TD>[timify $Data(Duration)]</TD></TR>\n"
         }
      }
   }
   clipboard append $html
   return 1
}

package provide catalog 1.0
