#------------------------------------------------------------------------------
# id3Tag 1.0
#
# ID3 Lookup functions for snackAmp Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/id3.tcl,v 1.53 2004/02/11 01:00:18 wilkason Exp $
}
##
# User API: read: id3Label
#           write: id3V1Modify
#
#------------------------------------------------------------------------------
# Function   : id3 handling routines
# Description: Create tag list for lookup in file
# Author     : Tom Wilkason
# Date       : 2/6/2001
#------------------------------------------------------------------------------
namespace eval id3Tag {
   variable ID3types
   variable ID3typesLU
   variable TagIDs
   variable id3v1Lookup
   variable id3v1ReverseLookup
   variable v1Genres
   variable mmMatch "<Edit For Multiple Tracks>"
   variable lastGenre "Not Set"
   variable TableCols
   variable TableJust
   variable TableType
   variable TableWidths
   # Includes tags up to V2.4
   set ID3types_ {
     AENC "Audio encryption"
     APIC "Attached picture"
     ASPI "Audio seek point index"
     COMM "Comments"
     COMR "Commercial frame"
     ENCR "Encryption method registration"
     EQU2 "Equalisation"
     EQUA "Equalization"
     ETCO "Event timing codes"
     GEOB "General encapsulated object"
     GRID "Group identification registration"
     IPLS "Involved people list"
     LINK "Linked information"
     MCDI "Music CD identifier"
     MLLT "MPEG location lookup table"
     MusicMatch "Unused"
     NCON "Unknown"
     OWNE "Ownership frame"
     PCNT "Play counter"
     POPM "Popularimeter"
     POSS "Position synchronisation frame"
     PRIV "Private frame"
     RBUF "Recommended buffer size"
     RVA2 "Relative volume adjustment"
     RVAD "Relative volume adjustment"
     RVRB "Reverb"
     SEEK "Seek"
     SIGN "Signature"
     SYLT "Synchronized lyric/text"
     SYTC "Synchronized tempo codes"
     TALB "Album/Movie/Show title"
     TBPM "BPM (beats per minute)"
     TCOM "Composer"
     TCON "Content type"
     TCOP "Copyright message"
     TDAT "Date"
     TDEN "Encoding"
     TDLY "Playlist delay"
     TDOR "Original release"
     TDRC "Recording"
     TDRL "Release"
     TDTG "Tagging"
     TENC "Encoded by"
     TEXT "Lyricist/Text writer"
     TFLT "File type"
     TIME "Time"
     TIPL "Involved people"
     TIT1 "Content group description"
     TIT2 "Title/songname/content description"
     TIT3 "Subtitle/Description refinement"
     TKEY "Initial key"
     TLAN "Language(s)"
     TLEN "Length"
     TMCL "Musician credits"
     TMED "Media type"
     TMOO "Mood"
     TOAL "Original album/movie/show title"
     TOFN "Original filename"
     TOLY "Original lyricist(s)/text writer(s)"
     TOPE "Original artist(s)/performer(s)"
     TORY "Original release year"
     TOWN "File owner/licensee"
     TPE1 "Lead performer(s)/Soloist(s)"
     TPE2 "Band/orchestra/accompaniment"
     TPE3 "Conductor/performer refinement"
     TPE4 "Interpreted, remixed, or otherwise modified by"
     TPOS "Part of a set"
     TPRO "Produced notice"
     TPUB "Publisher"
     TRCK "Track number/Position in set"
     TRDA "Recording dates"
     TRSN "Internet radio station name"
     TRSO "Internet radio station owner"
     TSIZ "Size"
     TSOA "Album sort order"
     TSOP "Performer sort order"
     TSOT "Title sort order"
     TSRC "ISRC (international standard recording code)"
     TSSE "Software/Hardware and settings used for encoding"
     TSST "Set subtitle"
     TXXX "User defined text information frame"
     TYER "Year"
     UFID "Unique file identifier"
     USER "Terms of use"
     USLT "Unsychronized lyric/text transcription"
     WCOM "Commercial information"
     WCOP "Copyright/Legal information"
     WOAF "Official audio file webpage"
     WOAR "Official artist/performer webpage"
     WOAS "Official audio source webpage"
     WORS "Official internet radio station homepage"
     WPAY "Payment"
     WPUB "Publishers official webpage"
     WXXX "User defined URL"
   }
   ;# Make a real list
   foreach {type description} $ID3types_ {
      lappend ID3types $type [string trim $description]
      set ID3typesLU($type) $description

   }
   array set id3v1Lookup {
      0  "Blues"        25 "Euro-Techno"       50 "Darkwave"          75 "Polka"            100 "Humour"        125 "Dance Hall"
      1  "Classic Rock" 26 "Ambient"           51 "Techno-Industrial" 76 "Retro"            101 "Speech"        126 "Goa"
      2  "Country"      27 "Trip Hop"          52 "Electronic"        77 "Musical"          102 "Chanson"       127 "Drum & Bass"
      3  "Dance"        28 "Vocal"             53 "Pop-Folk"          78 "Rock & Roll"      103 "Opera"         128 "Club House"
      4  "Disco"        29 "Jazz+Funk"         54 "Eurodance"         79 "Hard Rock"        104 "Chamber Music" 129 "Hardcore"
      5  "Funk"         30 "Fusion"            55 "Dream"             80 "Folk"             105 "Sonata"        130 "Terror"
      6  "Grunge"       31 "Trance"            56 "Southern Rock"     81 "Folk/Rock"        106 "Symphony"      131 "Indie"
      7  "Hip-Hop"      32 "Classical"         57 "Comedy"            82 "National Folk"    107 "Booty Bass"    132 "BritPop"
      8  "Jazz"         33 "Instrumental"      58 "Cult"              83 "Swing"            108 "Primus"        133 "Negerpunk"
      9  "Metal"        34 "Acid"              59 "Gangsta Rap"       84 "Fast-Fusion"      109 "Porn Groove"   134 "Polsk Punk"
      10 "New Age"      35 "House"             60 "Top 40"            85 "Bebop"            110 "Satire"        135 "Beat"
      11 "Oldies"       36 "Game"              61 "Christian Rap"     86 "Latin"            111 "Slow Jam"      136 "Christian Gangster"
      12 "Other"        37 "Sound Clip"        62 "Pop/Punk"          87 "Revival"          112 "Club"          137 "Heavy Metal"
      13 "Pop"          38 "Gospel"            63 "Jungle"            88 "Celtic"           113 "Tango"         138 "Black Metal"
      14 "R&B"          39 "Noise"             64 "Native American"   89 "Blue Grass"       114 "Samba"         139 "Crossover"
      15 "Rap"          40 "Alt. Rock"         65 "Cabaret"           90 "Avantegarde"      115 "Folklore"      140 "Contemporary Club"
      16 "Reggae"       41 "Bass"              66 "New Wave"          91 "Gothic Rock"      116 "Ballard"       141 "Christian Rock"
      17 "Rock"         42 "Soul"              67 "Psychedelic"       92 "Progressive Rock" 117 "Power Ballard" 142 "Merengue"
      18 "Techno"       43 "Punk"              68 "Rave"              93 "Psychedelic Rock" 118 "Rhythmic Soul" 143 "Salsa"
      19 "Industrial"   44 "Space"             69 "Showtunes"         94 "Symphonic Rock"   119 "Freestyle"     144 "Thrash Metal"
      20 "Alternative"  45 "Meditative"        70 "Trailer"           95 "Slow Rock"        120 "Duet"          145 "Anime"
      21 "Ska"          46 "Instrumental Pop"  71 "Lo-Fi"             96 "Big Band"         121 "Punk Rock"     146 "JPop"
      22 "Death Metal"  47 "Instrumental Rock" 72 "Tribal"            97 "Chorus"           122 "Drum Solo"     147 "Synthpop"
      23 "Pranks"       48 "Ethnic"            73 "Acid Punk"         98 "Easy Listening"   123 "A Cappella"    -1  "Unknown"
      24 "Soundtrack"   49 "Gothic"            74 "Acid Jazz"         99 "Acoustic"         124 "Euro-House"    255 "Unknown"
   }
   # Order shown on editor
   set v1Parts {
      "Title"
      "Artist"
      "Album"
      "Year"
      "Note"
      "Track"
      "SubGenre"
      "Genre"
   }
   foreach {index value} [array get id3v1Lookup] {
      lappend v1Genres $value
      set id3v1ReverseLookup($value) $index
   }
   set v1Genres [lsort $v1Genres]
   set TagIDs   [list Artist Title Album Track Media Year Genre Note Comp Enc Desc]
   variable oggSound [snack::sound -debug 0]
}
#------------------------------------------------------------------------------
# Return the offset to the start of music so the ID3V2 tag can
# be skipped by a streaming server
#------------------------------------------------------------------------------
proc id3Tag::id3V2Offset {fid} {
   # Have ID, get data
   fconfigure $fid -translation binary
   if {[catch {set block [read $fid 10]}] } then {
      return -code error $result
   }
   if {[catch {seek $fid 0 start} result] } then {
      return 0
   }
   ##
   # If this is an ID3 V2, need offset to start of stream
   #
   if {[string range $block 0 2] == "ID3"} {
      # Determine the frame length and read the rest of the id3 header
      binary scan [string range $block 6 9] "c1c1c1c1" a b c d
      set length [expr {($a<<21 | $b<<14 | $c<<7 | $d)  & 0xFFFFF }]
      if {$length < 0} {
         return 0
      } else {
         return $length
      }
   }
   return 0
}
#------------------------------------------------------------------------------
# Function   : ideTag::id3V2Modify
# Description: Write out the ID3V2 tag, requires reqwriting the entire file
# Author     : Tom Wilkason
# Date       : 1/30/2004
# Notes: Will skip embedded binary data
#        Won't handle compression or header extensions properly
#        Should write out older version of tag
#        todo
#        If new tag will fit within existing tag space then
#        we should overwrite existing tag and pad data.
#------------------------------------------------------------------------------
proc id3Tag::id3V2Modify {file _Data} {
   # get ID3V2 tag names from index names
   # Compute the total size of the old tag
   # open input file read
   # get V2 offset from original file
   # Compute size of new tag
   # if tag len <= existing tag len
      # open input for writing
   #else
      # will process new file, delete old
   #
   # open temp file write
   # build the header and write to temp file
   # write out the tags
   # Verify seek position with tag length (as a check)
   # seek to end of tag in input file
   # read entire inputfile into memory as binary
   # write out buffer to output file
   # close both files
}
#------------------------------------------------------------------------------
# Function   : ideTag::id3V2Get
# Description: Return a V2 tag from using a file name, a list of all known fields
# Author     : Tom Wilkason
# Date       : 3/17/2002
#------------------------------------------------------------------------------
proc id3Tag::id3V2Get {file} {
   Update
   variable ID3types
   variable id3v1Lookup
   variable ID3typesLU
   set result [list]
   if {[catch {set fid [open $file r]}]} {
      return $result
   }
   # Have ID, get data
   fconfigure $fid -translation binary
   if {[catch {set block [read $fid 10]}] } then {
      close $fid
      return $result
   }
set info {
The first part of the ID3v2 tag is the 10 byte tag header, laid out
   as follows:

     0-2 ID3v2/file identifier      "ID3"
     3-4 ID3v2 version              $04 00  (V2.4.0)
     5   ID3v2 flags                %abcd0000
     6-9 ID3v2 size             4 * %0xxxxxxx

   The first three bytes of the tag are always "ID3", to indicate that
   this is an ID3v2 tag, directly followed by the two version bytes. The
   first byte of ID3v2 version is its major version, while the second
   byte is its revision number. In this case this is ID3v2.4.0. All
   revisions are backwards compatible while major versions are not. If
   software with ID3v2.4.0 and below support should encounter version
   five or higher it should simply ignore the whole tag. Version or
   revision will never be $FF.
}
   if {[string range $block 0 2] == "ID3"} {
      set Pairs [list]
      # Determine the frame length and read the rest of the id3 header
      binary scan [string range $block 3 9] "h1h1h1c1c1c1c1" majVer minVer flag a b c d
      # 7 bits of each byte are used for a total of 28 bits (256Mb)
      # clamp at 1M (2^20) (for safety)
      set length [expr {($a<<21 | $b<<14 | $c<<7 | $d) & 0xFFFFF}]
      # Note: May want to clamp length because we don't care about embedded
      # cover art and lyrics.
      if {$length <= 0} {
         close $fid
         return [list]
      }
      set block [read $fid $length]
      # Scan for tags and save their location
      foreach {type description} $ID3types {
         # Note: This can be a bit slow if the tag space is large
         # May want to bail if we have the basic tags
         # Note: Since there is no tag order in the ID3V2
         #       we have to search all of it for each tag
         #       (there may be multiple instances of the same tag)
         set offset 0
         while {1} {
            set loc [string first $type $block $offset]
            if {$loc >= 0} {
               lappend Pairs [list $type $loc]
               set offset [expr {$loc+1}]
            } else {
               break
            }
         } ;# end while
      }
      # Sort by first to last and build up a list of the start
      # and end position of each tag then extract it and add
      # its description and the tag data to a list
      set newPairs [lsort -index 1 -integer $Pairs]
      set result [list]
      set numPairs [llength $newPairs]
      for {set i 0} {$i < $numPairs} {incr i} {
         set loc1 [lindex [lindex $newPairs $i] 1]
         set loc2 [lindex [lindex $newPairs [expr {$i+1}]] 1]
         ;# Clamp the amount of data we will use
         if {$loc2==""} {
            set loc2 [expr {$loc1+60}]       ;#last tag
         } elseif {$loc2>$loc1+512} {
            set loc2 [expr {$loc1+512}]
         }
         set nam1 [lindex [lindex $newPairs $i] 0]
         # Extract the tag with any codes
         # Note: If for some reason the tag (e.g. COMM) shows up more than once
         # the last one will be used.
         set String [string range $block [expr {$loc1 + 4+7}] [expr {$loc2 - 1}]]
         # Clean up result
         switch -- $nam1 {
            "NCON" {}
            "APIC" {}
            "PRIV" {}
            "TPOS" {}
            "COMM" {
               ;# First three bytes are the language
               set String [id3Clean [encoding convertfrom [string range $String 3 end]]]
               lappend result $nam1 $ID3typesLU($nam1) $String
            }
            "TCON" {
               ;# (num) indicates a V1 tag reference, so look it up
               set String [id3Clean $String]
               if {[regexp {\((.+)\)(.*)} $String -> v1 refine]} {
                  set String $id3v1Lookup($v1)
                  if {[string length $refine]} {
                     append String "-$refine"
                  }
               }
               lappend result $nam1 $ID3typesLU($nam1) $String
            }
            default {
               set String [id3Clean [encoding convertfrom $String]]
               lappend result $nam1 $ID3typesLU($nam1) $String
            }
         }
      }
      lappend result "Tag" "ID3 Tag Type" V2.$majVer.$minVer
   }
   close $fid
   return $result
}
#------------------------------------------------------------------------------
# Retrieve ogg data for a file
#------------------------------------------------------------------------------
proc id3Tag::id3oggGet {file} {
   variable oggSound
   variable id3v1Lookup
   $oggSound config -file $file
   if {[catch {$oggSound config -comment} data] } then {
      return [list]
   } else {
      foreach {entry} $data {
         foreach {tag value} [split $entry =] {
            set value [id3Clean $value]
            switch -- [string tolower $tag] {
               "title"       {lappend result "TIT2" "Title"   $value}
               "artist"      {lappend result "TPE1" "Artist"  $value}
               "album"       {lappend result "TALB" "Album"   $value}
               "genre"       {lappend result "TCON" "Content Type" $value}
               "date"        {lappend result "TYER" "Year"    $value}
               "tracknumber" {lappend result "TRCK" "Track"   $value}
               "comment"     {lappend result "COMM" "Comment" $value}
               default {}
            }
         }
      }
      lappend result "Tag" "ID3 Tag Type" OGG
   }
   $oggSound config -file ""
   return $result
}
#------------------------------------------------------------------------------
# Function   : ideTag::id3V1Get
# Description: Return a V1 tag from using a file name, a list of known fields
# Author     : Tom Wilkason
# Date       : 3/17/2002
#------------------------------------------------------------------------------
proc id3Tag::id3V1Get {file} {
   Update
   variable id3v1Lookup
   set result [list]
   if {[catch {set fid [open $file r]} ec]} {
      saLog "$ec ($file)"
      return [list]
   }
   fconfigure $fid -translation binary
   ##
   # ID3V1.2 tags are in the last 256 bytes of the file in a fixed format
   #
   if {[catch {seek $fid -256 end} ec] } {
      close $fid
      saLog "$ec ($file)"
      return [list]
   }
   if {[catch {set block [read $fid 256]} ec] } then {
      close $fid
      saLog "$ec ($file)"
      return [list]
   }
   close $fid
   binary scan $block "a3 a30 a30 a30 a15 a20 a3 a30 a30 a30 a4 a28 ccc" \
      ext extTitle extArtist extAlbum extComment extGenre \
      id title artist album  year comment zero track genre
   # Support ID3V1.2 Extensions
   if {[string equal $ext "EXT"]} {
      append title   $extTitle
      append artist  $extArtist
      append album   $extAlbum
      append comment $extComment
   }
   if {[string equal $id "TAG"]} {
      # V1.1 spec allows last comment string to be the track, if not then a comment
      # If a null char before a non-null Track, the use the Track otherwise append
      # to the Comment
      set ver "V1"
      if {$zero==0 && $track!=0} {
         if {[string is integer -strict $track]} {
            set ver "V1.1"
            set track [toUnsigned $track]
         } else {
            set track ""
         }
      } else {
         # eke out the last two chars and append to comment, they were not a track
         append comment [string trim [binary format c $zero] [binary format c $track]]
         set track ""
      }
      lappend result "TRCK" "Track"     $track
      lappend result "TIT2" "Title"     [id3Clean $title]
      lappend result "TPE1" "Artist"    [id3Clean $artist]
      lappend result "TALB" "Album"     [id3Clean $album]
      lappend result "TYER" "Year"      [id3Clean $year]
      lappend result "COMM" "Comment"   [id3Clean $comment]
      # Sub Genre and tag type
      if {[string equal $ext "EXT"]} {
         lappend result "TSST" "SubGenre" [id3Clean $extGenre]
         lappend result "Tag" "ID3 Tag Type" "V1.2"
      } else {
         lappend result "Tag" "ID3 Tag Type" "$ver"
      }
      if {![catch {set Genre $id3v1Lookup($genre)}]} {
         lappend result "TCON" "Content Type" $Genre
      } else {
         lappend result "TCON" "Content Type" $id3v1Lookup(12)
      }
   }
   return $result
}
#------------------------------------------------------------------------------
# Function   : id3Tag::cleanTrack
# Description: Format a track with atleast two digits
# Author     : Tom Wilkason
# Date       : 10/12/2002
#------------------------------------------------------------------------------
proc id3Tag::cleanTrack {track} {
   if {[catch {format "%2.2d" [string trimleft $track 0]} Track] } then {
      return $track
   } else {
      return $Track
   }
}

#------------------------------------------------------------------------------
# Function: id3Tag::isogg
#
# Return true if the file extension is for an ogg/vorbis file
#------------------------------------------------------------------------------
proc id3Tag::isogg {file} {
   return [string equal -nocase [file extension $file] ".ogg"]
}

#------------------------------------------------------------------------------
# Function: id3Tag::id3oggModify
#
# Update the Ogg tag with new data (if needed)
#------------------------------------------------------------------------------
proc id3Tag::id3oggModify {file _Data} {
   variable oggSound
   $oggSound config -file $file
   set oldTag [lsort [$oggSound config -comment]]
   upvar $_Data Data
   set newTag [list]
   foreach {idTag oggTag} {
      Title  TITLE
      Artist ARTIST
      Album  ALBUM
      Track  TRACKNUMBER
      Genre  GENRE
      Note   COMMENT
      Year   DATE
   } {
      if {[info exists Data($idTag)]} {
         lappend newTag "$oggTag=$Data($idTag)"
      }
   }
   set newTag [lsort $newTag]
   if {![string equal $oldTag $newTag]} {
      saLog "Changing\n$oldTag\nto\n$newTag"
      $oggSound config -comment $newTag
   }
   $oggSound config -file ""
   return 1
}
#------------------------------------------------------------------------------
# Function   : id3V1Modify
# Description: Modify the ID3V1 tag of a file
# Author     : Tom Wilkason
# Date       : 2/7/2001
#------------------------------------------------------------------------------
proc id3Tag::id3V1Modify {file _Data} {
   Update
   upvar $_Data Data
   variable v1Parts
   variable id3v1Lookup
   variable id3v1ReverseLookup
   variable mmMatch
   variable lastGenre
   ##
   # Read in current ID3 data then compare to determine if any changes were made
   #
   set tagExists [id3Label $file id3V1Data V1]
   if {!$tagExists} {
      if {[isogg $file]} {
         if {[id3Label $file id3altData ogg]} {
            array set id3V1Data [array get id3altData]
         }
      } else {
         if {[id3Label $file id3altData V2]} {
            array set id3V1Data [array get id3altData]
         }
      }
      set offset 0
   } else {
      # V1 or V1.2 Data
      if {$id3V1Data(Tag)=="V1.2" && $::snackAmpSettings(ID3V12)} {
         set offset -256
      } else {
         set offset -128
      }
   }
   set diff 0
   # Check each field
   foreach {element} $v1Parts {
      ##
      # If tags don't match or incoming tag has data and existing tag doesn't
      #
      if {[hasData Data $element]} {
         if {[hasData id3V1Data $element] && ![string equal $id3V1Data($element) $Data($element)]} {
            # Both have data, but different
            set diff 1
         } elseif {![hasData id3V1Data $element]} {
            # New incoming field
            set diff 1
         }
         # For multi-match tags, either replace or discard them
         # Ignore the < Match > fields, use existing data
         # Note: TFW If id3V1Data($element) doesn't exist (new tag) then what should we do? Use blanks?
         if {$Data($element) == $mmMatch} {
            if {[info exists id3V1Data($element)]} {
               # Use existing data (not the multi-patch string)
               set Data($element) $id3V1Data($element)
            } else {
               # Don't write that portion of tag
               set Data($element) ""
            }
         }
      } elseif {[info exists id3V1Data($element)]} {
         # Incoming blank, copy over existing if it exists
         set Data($element) $id3V1Data($element)
      } else {
         # Otherwise just blank it out
         set Data($element) ""
      }
   }
   ##
   if {$diff == 0} {
      return {}
   }
   ##
   # We know we have changes, ready for write
   #
   if {![file writable $file] || $::snackAmpSettings(ID3ReadOnly)} {
      tk_messageBox -type ok -icon warning -message "You don't have write permission for $file"
      return {}
   }
   set mtime [file mtime $file]
   if {[catch {set fid [open $file r+]}]} {
      tk_messageBox -type ok -icon warning -message "Could not open file $file for writing"
      return {}
   }
   fconfigure $fid -translation binary
   if {[catch {seek $fid $offset end} result] } {
      close $fid
      return {}
   }
   ##
   # Handle Genre, default to 12 (Other)
   set genre $Data(Genre)
   if {[info exists id3v1ReverseLookup($genre)]} {
      set Genre $id3v1ReverseLookup($genre)
      set lastGenre $genre
   } else {
      set Genre 12
   }

   ##
   # Handle Track, use zero (blank) if not valid, don't interpret
   # octal numbers as octal
   set Track [string trimleft $Data(Track) 0]
   if {[string is integer -strict $Track]} {
      set T c
   } else {
      set Track "\0"
      set T a
   }
   ##
   # Write either V1 or V1.2 Tags
   if {$::snackAmpSettings(ID3V12)} {
      # Support ID3V1.2 Extensions
      foreach {item start} {
         Title 30 Artist 30 Album 30 Note 28 SubGenre 0
      } {
         set ext($item) [string range $Data($item) $start end]
      }
      set fstring "a3 a30 a30 a30 a15 a20 a3 a30 a30 a30 a4 a28 c${T} c"
      set block [binary format $fstring \
         EXT \
         [encoding convertto $ext(Title)] \
         [encoding convertto $ext(Artist)] \
         [encoding convertto $ext(Album)] \
         $ext(Note) \
         $ext(SubGenre) \
         TAG \
         [encoding convertto $Data(Title)] \
         [encoding convertto $Data(Artist)] \
         [encoding convertto $Data(Album)] \
         $Data(Year) \
         $Data(Note) \
         0 $Track $Genre]
   } else {
      set fstring "a3 a30 a30 a30 a4 a28 c${T} c"
      set block [binary format $fstring \
         TAG \
         [encoding convertto $Data(Title)]  \
         [encoding convertto $Data(Artist)] \
         [encoding convertto $Data(Album)]  \
         $Data(Year)   \
         $Data(Note)   \
         0 $Track $Genre]
   }
   puts -nonewline $fid $block
   close $fid
   ##
   # Make sure we don't have fields we don't intend to have
   #
   foreach {field} [list size sampRate FullData Played Gain Rating Votes date Mounted LastPlay Votes Start Stop Rate Duration] {
      if {[info exists Data($field)]} {
         debug "Found unexpected data"
         parray Data
         break
      }
   }
   ##
   # Update the database with this info
   #
   db::cacheFile $file Data 0

   ##
   # Restore the old timestamp if needed
   if {$::snackAmpSettings(preserveTime)} {
      if {[catch {file mtime $file $mtime} result] } then {
         saLog "Could not reset file time attribute for $file"
      }
   }
   return 1
}
#------------------------------------------------------------------------------
# Function   : id3Tag::removeTag
# Description: Remove a tag from a file it one exists
#              Does not work, tcl won't truncate
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Tag::removeTag {files} {
   foreach {file} $files {
      if {[catch {set fid [open $file r+]}]} {
         tk_messageBox -type ok -icon warning -message "Could not open file $file for writing"
         return 0
      }
      fconfigure $fid -translation binary
      ##
      # Now check for ID3V1 tags, these are much easier
      # Return codes for ID3V2 so only one tag type is dealt with
      # ID3V1 tags are in the last 128 bytes of the file in a fixed format
      if {[catch {seek $fid -128 end} result] } {
         close $fid
         return 0
      }
      # TFW: Here is where we know if a tag exists
      set block [read $fid 128]
      if {[string range $block 0 2] != "TAG"} {
         # No Tag
         seek $fid 0 end
      } else {
         # Has a tag, close will truncate here
         seek $fid -128 end
         saLog "Removing tag for $file"
      }
      # Note: This function does not really truncate the file.
      close $fid
   }
   return 1
}
#------------------------------------------------------------------------------
# Function   : id3Tag::padTo
# Description: Pad/truncate a string to some length
# Author     : Tom Wilkason
# Date       : 3/17/2002
#------------------------------------------------------------------------------
proc id3Tag::padTo {string len} {
   return [string range [format "%-${len}s" $string] 0 [incr len -1]]
}
#------------------------------------------------------------------------------
# Function   : formatID3Data {List}
# Description: If a list of ID3 info is passed in, this will format it for
#              use on a balloon pop-up. Non-blank entries are not returned.
# Author     : Tom Wilkason
# Date       : 11/11/2001
#------------------------------------------------------------------------------
proc id3Tag::formatID3Data {_Data} {
   variable TagIDs
   upvar $_Data Data
   set Info {}
   # for each tag, if it exists then append the info
   foreach Tag $TagIDs {
      if {[hasData Data $Tag]} {
         append Info "$Tag \t: $Data($Tag)\n"
      }
   }
   return $Info
}
#------------------------------------------------------------------------------
# Function   : id3Clean
# Description: Remove garbage from an ID3 tag
# Author     : Tom Wilkason
# Date       : 2/11/2001
#------------------------------------------------------------------------------
proc id3Tag::id3Clean {String {trim 1}} {
   regsub -all  -- {\0|||[[:cntrl:]]} $String {} String
   if {$trim} {
   # trimming causes problems with join 1.2 tags, need to retain the spaces
      return [string trim $String]
   } else {
      return $String
   }
}
#------------------------------------------------------------------------------
# Function   : id3Label
# Description: Return an appropriate ID3 Label in an array
#              Additional tags are appended to the Desc field (V2 has soo many...)
# Author     : Tom Wilkason
# Date       : 2/7/2001
#------------------------------------------------------------------------------
proc id3Tag::id3Label {file _Array {types "V1 V2 ogg"}} {
   upvar $_Array Array
   # Get data V1 then V2
   foreach type $types {
      set data [id3${type}Get $file]
      if {[llength $data] > 0} {
         foreach {TagID Description Value} $data {
            if {[string length $Value] > 0} {
               #set Value [string map {\{ ( \} ) \n " "} $Value]
               set Value [id3Clean $Value]
               switch -- $TagID {
                  TPE1 {set Array(Artist)   [encoding convertfrom $Value]}
                  TIT2 {set Array(Title)    [encoding convertfrom $Value]}
                  TALB {set Array(Album)    [encoding convertfrom $Value]}
                  TRCK {set Array(Track)    $Value}
                  TMED {set Array(Media)    $Value}
                  TYER {set Array(Year)     $Value}
                  TCON {set Array(Genre)    $Value}
                  COMM {append Array(Note) [encoding convertfrom $Value]}
                  TIT1 {append Array(Note) [encoding convertfrom $Value]}
                  TXXX {append Array(Note) [encoding convertfrom $Value]}
                  TCOM {set Array(Comp)     $Value}
                  TENC {set Array(Enc)      $Value}
                  TSST {set Array(SubGenre) $Value}
                  Tag  {set Array(Tag)      $Value}
                  TLEN {}
                  PCNT {}
                  default {
                     append Array(Desc) "$Description: $Value"
                  }
               }
            }
         }
         #set Array(Tag) $type
         return 1
      } ; #end if data
   }
   return 0
}
#------------------------------------------------------------------------------
# Function   : id3Tag::blankForNull
# Description: Return a blank or real data for some array element
#
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Tag::blankForNull {_Data element} {
   upvar $_Data Data
   if {[info exists Data($element)]} {
      return $Data($element)
   } else {
      return ""
   }
}
#------------------------------------------------------------------------------
# Function   : id3Tag::appendIfNotBlank
# Description:
# Author     : Tom Wilkason
# Date       : 11/11/2001
#------------------------------------------------------------------------------
proc id3Tag::appendIfNotBlank {_Array Title Data} {
   upvar $_Array Array
   if {[string length $Data]>0} {
      append Array "$Title $Data\n"
   }
}
#------------------------------------------------------------------------------
# Function: id3Editor::guessArtist/guessTitle/guessAlbum
#
# Make an estimate of the artist/title & album names, used if no ID3 tag info exists
#------------------------------------------------------------------------------
proc id3Tag::guessArtist {file} {
    return [lindex [lrange [file split [file rootname $file]] end-$::snackAmpSettings(ListDepth) end] 0]
}
proc id3Tag::guessTitle {file} {
   return [file rootname [file tail $file]]
}
proc id3Tag::guessAlbum {file} {
   return [lindex [lrange [file split [file rootname $file]] end-$::snackAmpSettings(ListDepth) end] end-1]
}


package provide snID3 1.0



