#----------------------------------------------------------------------------
# Establish a client interface for an IRC remote control server
# This currently works with the ATI remote
# Requires WinLIRC ATI Remote plug-in http://winlirc.sourceforge.net/
# If other LIRC remotes are used, may require mods to switch statement below
#----------------------------------------------------------------------------
##
# IR Handler for remote controls
#
namespace eval irc {
   variable socket 0
   variable enabled
   variable ircCheck
   variable ircEnable
   variable port
   variable host
   upvar #0 snackAmpSettings(ircEnabled) enabled
   upvar #0 snackAmpSettings(ircCheck)   ircCheck
   upvar #0 snackAmpSettings(ircEnable)  ircEnable
   upvar #0 snackAmpSettings(ircPort)    port
   upvar #0 snackAmpSettings(ircHost)    host
}
#----------------------------------------------------------------------------
# Install the LIRC socket interface
#----------------------------------------------------------------------------
proc irc::Server {args} {
   variable socket
   variable port
   variable host
   if {[string length [string trim $host]]==0} {
      set host [info hostname]
   }
   catch {close $socket} result
   if [catch {set socket [socket $host $port]} result] {
      saLog "Error Establishing Server $::errorInfo"
   } else {
      fconfigure $socket -translation auto
      fileevent $socket readable [list irc::Process $socket]
      saLog "IRC Server Established on port $port"
   }
}
proc irc::Close {args} {
   variable socket
   catch {close $socket} result
}
#########################################################################
# Function    : irc::Process
# Description : Called from fileevent on a readable input on my port
# Author      : Tom Wilkason
# Date        : 6/2/1999
#########################################################################
proc irc::Process {sock} {
   global snackAmpSettings
   variable socket
   variable enabled
   variable ircCheck
   variable ircEnable
   ;# If eof or partial line the close up the channel now
   if {[eof $sock] || [catch {gets $sock line}]} {
      # end of file or abnormal connection drop
      close $sock
      set socket 0
      # Re-establish server
      after 200 irc::Server
   } else {
      # Process the received command line
      # The forth param is the remote type e.g. "atiremotewonder"
      set remote [lindex $line 3]
      # The third parameter is the ascii command
      set command [lindex $line 2]
      set ::snackAmpSettings(ircLastKey) $command
      ##
      # Check for an enable key, otherwise disable
      #
      if {[lsearch -exact $ircCheck $command]>=0} {
         if {[string equal $command $ircEnable]} {
            set enabled 1
         } else {
            set enabled 0
         }
      }
      if {$enabled} {
         switch -- [string tolower $command] {
            vol_up {bumpLevel 2}
            vol_down {bumpLevel -2}
            fastforward {Next}
            rewind {Prev}
            1 -
            2 -
            3 -
            4 -
            5 {
               db::setVal $::TA(currentMF) Rating [string repeat * $command]
            }
            pause {Pause}
            stop {Stop}
            play {Play}
            channel_up {bumpScaling .1}
            channel_down {bumpScaling -.1}
            mute {Pause}
            right {macros::skipPlay 10}
            left {macros::skipPlay -10}
            record {taExit}
            up {wm deiconify .saTop}
            down {wm iconify .saTop}
            ok {::browser::urlOpen $::snackAmpSettings(internalWebBase)/remote/index.html}
            default {
            }
         }
      }
   }
}
#----------------------------------------------------------------------------
# Fire up the server
#----------------------------------------------------------------------------
irc::Server
package provide ircserver 1.0
