#------------------------------------------------------------------------------
# macros 1.0
#
# High level commands for SnackAmp
#
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/macros.tcl,v 1.6 2004/02/11 00:58:34 wilkason Exp $
}
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# Macros Namespace
#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
namespace eval macros {
   namespace export *
}
#------------------------------------------------------------------------------
# Function:  macros::PlayOffset
#
# Play a music file at some offset, likely bound to slider control
#------------------------------------------------------------------------------
proc macros::PlayOffset {{manual 0}} {
   Trace
   global TA LB snackAmpSettings
   # Make sure we have an open file
   set soundLength [soundControl::length]
   if {$soundLength > 0} {
      # Some files are fast and some are slow to respond to -start = 0
      if {$manual == 0} {
         set offset [expr {int(double($gui::trackPosition)/$gui::songDuration*$soundLength)}]
      } else {
         set offset $manual
      }
      if {[catch {
         soundControl::stop
         ;# Inverse active track
         LB::inverseVideoFile $TA(currentMF) 1
         # Workaround for bug on linux/ogg where file plays from last pos, not from -start
         if {$::tcl_platform(platform) == "unix"} {
            soundControl::sound [soundControl::fileName]
         }
         soundControl::play -start $offset -command [soundControl::getNextCommand $::snackAmpSettings(trackGap)]
      } result]} then {
         saLog "PlayOffset: $result"
         return
      }
      if {![afterExists soundControl::Timer]} {
         soundControl::Timer
      }
   }
}
#------------------------------------------------------------------------------
# Function   : bumpScaling
# Description: Adjust the current scaling for the playing song for some amount
# Author     : Tom Wilkason
# Date       : 1/9/2002
#------------------------------------------------------------------------------
proc macros::bumpScaling {percent} {
   global TA
   set gain [snack::audio scaling]
   gainControl::add $TA(currentMF) [expr {$gain*(1.0+$percent)}]
   snack::audio scaling [expr {$gain*(1.0+$percent)}]
}
#------------------------------------------------------------------------------
# Function   : Assign
# Description: Routine to assign global hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Assign {key Command} {
   if {[catch {bind .saTop $key $Command} result]} {
      saLog "FilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : FilterAssign
# Description: Routine to assign filter box hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::FilterAssign {key Command} {
   if {[catch {bind $::FI $key $Command} result]} {
      saLog "FilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : PlaylistAssign
# Description: Routine to assign play list hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::PlaylistAssign {key Command} {
   if {[catch {bind [LB::body] $key $Command} result]} {
      saLog "PlaylistAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : APL_ListAssign
# Description: Routine to assign APL play list hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APL_ListAssign {key Command} {
   if {[APLExists]} {
      set plb [$::PLB bodypath]
      if {[catch {bind $plb $key $Command} result]} {
         saLog "APL_ListAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL_TreeAssign
# Description: Routine to assign APL tree hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APL_TreeAssign {key Command} {
   if {[APLExists]} {
      if {[catch {bind [winfo parent $::TREE] $key $Command} result]} {
         saLog "APL_TreeAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL_TreeAssign
# Description: Routine to assign APL tree hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APL_TreeNodeAssign {key Command} {
   if {[APLExists]} {
      if {[catch {$::TREE bind x $key $Command} result]} {
         saLog "APL_TreeNodeAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APLFilterAssign
# Description: Routine to assign APL filter box hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APLFilterAssign {key Command} {
   if {[APLExists]} {
      if {[catch {bind $::SAFind $key $Command} result]} {
         saLog "APLFilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}

#------------------------------------------------------------------------------
# Function   : editPLID3
# Description: Edit the highligted files ID3 info
# Author     : Tom Wilkason
# Date       : 3/18/2002
#------------------------------------------------------------------------------
proc macros::SelectFile {{playing 0}} {
   if {!([winfo exists .wFC] && [wm state .wFC] == "normal")} {
      AutoPlaylist
   }
   if {$playing} {
      APL::SelectFile [soundControl::fileName]
   } else {
      APL::SelectFile [LB::nameActive]
   }
}

#------------------------------------------------------------------------------
# Function: macros::EditPLID3
#
# Invoke the id3 editor (in group mode) on the selected playlist file
#------------------------------------------------------------------------------
proc macros::EditPLID3 {args} {
   package require id3Editor
   set files [LB::nameSelected 1]
   if {[llength $files] > $::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Editor::id3Editor [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
}

#------------------------------------------------------------------------------
# Function: macros::PLNameToTag
#
# Invoke the name to tag editor on the playlist selected files
#------------------------------------------------------------------------------
proc macros::PLNameToTag {smart} {
   package require id3Editor
   id3Editor::id3NameToTag [LB::nameSelected 1] $smart
}

#------------------------------------------------------------------------------
# Function: macros::PLTagToName
#
# Invoke the tag to name editor on the playlist selected files
#------------------------------------------------------------------------------
proc macros::PLTagToName {args} {
   package require id3Editor
   id3Editor::id3TagToName [LB::nameSelected 1]
}

#------------------------------------------------------------------------------
# Function: macros::removeTag
#
# Remove an ID3 tag (not currently implemented)
#------------------------------------------------------------------------------
proc macros::removeTag {args} {
   id3Tag::removeTag [LB::nameSelected 1]
}

#------------------------------------------------------------------------------
# Function: macros::EditPlaying
#
# Edit the tag info on the currently playing track
#------------------------------------------------------------------------------
proc macros::EditPlaying {args} {
   package require id3Editor
   id3Editor::id3Editor [list $::TA(currentMF)]
}
#------------------------------------------------------------------------------
# Function   : macros::Highlight
# Description: macros::Highlight one and only one selection
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc macros::Highlight {i {inverse 0}} {
   global TA
   LB::highlightIndex $i $inverse
   if {[APLExists] && $::snackAmpSettings(syncAPL)} {
      set TA(entryName) [LB::nameFromIdx $i 0]
      APL::SelectFile $TA(entryName)
   }
}
#------------------------------------------------------------------------------
# Function   : Swap
# Description: Move a selection up/down the list (bind keys to this)
# Author     : Tom Wilkason
# Date       : 10/29/2001
#------------------------------------------------------------------------------
proc macros::Swap {how} {
   global LB
   if {[llength [LB::idxSelected]] > 0} {
      set selected [lsort -integer [LB::idxSelected]]
      set first [lindex $selected 0]
      set last [lindex $selected end]
      if {$how=="up"} {
         ##
         # Move the item above the selection below the selection
         #
         set moveIndex [expr {$first-1}]
         if {$moveIndex >= 0} {
            LB::moveByIndex $moveIndex $last
         }
      } else {
         ##
         # Move the item below the selection above the selection
         #
         set moveIndex [expr {$last+1}]
         if {$moveIndex < [LB::idxEnd]} {
            LB::moveByIndex $moveIndex $first
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : PlayGain/getPlayGain
# Description: Provide x2 gain control to better match audio experience
# Author     : Tom Wilkason
# Date       : 10/19/2001
#------------------------------------------------------------------------------
proc macros::PlayGain {value} {
   set newValue [expr {round(pow($value,2)/40.0)}]
   set ::TA(line,$::snackAmpSettings(lineSelect)) $newValue
}

#------------------------------------------------------------------------------
# Function   : macros::getPlayGain
# Description: Return the current gain setting log scaled
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::getPlayGain {args} {
   set audioGain $::TA(line,$::snackAmpSettings(lineSelect))
   return [expr {pow($audioGain*40.,0.5)}]
}
#------------------------------------------------------------------------------
# Function   : bumpLevel
# Description: Adjust the current play level by some amount
# Author     : Tom Wilkason
# Date       : 11/9/2001
#------------------------------------------------------------------------------
proc macros::bumpLevel {factor} {
   set ::TA(gain) [expr {$::TA(gain)+$factor}]
   PlayGain $::TA(gain)
}
#------------------------------------------------------------------------------
# Function   : Console
# Description: Toggle the console for debugging purposes (WIN32 only)
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc macros::Console {args} {
   global TA LB
   if {$TA(Console)==0} {
      console hide
   } else {
      console show
   }
}
#------------------------------------------------------------------------------
# Function   : applyCurrentAPL
# Description: Apply the latest Auto Playlist settings, ignoring any filtering
#              in effect.
# Author     : Tom Wilkason
# Date       : 11/8/2001
#------------------------------------------------------------------------------
proc macros::applyCurrentAPL {{resort 1}} {
   Trace
   # Re-sort if required, otherwise use the existing sorting order
   soundControl::latency long
   Wait 1
   if {$resort} {
      # Cut currently playing entry
      filtering::createMasterList
      Update
      # position it at top
   }
   ##
   # Change this to a specific function to apply an APL
   #
   catalog::applyAPL ;# right now, a sort is happening in poplist
   Wait 0
   soundControl::latency
}
#------------------------------------------------------------------------------
# Function   : RenameFileIP
# Description: Rename a file directly from the listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
#------------------------------------------------------------------------------
proc macros::RenameFile {{fullName {}}} {
   set W $::LB
   if {[string length $fullName] == 0} {
      set fullName [LB::nameActive]    ;# lookup long name
   }
   if {[string length $fullName]==0} {
      return
   } elseif {![file exists $fullName]} {
      tk_messageBox -icon warning -message "Could not rename '$fullName'"
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName
   ##
   # Block on dialog
   #
   set name [catalog::renameInPlaceDialog $W 0 $oldName]
   ##
   # Back from window closure
   #
   if {[string length $::TA(rename)]>0 && $::TA(rename) != $oldName} {
      set newFullName [file join $dirName $::TA(rename)]
      if {[catch {catalog::RenameFile $fullName $newFullName} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$result"
      } else {
         catalog::RenameEntry $fullName $newFullName
      }
   }
}
#------------------------------------------------------------------------------
# Function   : macros::RenameFile
# Description: Rename a file using a dialog, called from key binding or menu
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc macros::RenameFileDialog {{fullName {}}} {
   if {[string length $fullName] == 0} {
      set fullName [LB::nameActive]    ;# lookup long name
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName

   set answer [getEntryDialog "$oldName" newName "Rename File"]
   if {$answer == "ok" && $newName != $oldName} {

      set newFullName [file join $dirName $newName]
      if {[catch {catalog::RenameFile $fullName $newFullName} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$result"
      } else {
         catalog::RenameEntry $fullName $newFullName   ;# update full sorted list
      }
   }
}

#------------------------------------------------------------------------------
# Function   : SaveFolderFile
# Description: Save the current folder settings to a file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::SaveFolderFile {{fName ""}} {
   global TA LB snackAmpSettings
   ##
   # If a list for GlobData is passed in, the second item should be the name to save as (if different)
   #
   if {$fName == ""} {
      set types [list [list "Auto-Playlists Files" .tafs] [list "All Files" "*"]]
      set fName [tk_getSaveFile -title "Save Auto-Playlists File" -filetypes $types \
         -initialdir $::HOME -defaultextension .tafs]
      catch {focus $ThisWin}
      if { $fName == "" } {
         return  ""
      }
   }
   # Force extension in case dialog lost it (as on linux)
   set fName "[file rootname $fName].tafs"
   if {[catch {open $fName w} InFile]} {
      set result [tk_messageBox -type ok -title "SnackAmp - Please Respond" -icon warning \
         -message "Save Error: Unable to open '$fName' for writing"]
      return ""
   }
   fconfigure $InFile -buffersize 100000

   set snackAmpSettings(lastFolderFile) $fName
   set TA(AutoPlayListSelection) [file rootname [file tail $fName]]
   if {$InFile!=0} {
      foreach {File} [aplFolder::foldersMatchingPattern * 1] {
         puts $InFile [list $File [aplFolder::Get $File]]
      }
      close $InFile
      if {[APLExists]} {
         Tree::setDirty $::TREE 0
      }
      set snackAmpSettings(lastFolderFile) $fName
      return $fName
   } else {
      return -1
   }
}

#------------------------------------------------------------------------------
# Function   : removeAPLFile
# Description: Remove an APL file, called from menu callback
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::removeAPLFile {fName} {
   set result [tk_messageBox -type yesno -title "Please Answer" \
      -icon question -message "Are you sure you want to remove APL setting '[file rootname [file tail $fName]]'?"]
   if {[string equal $result "yes"]} {
      if {[catch {file delete -force $fName} result] } then {
         tk_messageBox -type ok -title "Error" \
            -icon warning -message "Failed to remove '[file rootname [file tail $fName]]'\n$result"
      }

   }
}
#------------------------------------------------------------------------------
# Function   : removePLFile
# Description: Remove an playlist file, called from menu callback
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::removePLFile {fName} {
   set result [tk_messageBox -type yesno -title "Please Answer" \
      -icon question -message "Are you sure you want to playlist '[file rootname [file tail $fName]]'?"]
   if {[string equal $result "yes"]} {
      if {[catch {file delete -force $fName} result] } then {
         tk_messageBox -type ok -title "Error" \
            -icon warning -message "Failed to remove '[file rootname [file tail $fName]]'\n$result"
      }

   }
}
#------------------------------------------------------------------------------
# Function   : readAPLFile
# Description: Read a file and change the favorate folders
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::readAPLFile {fName {apply 0}} {
   Trace
   global TA snackAmpSettings
   Wait 1
   ;# Build default APL folders if not yet done
   # Save a corresponding playlist file
   set lf $::snackAmpSettings(lastFolderFile)
   if {[file exists $fName]} {
      if {[aplFolder::entries]==0} {
         aplFolder::createTeFolders 0 1
      }
      ;# Can we remove the option below?
      ##
      # TBD: Move this code below to aplFolder namespace
      #
      if {$::snackAmpSettings(savePlonApply) && [file exists $lf] && $lf!=$fName} {
         set plfile "[file rootname $lf].m3u"
         file delete -force $plfile
         SavePlaylistFile $::LB replace all $plfile
         Update
      }
      if {[catch {
         set fid [open $fName r]
      } result] \
      } then {
         tk_messageBox -icon warning -type ok -title "Open Error" -message $result
         return
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]
      close $fid
      # Invalidate all current settings
      aplFolder::SetSome * -1
      # Only update values if an entry already exists from prior scan
      # This prevents garbage if directories were moved
      foreach line $full {
         foreach {Index Value} $line {
            aplFolder::SetIfExists $Index $Value
         }
      }
      Update
      ##
      # Some folders will still be tagged with -1 from
      # above invalidation, meaning they weren't in the new autoplaylist
      # file. If so, then set them to the value of their parent if possible.
      # Also, set any parents to a 2 if they have children with checks.
      #
      foreach {Index} [aplFolder::foldersMatchingPattern * 1] {
         aplFolder::setTeDefault $Index
      }
      if {[winfo exists .wFC]} {
         APL::refreshTree
      }
      ;# See if we need to apply it also
      if {$apply} {
         applyCurrentAPL 0
         Wait 1
      }
   } else {
      # If no default file create one with everything one
      ##
      # Create the default Folder list based on files
      # New ones are turned off by default.
      # We should already have the teFolders set to some value
      #
      aplFolder::SetSome * 1
      SaveFolderFile $fName
      applyCurrentAPL 0
      Wait 1
      waitWindow::close
      set result [tk_messageBox -type yesno -title "SnackAmp - Please Respond" -icon question \
         -message "All of your Auto-playlist folders containing music are set true \nby default, would you like to change them now?"]
      if {$result == "yes"} {
         AutoPlaylist
      }

   }
   set snackAmpSettings(lastFolderFile) $fName
   if {[APLExists]} {
      Tree::setDirty $::TREE 0
   }
   set TA(AutoPlayListSelection) [file rootname [file tail $fName]]
   Wait 0
}
# trace variable teFolders w ttttt
# proc ttttt {args} {
#    Trace 1
# }
#------------------------------------------------------------------------------
# Function   : saveCurrentAPL
# Description: Save and then apply the current Auto-playlist
# Author     : Tom Wilkason
# Date       : 11/8/2001
#------------------------------------------------------------------------------
proc macros::saveCurrentAPL {{apply 1}} {
   global snackAmpSettings
   SaveFolderFile $snackAmpSettings(lastFolderFile)
   if {$apply} {
      applyCurrentAPL 0
   }
   if {[APLExists]} {
      Tree::setDirty $::TREE 0
   }
}
#------------------------------------------------------------------------------
# Function   : Clear
# Description: Clear the playlist
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc macros::Clear {args} {
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to clear the playlist"]=="yes"} {
      global TA LB
      LB::deleteAll
   }
}
#------------------------------------------------------------------------------
# Function   : OpenFile
# Description: Open a single file for playing, add it to play list and play it
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::OpenFile {args} {
   global TA LB
   if {[catch {
      # This isn't in earlier snack versions. Also needs
      # Tcl 8.43+ to work, therefore the catch.
      set files [snack::getOpenFile -format MP3 -multiple 1]
   } result] } then {
      set files [snack::getOpenFile -format MP3]
   }
   # Here is a funky way to tell if a list of files was returned
   # Look at the string rep.
   if {![string match "\{*" $files]} {
      LB::InsertPlay $files
   } elseif {[llength $files] > 0} {
      LB::userInsertList $files
   }
}
#------------------------------------------------------------------------------
# Open up a URL, save a shortcut and play it
#------------------------------------------------------------------------------
proc macros::OpenURL {args} {
   if {[playurl::checkuskfolder]} {
      set shortcut [playurl::addPromptUrl $::snackAmpSettings(uskDir)]
      LB::InsertPlay $shortcut
   }
}
#------------------------------------------------------------------------------
# Function   : macros::Append
# Description: Wrapper for LB::AppendFile for macros & registry
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Append {file} {
   LB::AppendFile $file
}
#------------------------------------------------------------------------------
# Function   : macros::PlayResume
# Description: Play button binding, resume play if paused otherwise
#              play active
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::PlayResume {args} {
   switch -- [soundControl::playState] {
      "Playing" {
         # Restart
         Play
      }
      "Paused" {
         # Resume play
         Pause
      }
      "Stopped" {
         # Play
         Play
      }
   }
}
#------------------------------------------------------------------------------
# Function   : macros::skipPlay
# Description: Skip some number of seconds in the playing track
# Author     : Tom Wilkason
# Date       : 8/9/2003
#------------------------------------------------------------------------------
proc macros::skipPlay {amount} {
   set elapsed [soundControl::position]
   set ::TA(setdrag) 1
   set newTime [expr {$elapsed+$amount}]
   if {$newTime < 0} {
      set newTime 0
   }
   gui::setTime $newTime
   PlayOffset
}
#------------------------------------------------------------------------------
# Function   : Play
# Description: Play a list item. Bound to Play key but used elsewhere.
#              Updates the GUI with the song info. Allows an offset within
#              the song to be passed in.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Play {{What active} {offset -1} {pause 0}} {
   Trace
   global TA LB snackAmpSettings
   ##
   # Determine if the current active index
   # is a valid song. If so play it. Otherwise, flag it and jump to the next one.
   #
   if {$snackAmpSettings(NextAfterCurrnt)} {
      set i [LB::idxPlaying]
      if {$i < 0} {
         set i [LB::idxActive]
      }
   } else {
      set i [LB::idxActive]
   }
   ##
   # Based on prev, current or next position cursor to proper
   # file
   #
   switch -- $What {
      current {
         Stop
         if {$i == "" || $i<0} {
            set i 0
         }
      }
      active {
         # Always use active (e.g double click)
         Stop
         set i [LB::idxActive]
         if {$i == "" || $i<0} {
            set i 0
         }
      }
      next {
         # Only called, on automatic transition to next song
         # Can be called just before other song complete.
         # Don't issue a stop to keep the snack queue flowing.
         set soundControl::playState "Stopped"
         killAfter soundControl::Timer
         gui::setTime 0
         if {$i == "" || $i<0} {
            set i 0
         } else {
            incr i
         }
         # check for continuous play
         if {$i >= [LB::idxEnd]} {
            if {$snackAmpSettings(ContinuousPlay)} {
               set i 0
               # Resort on a wrap if some form of random play
               filtering::sortPlaylist
            }
         }
      }
      prev {
         if {$i == ""} {
            set i 0
         } elseif {$i==0} {
            set i [LB::idxEnd]
         } else {
            if {$i > 0} {
               incr i -1
            }
         }
      }
      default {
         saLog "Invalid subcommand :Play $What"
      }
   }
   ##
   # Get the track name at this index, translate if possible
   #
   set SelectionName [LB::nameFromIdx $i 1]
   set TA(currentMF) $SelectionName

#    set Selection [LB::keyFromIndex $i]
#    # Check file existance
#    if {$Selection=={}} {
#       return
#    }
#    # currentMF is song selected for play
#    set TA(currentMF) [LB::keyToName $Selection]
#    set SelectionName [LB::keyToName $Selection]
   ##
   # If file does not exist then assume it needs to be mounted
   #
   if {[string length $SelectionName] == 0} {
      # macros::Highlight this file then skip to the next one
      # if we haven't been commanded to stop
      saLog "Play: Could not validate '[LB::nameFromIdx $i]' as valid file"
      set ::songArray(Title) "'[LB::nameFromIdx $i]' invalid"
      set ::songArray(Artist) "File not verifiable"
      set ::songArray(Album) "Error"
      LB::activate $i
      set LB::indexPlaying $i
      LB::invalidateIndex $i
      if {$::TA(STOP) == 0} {
         after 1000 Next
         # Next will clear background so delay
         after 1500 [list LB::invalidateIndex $i]
      }
      return
   }
   # File exists, inverse video it showing it is playing
   # large playlists are slow to inverse, so delay the inverse
   # till after track is playing
   set LB::indexPlaying $i
   #LB::inverseVideoIndex $i
   after idle [list LB::inverseVideoIndex $i]
#   after idle [list LB::inverseVideoFile $TA(currentMF) 1]
   ##
   # See if user changed the slider to adjust where to play
   # will reset any fading that happened
   #
   PlayGain $TA(gain)
   ##
   # If this is a url to play, then play it directly
   #
   if {[playurl::isusk $SelectionName]} {
      soundControl::reset url
      foreach {index} [array names ::songArray] {
         set ::songArray($index) ""
      }
      id3Tag::id3Label $SelectionName ::songArray

      set ::songArray(Title) "Connecting ... $::songArray(Title)"
      set ::songArray(Album) "$SelectionName"
      if {[catch {playurl::playusk [soundControl::sound] $SelectionName [soundControl::getNextCommand] ::songArray} result] } then {
         saLog $::errorInfo
         set ::songArray(Title) "Error: $result"
         after 200 [list eval [soundControl::getNextCommand]]
         return
      } else {
         soundControl::pauseRelief 0
         set max [expr {3600*24}]      ;# 24 hours
         set gui::songDuration $max
         if {$::TA(showTicks)} {
            $gui::tsScale configure -to $max -tickinterval [expr {3600*4}]
         } else {
            $gui::tsScale configure -to $max
         }
         balloon::define $::SA.panel.stat.l.dtTitle $SelectionName
         if {[winfo exists .miniPlayer]} {
            balloon::define .miniPlayer.tt.title $SelectionName
         }
         catch {$miniPlayer::mpSS configure -to $max}
         set songTitle [string map {// /} [file rootname [join [lrange [file split $TA(currentMF)] end-1 end] /]]]
         wm title $::SA "$songTitle"
         set TA(songTitle) $songTitle

         if {![afterExists soundControl::Timer]} {
            soundControl::Timer
         }
      }
      return
   }
   ##
   # Process a file
   #
   ##
   # It seems faster on play -start to have the short file name and cd to that directory
   # Also, UNC names are really slow (on Win2K anyway)!
   # Strangly, network drive mappings are also fast (if not using UNC)
   #
   soundControl::reset file
   cd [file dirname $SelectionName]
   if {[catch {soundControl::fileName $SelectionName} result]} {
      tk_messageBox -icon warning -type ok -title "sound error" \
         -message "Error creating sound object using $SelectionName\n$result\n$::errorInfo"
      return
   }
   ##
   # Check for an invalid header
   #
   foreach {sndSamples sndRate x x x x sndExt} [soundControl::sinfo] {break}
   if {$sndSamples == -1 || $sndRate==0 || [lsearch -exact $::TA(upcaseExtensions) $sndExt] < 0} {
      soundControl::fileName ""
      saLog "Invalid file: $SelectionName"
      set ::songArray(Title) "Invalid header:$SelectionName"
      update
      if {$snackAmpSettings(AutoDelete)} {
         Delete
      }
      after 1000 [list eval [soundControl::getNextCommand]]
      return
   }
   # #
   # Get ID3 and detailed song info and display it, if ID3 not available
   # Use file name as song title
   catalog::trackTextData $SelectionName ::songArray
   #catalog::trackFullData $SelectionName ::songArray
   if {$::songArray(ID3)=={}} {
      set ::songArray(Title) [stringTitle [file rootname [file tail $TA(currentMF)]]]
   }
   # This should be use when applying an APL
   if {$What == "next"} {
      # Check for minimum rating before play
      set rating [ratingToNum [db::getVal $SelectionName Rating]]
      if {$rating > 0 && ($rating < [ratingToNum $::snackAmpSettings(minTxtRating)])} {
         set ::songArray(Title) "(Rating) Skipping $::songArray(Title)"
         saLog "(Rating) Skipping $SelectionName"
         if {$::TA(STOP)==0} {
            after idle Next
         }
         return
      }
      ##
      # Verify minimum bit rate
      #
      set rate [lindex $::songArray(Rate) 0]
      if {$rate < $::TA(minRate)} {
         set ::songArray(Title) "(Bitrate) Skipping $::songArray(Title)"
         saLog "(Bitrate) Skipping $SelectionName"
         if {$::TA(STOP)==0} {
            after idle Next
         }
         return
      }
   }
   ##
   # Adjust gain if the song has an auto level setting (and it is enabled)
   #
   if {$snackAmpSettings(autoLevel)} {

      set value [gainControl::createIfNeeded $SelectionName]
      ##
      # Pre compute the next track in the playlist if possible
      # Wait about 20 seconds to do so
      set nextSelection [LB::nameFromIdx [expr {$i+1}] 1]
      if {[string length $nextSelection]} {
         after 20000 [list gainControl::createIfNeeded $nextSelection]
      }
      # Clamp db Gain at runtime
      if {[catch {toDb $value} db] } then {
         set db 0.0
      } else {
         if {$db > $::snackAmpSettings(maxGain)} {
            set db $::snackAmpSettings(maxGain)
         } elseif {$db < $::snackAmpSettings(minGain)} {
            set db $::snackAmpSettings(minGain)
         }
      }
      snack::audio scaling [fromDb $db]
   } else {
      snack::audio scaling 1.0
   }
   # Fires Next command at end of this file
   if {$offset < 0} {
      set offset [gainControl::getStart [soundControl::fileName]]
   }
   soundControl::play \
      -start $offset \
      -command [soundControl::getNextCommandBump [soundControl::fileName] $::snackAmpSettings(trackGap)]
   if {$pause} {
      Pause
   }
#       set ::songArray(Title) "Error: $SelectionName"
#       snack::audio selectOutput $::snackAmpSettings(outputDevice)
#       saLog "Play: $result\n$::errorInfo"
#       after 1000 [list eval [soundControl::getNextCommand]]
#       return

   # Only log this if going to a file
   if {$::snackAmpSettings(logFile)} {
      saLog "Playing $TA(currentMF)"
   }
   ##
   # What ~10 ticks on 5 or multiple sec intervals
   #
   set trackTime  [soundControl::duration]
   if {$TA(showTicks)} {
      set Interval [expr {(round($trackTime/3.3)/10)*10}]
      if {$Interval == 0} {
         set Interval 10
      }
      $gui::tsScale configure -to $trackTime -tickinterval $Interval
   } else {
      $gui::tsScale configure -to $trackTime
   }
   # Miniplayer (catch if not up)
   if {[info exists miniPlayer::mpSS] && [winfo exists $miniPlayer::mpSS]} {
      $miniPlayer::mpSS configure -to $trackTime
   }
   ##
   # Song Title is shown in the title bar
   #
   set songTitle [string map {// /} [file rootname [join [lrange [file split $TA(currentMF)] end-1 end] /]]]
   # If the folder name is the same as the first part of file name,
   # Then we can ignore the folder name
   if {[string match -nocase "[file dirname $songTitle]*" [file tail $songTitle]]} {
      set songTitle [file tail $songTitle]
   }
   wm title $::SA "$songTitle"
   set TA(songTitle) $songTitle

   #set TA(songInfo) "$Details(fSize) - $Details(fDate) - $Details(Rate)"
   set gui::songDuration $::songArray(Duration)
   if {$snackAmpSettings(autoSavePlaylist)} {
      catalog::SavePlaylistCache
   }
   if {![afterExists soundControl::Timer]} {
      soundControl::Timer
   }
}
#----------------------------------------------------------------------------
# Play next sound file in list
#----------------------------------------------------------------------------
proc macros::Next {} {
   Trace
   Stop
   Play next
}
#----------------------------------------------------------------------------
# Play the next track right now
#----------------------------------------------------------------------------
proc macros::NextNow {} {
   Trace
   Stop
   Play next
}
#----------------------------------------------------------------------------
# Play previous sound file in list
#----------------------------------------------------------------------------
proc macros::Prev {} {
   Trace
   Stop
   after idle Play prev
}
#----------------------------------------------------------------------------
# Shuffle the selected files in the playlist
#----------------------------------------------------------------------------
proc macros::shuffleSelection {args} {
   LB::shuffleSelection ::LB
}
#----------------------------------------------------------------------------
# Sort the selected files in the playlist
#----------------------------------------------------------------------------
proc macros::sortSelection {args} {
   LB::sortSelection ::LB
}
#----------------------------------------------------------------------------
# Stop playback of a file and cancle the update timer
#----------------------------------------------------------------------------
proc macros::Stop {args} {
   Trace
   global TA LB
   killAfter soundControl::Timer
   if {[snack::audio active]} {
      soundControl::stop
      gui::setTime 0
   }
   soundControl::pauseRelief 2
   playurl::cleanupLast
}
#------------------------------------------------------------------------------
# Move the cursor to the currently playing file,
# assign to <Home> key
#------------------------------------------------------------------------------
proc macros::locatePlaying {args} {
   LB::seePlaying
}

#------------------------------------------------------------------------------
# Pause/resume playback
#------------------------------------------------------------------------------
proc macros::Pause {args} {
   Trace
   global TA LB
   if {[snack::audio active]} {
      if {$soundControl::paused==1} {
         soundControl::pause 0
         soundControl::Timer
      } else {
         killAfter soundControl::Timer
         soundControl::pause 1
      }
   }
}
#------------------------------------------------------------------------------
# Callback to remove the current selections from the display
#------------------------------------------------------------------------------
proc macros::Cut {{playNext 0}} {
   Trace
   global TA LB
   set playActive 0
   set playing [LB::idxPlaying]
   if {[llength [LB::idxSelected]] > 0} {
      set Index [LB::idxAnchor]
      foreach {ind} [lsort -decreasing -integer [LB::idxSelected]] {
         if {$playing==$ind} {
            set playActive [soundControl::isPlaying]
            if {$playActive} {
               Stop
            }
         }
         LB::deleteIndex $ind
      }
      # Stay on 0 if that is what we were on
      ##macros::Highlight [expr {$Index-0}]
      gui::updateFileDisplay
      # Set selection to current active
      LB::setSelection [LB::idxActive]
      # This will play the next song
      if {$playNext && $playActive} {
         Play
      }
   }
}
#------------------------------------------------------------------------------
# Cut the listbox to the clibboard
# (only used for main playlist)
#------------------------------------------------------------------------------
proc macros::clipboardCut {lb {append 1}} {
   Trace
   global TA
   if {[llength [$lb curselection]] > 0} {
      clipboardCopy $lb $append
      Cut 1
      gui::updateFileDisplay
   }
}
#------------------------------------------------------------------------------
# Copy the listbox files to the clipboard
#------------------------------------------------------------------------------
proc macros::clipboardCopy {lb {append 1}} {
   Trace
   global TA LB
   if {[llength [$lb curselection]] > 0} {
      if {$append == 0} {
         clipboard clear
      } else {
         clipboard append " "
      }
      foreach {ind} [lsort -integer [$lb curselection]] {
         switch -glob -- $lb {
            ".saTop*"  {lappend data [LB::keyToName [LB::keyFromIndex $ind]]}
            ".wFC*"    {lappend data [aplLB::rowIndex2file $ind]}
         }
      }
      clipboard append $data
   }
}
#------------------------------------------------------------------------------
# Paste the listbox files to the playlist
# Note: Only works with LB playlist
#------------------------------------------------------------------------------
proc macros::clipboardPaste {lb {clear 1}} {
   Trace
   global TA
   if {[catch {selection get -selection CLIPBOARD} files]} {
      return
   }
   set current [$lb index anchor]
   set processed [list]
   foreach {file} $files {
      # Only insert files that are in the master list
      # And not already processed (i.e. no duplicates)
      #W: move to shared
      if {[lsearch -exact $processed $file] < 0 && [db::whereInDb $file] >= 0} {
         lappend processed $file
      }
   }
   if {[llength $processed]>0} {
      LB::insertList [LB::idxActive] $processed
   }
   if {$clear} {
      clipboard clear
   }
}
#------------------------------------------------------------------------------
# Callback to remove the current selections from the disk
#------------------------------------------------------------------------------
proc macros::Delete {{Next 1}} {
   Trace
   global TA LB
   set goNext 0
   if {[llength [LB::idxSelected]] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to remove files"
         return $goNext
      }
      set Index [LB::idxAnchor]

      set selection [LB::idxSelected]
      foreach {ind} $selection {
         lappend Files [LB::keyFromIndex $ind]
      }

      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to delete:\n[join $Files \n]?"]
      if {$result == "yes"} {
         foreach {Entry} $Files {
            set File [LB::keyToName $Entry]
            set goNext [catalog::DeleteFile $File]
            Cut $goNext
         }
         ##macros::Highlight $Index
      }
   }
   ##
   # Play the next song if we deleted the current one and we want to
   #
   if {$Next && $goNext} {
      Play
   }
   gui::updateFileDisplay
   return $goNext
}
#----------------------------------------------------------------------------
# Bring up or down the Auto Playlist window, depending on its state
#----------------------------------------------------------------------------
proc macros::AutoPlaylist {args} {
   if {![winfo exists .wFC]} {
      package require saTree
      package require autoPlayList
      $::TA(treeButton) configure -relief sunken
      # First make sure we have the proper roots for the tree viewer
      APL::FolderTree
      update  ;# let it draw
      # May not have a selection
      catch {APL::SelectFile $::TA(currentMF)}
   } elseif {[wm state .wFC] != "normal"} {
      $::TA(treeButton) configure -relief sunken
      APL::FolderTree
      update  ;# let it draw
      catch {APL::SelectFile $::TA(currentMF)}
   } else {
      wm withdraw .wFC
      $::TA(treeButton) configure -relief flat
   }
}
#------------------------------------------------------------------------------
# Function   : macros::findNextSong
# Description: Locate the next occurance of a song in the playlist based on the
#              current filter value
# Author     : Tom Wilkason
# Date       : 11/08/2001
#------------------------------------------------------------------------------
proc macros::findNextSong {{top 0}} {
   global TA snackAmpSettings
   # Get the current playlist
   set regPattern [list]                                ;#retain wildcards
   set pattern [split [string map {" or " "|" " OR " "|" "*" ".*" "?" "."} [catalog::globEscape $TA(filter)]] |]
   foreach part $pattern {
      lappend regPattern "($part)+"
   }
   set regPattern [join $regPattern |]
   if {$top} {
      set i -1
   } else {
      set i [lindex [LB::idxSelected] 0]
      if {$i=={}} {
         set i -1
      }
   }
   set ok 2
   ##
   # Allow one wrap around if at first not found
   #
   if {[catch {
      while {$ok} {
         incr i
         foreach {Entry} [LB::keyFromIndex $i end] {
            if {[regexp -nocase -- $regPattern [LB::keyToName $Entry]]} {
               macros::Highlight $i
               return
            }
            incr i
         }
         set i -1
         incr ok -1
      }
   } result]} {
      #saLog "findNextSong $result"
   }
}
#------------------------------------------------------------------------------
# Function   : macros::playSpecificSong
# Description: Locate and Play a specific song in the playist, ignore if not in playlist
# Author     : Tom Wilkason
# Date       : 4/28/2002
#------------------------------------------------------------------------------
proc macros::playSpecificSong {song} {
   Trace
   set i 0
   set i [lsearch -exact [LB::keyOrNamesAll] $song]
   if {$i >= 0} {
      LB::activate $i
      Play active
   }
}
#------------------------------------------------------------------------------
# Function   : macros::clearFilter
# Description: Clear the current filter entry box
# Author     : Tom Wilkason
# Date       : 11/9/2001
#------------------------------------------------------------------------------
proc macros::clearFilter {args} {
   set ::TA(filter) {}
}

#------------------------------------------------------------------------------
# Function: macros::selectAll
#
# Select all the files in the playlist
#------------------------------------------------------------------------------
proc macros::selectAll {args} {
   LB::setSelection 0 end
}
#------------------------------------------------------------------------------
# Function   : macros::filterMusic
# Description: Restrict files shown based on the filter & folder settings
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc macros::filterMusic {how {what def}} {
   filtering::filterMusic $how $what
}
#------------------------------------------------------------------------------
# Function   : macros::SavePlaylistFile
# Description: Save current files in list
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::SavePlaylistFile {ListBox {how "replace"} {what "selected"} {fName {}}} {
   Trace
   global TA LB
   ##
   # If a list for GlobData is passed in, the second item should be the name to save as (if different)
   #
   if {$how=="replace"} {
      set access "w"
   } else {
      set access "a+"
   }
   # Use different dialog depending on what you want to do
   array set tk_dialog {
      replace tk_getSaveFile
      append tk_getOpenFile
   }
   if {$what == "selected"} {
      set Indicies [LB::idxSelected]
   } else {
      set Indicies [LB::idxAll]
   }
   if {[string length $fName]==0} {
      set types [list [list "Playlist files" {.m3u .taf}] [list "All Files" "*"]]

      set fName [$tk_dialog($how) -title "Save [llength $Indicies] tracks to Playlist File" -filetypes $types \
         -initialdir $::HOME -defaultextension .m3u]
      catch {focus $ThisWin}
      if { $fName == "" } {
         return  ""
      }
      # Set default extension to m3u
      if {[file extension $fName]==""} {
         set fName "$fName.m3u"
      }
   } elseif {[file exists $fName]} {
      if {$how=="replace" && [tk_messageBox -icon question -type yesno -title "Overwrite File" -message "Is it ok to $how $fName with [llength $Indicies] tracks?"]!="yes"} {
         return {}
      }
   }
   if {[catch {open $fName $access} InFile]} {
      tk_messageBox -icon warning -type ok -title "File Error" -message "Save Error: Unable to open '$fName' for writing"
      return {}
   }
   set TA(PlayListSelection) [file rootname [file tail $fName]]
   set TA(PlayListFile) $fName
   set ext [file extension $fName]
   if {[string equal $ext .m3u]} {
      puts $InFile "#\n# Playlist file: [file nativename $fName]"
      puts $InFile "# Saved on [clock format [clock seconds]] by user $::tcl_platform(user) on host [info hostname]\n#"
   }
   if {$InFile!=0} {
      ##
      # Process either selected or entire list
      #
      set i 1
      set names [list]
      if {$ListBox == $::LB} {
         foreach {Index} $Indicies {
            set File [LB::keyToName [LB::keyFromIndex $Index]]
            lappend names $File
            if {[string equal $ext .m3u]} {
               set File [file nativename $File]
            }
            puts $InFile $File
         }
      } elseif {$ListBox == $::PLB} {
         if {$what == "selected"} {
            set names [aplLB::nameSelected 0]
         } else {
            set names [aplLB::getAllNames 1]
         }
         foreach {File} $names {
            if {[string equal $ext .m3u]} {
               set File [file nativename $File]
            }
            incr i
            puts $InFile $File
         }
      } else {
         saLog "unknown listbox '$ListBox'"
      }
      #W: Only save if playing is in list
      if {[lsearch -exact $names $::TA(currentMF)]>= 0} {
         puts $InFile "#Playing=$::TA(currentMF)"
      }
      close $InFile
      return $fName
   } else {
      return -1
   }
}
#------------------------------------------------------------------------------
# Function   : macros::ReadPlaylistFile
# Description: Verify playlist file and notify user of any problems
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::ReadPlaylistFile {fName {how "replace"}} {
   # If file doesn't exist, then prompt for it
   if {$fName == {} || ![file exists $fName]} {
      set types [list [list "Playlist files" {.m3u .taf}] [list "All Files" "*"]]
      set fName [tk_getOpenFile -title "Open Playlist File" -filetypes $types \
         -initialdir $::HOME]
      if { $fName == "" } {
         return  ""
      }
   }
   set failed [ProcessPlaylistFile $fName $how 1]
   set numFailed [llength $failed]
   if {$numFailed > 0} {
      ##
      # Tell user how many failed (if any)
      #
      if {$numFailed > 10} {
         set failed [lrange $failed 0 9]
         lappend failed "...."
      }
      tk_messageBox -icon warning -type ok -title "Invalid Files" \
         -message "A total of $numFailed files do no exist, and could not be added to the playlist\n[join $failed \n]"
   }
   return $numFailed
}
#------------------------------------------------------------------------------
# Function   : macros::ProcessPlaylistFile
# Description: Read a file and append/replace current list
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::ProcessPlaylistFile {fName {how "replace"} {resume 0}} {
   Trace
   global TA LB snackAmpSettings
   if {[catch {
      set fid [open $fName r]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
      return
   }
   Wait 1
   set full [split [string trim [read $fid]] \n]
   close $fid
   # Insert the files if they are not already in the list
   # Translate Removable media if necessary.
   # Also support winamp format playlists, this means no files can
   # start with # or else they will be ignored (that is a comment in m3u)
   set here [file dirname $fName]
   set push [pwd]
   cd $here
   set comment "\#*"
   set failed [list]
   set i 0
   set cp [LB::idxActive]
   set toinsert [list]
   set processLines [llength $full]
   foreach file $full {
      incr i
      if {! [string match $comment $file]} {
         # If a url, create a usk and process the usk file
         if {[string match -nocase "http://*" $file]} {
            # Create a usk (url) shortcut, then append it
            # Note: trying tail of file name, works if a file if a stream
            # may not work as well.
            set file [playurl::addUrl $::snackAmpSettings(uskDir) [file tail $file] $file]
         } else {
            # This will be ignored if abs path names
            set file [file join $here $file]
         }
         # Verify existence (can be slow on a net)
         if {[set rfile [catalog::TranslateMount $file]] != {} } {
            lappend toinsert $rfile
         } else {
            lappend failed $file
         }
         if {$i%50==0} {
            waitWindow::show "($i of $processLines) Playlist Files Verified"
            update
         }   ;# let the music play
      } else {
         if {$resume && [string match "#Playing*" $file]} {
            set resumeFile [lindex [split $file =] 1]
            if {$::snackAmpSettings(restorePlaying) && [file exists $resumeFile]} {
               set ans [tk_messageBox -type yesno -icon question -message "Do you want to resume playing of \n$resumeFile"]
               if {$ans=="yes"} {
                  after 1000 [list PlayFile $resumeFile]
               }
            }
         }
      }
   }
   ;# Insert the tracks
   if {[llength $toinsert] > 0} {
      if {$how=="replace"} {
         LB::deleteAll
         LB::insertList 0 $toinsert
      } else {
         LB::insertList $cp $toinsert
      }
   }
   Wait 0
   waitWindow::close
   set TA(PlayListFile) $fName
   set TA(PlayListSelection) [file rootname [file tail $fName]]
   cd $push
   gui::updateFileDisplay
   return $failed
}
#------------------------------------------------------------------------------
# Function   : macros::Insert
# Description: Wrapper for LB::userInsertList for macros & registry
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Insert {file} {
   LB::userInsertList [list $file]
}
#------------------------------------------------------------------------------
# Function   : PlayFile
# Description: Play a file at the top of the list (linked to DDE command on Windows)
#              Set this up early so a file association doesn't time out when launching
#              SnackAmp on a slow computer.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::PlayFile {file {justInsert 0}} {
   Trace
   set file [string map {\\ /} $file]
   set ext [string tolower [file extension $file]]
   switch -- $ext {
      .m3u
      -
      .pls {
         # If coming from a web site, create a usk file,
         # otherwise, read in the playlist and process each line
         if {[string match -nocase "http://*" $file]} {
            playurl::checkuskfolder
            set contents [split [playurl::getUrlBody $file] \n]
         } else {
            if {[file readable $file]} {
              set fid [open $file r]
              set contents [split [string trim [read $fid]] \n]
              close $fid
            } else {
               tk_messageBox -icon warning -type ok -message "File '$file' cannot be read"
               return
            }
         }
         set fileTail [file rootname [file tail $file]]
         ;# Handle playlist types uniquely
         switch -- $ext {
            .m3u {
                set toappend [list]
                set i 0
                foreach {row} $contents {
                   set drow [HttpD::Decode $row]  ;# not sure if this is needed
                   if {![string match "\#*" $drow]} {
                      set fullFile [file join [file dirname $file] $drow]
                      if {[string match -nocase "http://*" $drow]} {
                         # Create a usk (url) shortcut, then append it
                         set name [file tail $drow]
                         # Note: trying tail of file name, works if a file if a stream
                         # may not work as well.
                         set fname [playurl::addUrl $::snackAmpSettings(uskDir) "$name" $row]
                         lappend toappend $fname
                         incr i
                      } elseif {[file exists $fullFile]} {
                         lappend toappend $fullFile
                      } else {
                         saLog "File:'$row' does not exist"
                      }
                   }
                }
                set added 0
                set end [LB::userInsert]
                ;# If we inserted some, and not just inserting, play the first one
                if {[LB::userInsertList $toappend] > 0 && ($justInsert==0) } {
                   LB::activate $end
                   Play active
                }
            }
            .pls {
               set title $fileTail
               set i 0
               set Fnum 0
               set Tnum 0
               foreach {row} $contents {
                  if {[regexp -- {File([0-9]+)=(.*)} $row -> Fnum url]} {
                     set pls($Fnum,url) $url
                  }
                  if {[regexp -- {Title([0-9]+)=(.*)} $row -> Tnum title]} {
                     set title [string trim [file rootname $title] .]
                     set pls($Tnum,title) $title
                  }
               }
               set i 1
               foreach {index} [lsort -dictionary [array names pls *,url]] {
                  foreach {num type} [split $index ,] {break}
                  if {[info exists pls($num,title)]} {
                     set title $pls($num,title)
                  } else {
                     set title $fileTail.$num
                  }
                  set fname [playurl::addUrl $::snackAmpSettings(uskDir) $title $pls($index)]
                  if {$i==1} {
                    set end [LB::userInsert]
                  }
                  LB::userInsert $fname
                  incr i
               }
               if {$justInsert==0 && [info exists end]} {
                  LB::activate $end
                  gui::updateFileDisplay
                  Play active
               }

            }
         }
      }
      default {
         # usk shortcut files also handled here
         if {[file readable $file]} {
            if {$justInsert==1} {
               LB::userInsertList [list $file]
            } else {
               LB::InsertPlay $file
            }
         } else {
            tk_messageBox -icon warning -type ok -message "File '$file' cannot be read"
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : Exit
# Description: Exit gracefully
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc macros::taExit {args} {
   global TA snackAmpSettings
   catch {
      winico taskbar delete $TA(Icon)
   }
   #wm withdraw $::SA
   catch {APL::close}
   catalog::SavePlaylistCache
   # Stop after we save the cache
   soundControl::stop
   HttpD::Shutdown
   set snackAmpSettings(minWidth) [winfo width $::SA]
   foreach {w h x1 x2} [split [wm geometry $::SA] "x +"] {break}
   set snackAmpSettings(playerPos) [list $x1 $x2]
   settings::saveSettings
   gainControl::saveSettings
   db::Commit
   db::closeDB
   # close any remaining channels
   foreach {channel} [file channels] {
      catch {close $channel}
   }
   exit
}
#------------------------------------------------------------------------------
# Copy the selection to a staging file for burning to a CD
#------------------------------------------------------------------------------
proc macros::linkToStaging {lb {append 1}} {
   Trace
   global TA LB
   set data [list]
   if {[llength [$lb curselection]] > 0} {
      foreach {ind} [lsort -decreasing -integer [$lb curselection]] {
         switch -glob -- $lb {
            ".saTop*"  {lappend data [LB::keyToName [LB::keyFromIndex $ind]]}
            ".wFC*"    {lappend data [aplLB::rowIndex2file $ind]}
         }
      }
   }
   ;# Create N lists of files, for each root drive
   foreach file [lflip $data] {
      lappend toStage([fileRoot $file]) $file
   }
   ;# For each root drive, process the files
   foreach folder [array names toStage] {
      set staging [file join $folder "staging"]
      if {[file exists $staging]} {
         ;# First, find the maximum number to start with
         set i 0
         set existing [lsort -dictionary [glob -nocomplain [file join $staging *]]]
         foreach track $existing {
            set front [string trimleft [lindex [split [file tail $track] -] 0] "0"]
            if {[string is integer -strict $front] && $front > $i} {
               set i $front
            }
         }
         incr i
         saLog "Starting with $i"
         ##
         # Now link the files in
         #
         foreach file [lflip $toStage($folder)] {
            set num [format "%04d - " $i]
            set newName [file join $staging $num[file tail $file]]
            if {![file exists $newName]} {
               saLog "Creating $newName"
               if {[catch {file link $newName $file} result]} {
                  saLog $result
               } else {
                  incr i
               }
            }
         }
      } else {
         puts "$staging doesn't exist"
      }
   }
}

package provide macros 1.0
