#------------------------------------------------------------------------------
# snackAmp - minPlayer package 1.0 for snackAmp
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/miniPlayer.tcl,v 1.23 2004/02/05 04:57:22 wilkason Exp $
}

#------------------------------------------------------------------------------
# Function   : launch
# Description: Mini-player that pops up from the taskbar
# Author     : Tom Wilkason
# Date       : 2/1/2001
#------------------------------------------------------------------------------
namespace eval miniPlayer {
   variable mpSS ""
}
proc miniPlayer::launch {X Y} {
   set w .miniPlayer
   set ::snackAmpSettings(activeMiniPlayer) 1
   if {[winfo exists $w]} {
      focus -force $w.tt.title
      wm deiconify $w
      return
   }
   ##
   # Create miniPlayer window with no border, we will handle dragging
   #
   toplevel $w -relief raised -borderwidth 5 -class SnackAmp
   wm focusmodel $w active  ;# Does this force a focus on pop-up?
   # Bottom Frame
   # Song Title
   pack [frame $w.tt] -expand false -fill x -side top
   label $w.tt.title -textvariable ::TA(songTitle) \
      -bg $::snackAmpSettings(bg) -fg $::snackAmpSettings(fg) -font songInfo -width 20 \
      -anchor w -relief sunken -cursor fleur
   settings::regColor $w.tt.title -bg ::snackAmpSettings(bg) -fg ::snackAmpSettings(fg)
   pack $w.tt.title -expand true -fill x -side left

   # Close miniPlayer
   button $w.tt.close -command {miniPlayer::close} -image saImages::Close -padx 0 -pady 0 -borderwidth 1
   balloon::button $w.tt.close "Close mini-player"
   pack $w.tt.close -side right -padx 0 -pady 0

   frame $w.frame
   # Volume Slider
   scale $w.frame.volume -show no -orient horizontal -width 10 -command {PlayGain}\
         -len 120 -var TA(gain) -from 0. -to 64. -sliderlength 15 \
         -activebackground blue -troughcolor $::snackAmpSettings(bg)
   settings::regColor $w.frame.volume -troughcolor ::snackAmpSettings(bg)

   balloon::define $w.frame.volume "Adjust the volume"

   # Time Slider
   set trackTime  [soundControl::duration]

   set miniPlayer::mpSS [scale $w.frame.ss -show no \
      -orient horiz \
      -width 10 \
      -len 120 \
      -from 0. \
      -to $trackTime \
      -var gui::trackPosition \
      -sliderlength 5 \
      -activebackground blue \
      -troughcolor $::snackAmpSettings(bg) \
      -command gui::updateTime] ;#-cursor sb_h_double_arrow
   settings::regColor $miniPlayer::mpSS -troughcolor ::snackAmpSettings(bg)

   #balloon::define $miniPlayer::mpSS "Show/Adjust playback location within file"
   bind $miniPlayer::mpSS <ButtonPress-1> {
      set TA(setdrag) 0
   }
   # Show time when scale is dragged
   bind $miniPlayer::mpSS <B1-Motion> {
      miniPlayer::mpTime %W %X %Y
   }
   bind $miniPlayer::mpSS <ButtonRelease-1> {
      set TA(setdrag) 1
      PlayOffset
      destroy %W.balloon
   }
   bind $miniPlayer::mpSS <ButtonPress-2> {set TA(setdrag) 0}
   bind $miniPlayer::mpSS <ButtonRelease-2> {set TA(setdrag) 1; PlayOffset}
   bind $miniPlayer::mpSS <ButtonPress-3> {set TA(setdrag) 0}
   bind $miniPlayer::mpSS <ButtonRelease-3> {set TA(setdrag) 1; PlayOffset}

   pack $w.frame -padx 4 -pady 4
   pack $w.frame.volume -fill x -side left
   pack $w.frame.ss -fill x -side left
   balloon::widgetDefine .miniPlayer.tt.title

   # Since this is override direct, we have to handle drags
   bind  .miniPlayer <Enter> {
      focus -force .miniPlayer.tt.title
      wm deiconify .miniPlayer
   }
   # Pop-up info on the song
   if {$::tcl_platform(platform) != "windows"} {
      bind .miniPlayer <Button-5> [list bumpLevel -1]
      bind .miniPlayer <Button-4> [list bumpLevel 1]
   } else {
      bind .miniPlayer <MouseWheel> {
         set amount [expr {%D/abs(%D)*2}] ;# get +/- 2
         bumpLevel $amount
      }
   }
   bind  .miniPlayer.tt.title <ButtonPress-1> {
      # Save offset from corner to mouse
      set W [winfo toplevel %W]
      set ::SPLASH(X) [expr {%X - [winfo x $W]}]
      set ::SPLASH(Y) [expr {%Y - [winfo y $W]}]
      # Kill the little balloon
      destroy .miniPlayer.tt.title.balloon
   }
   # System Menu on RMB
   bind  .miniPlayer.tt.title <ButtonPress-3> {
      destroy .miniPlayer.tt.title.balloon
      tk_popup $::SA.bRmenubar %X %Y
   }
   bind  .miniPlayer.tt.title <B1-Motion> {
      set x [expr {%X - $::SPLASH(X)}]
      set y [expr {%Y - $::SPLASH(Y)}]
      wm geometry [winfo toplevel %W] "+$x+$y"
      set ::snackAmpSettings(positionMiniPlayer) [list $x $y]
      destroy .miniPlayer.tt.title.balloon
   }
   # Buttons for audio control
   set buttonData {
      bitmap snackPlayPrev Prev       "Play previous in list"
      bitmap snackPlay     PlayResume "Play Selection"
      bitmap snackPause    Pause      "Pause/Resume playback"
      bitmap snackStop     Stop       "Stop playing"
      bitmap snackPlayNext Next       "Play next in list"
   }
   # build each one
   set tf $w.frame
   foreach {bmType File Command Help} $buttonData {
      button  $tf.$File -$bmType $File -command $Command -borderwidth 1
      pack $tf.$File -side left -padx 0 -pady 0
      balloon::button $tf.$File $Help
   }
   # Support hiding of main window
   pack [checkbutton $tf.hide \
      -selectimage saImages::arrowhide -image saImages::arrowshow -indicatoron 0 -bd 0 \
      -selectcolor $::snackAmpSettings(bg) -variable TA(Hidden) -command {
      switch -- [wm state $::SA] {
         normal {wm withdraw $::SA; set TA(Hidden) 1}
         withdrawn {wm deiconify $::SA; set TA(Hidden) 0}
         iconic {wm withdraw $::SA; set TA(Hidden) 1}
         default {wm deiconify $::SA; set TA(Hidden) 0}
      }
   }] -side left
   balloon::define $tf.hide "Hide/Show the main snackAmp window."
   # Icon to pin it if can pin
   if {$ontop::enabled} {
      pack [checkbutton $tf.pin -image saImages::unpinit -selectimage saImages::pinit -indicatoron 0 -bd 0 \
         -selectcolor $::snackAmpSettings(bg) \
         -variable ::TA(.miniPlayer,pinned) \
         -command {ontop::control .miniPlayer $::TA(.miniPlayer,pinned)}] -side left
      balloon::define $tf.pin "Press to keep the mini player on top."
   }
   ##
   # Figure out where to put the window, TL corner is window 0,0
   #
   # Hide it and take away window decoration
   wm withdraw $w
   wm overrideredirect $w 1
   focus -force $w
   # with it hiden, let the layout take place so we know its geometry
   update idletasks

   # Get miniplayer size  and position at bottom right of usable area
   set rX [winfo reqwidth $w]
   set rY [winfo reqheight $w]
   # Don't place over taskbar
   # Locate the player in the right hand bottom corner then pop it up.
   if {($X+$Y)==0} {
      set x [expr {$::TA(screen_right)-$rX}]
      set y [expr {$::TA(screen_bottom)-$rY}]
   } else {
      set x $X
      set y $Y
   }
   wm geometry $w "+$x+$y"
   set ::snackAmpSettings(positionMiniPlayer) [list $x $y]
   wm deiconify $w
}
#------------------------------------------------------------------------------
# Function   : miniPlayer::mpTime
# Description: Show time on miniplayer when dragging mouse
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc miniPlayer::mpTime {W X Y} {
   set time [clock format [expr {int($gui::trackPosition)}] -format %T -gmt 1]
   balloon::show $W $X $Y "Time: $time"
}
#------------------------------------------------------------------------------
# Function   : miniPlayer::toggle
# Description: Toggle the mini player visibility
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc miniPlayer::toggle {{X 0} {Y 0}} {
   global TA
   if {$::snackAmpSettings(activeMiniPlayer) == 0} {
      miniPlayer::close
   } else {
      miniPlayer::launch $X $Y
   }
}
#------------------------------------------------------------------------------
# Function   : miniPlayer::close
# Description: Close the miniplayer and make sure main window is not also
#              withdrawn on non-windows systems (i.e. no task bar)
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc miniPlayer::close {args} {
   global TA
   set w .miniPlayer
   wm withdraw $w
   set ::snackAmpSettings(activeMiniPlayer) 0
   # If no taskbar icon, then make sure both windows
   # not closed.
   if {$::tcl_platform(platform) != "windows"} {
      if {[wm state $::SA] == "withdrawn"} {
         wm deiconify $::SA
         set TA(Hidden) 0
      }
   }
}
#------------------------------------------------------------------------------
# Function   : miniPlayer::iconCallback
# Description: Called when taskbar icon is "clicked"
# Author     : Tom Wilkason
# Date       : 1/4/2001
#------------------------------------------------------------------------------
proc miniPlayer::iconCallback {message win X Y} {
   switch -- $message {
      WM_LBUTTONDOWN {
         miniPlayer::launch 0 0
         Pause
      }
      WM_LBUTTONUP {}
      WM_MOUSEMOVE {}
      WM_RBUTTONUP {}
      WM_RBUTTONDOWN {
         if {[winfo exists .miniPlayer] && [winfo viewable .miniPlayer]} {
            miniPlayer::close
         } else {
            miniPlayer::launch 0 0
         }
      }
      default {}
   }
}

package provide miniPlayer 1.0
