#------------------------------------------------------------------------------
# snackAmp - saFiltering package 1.0 for snackAmp
# provides music filtering utilities
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/saFiltering.tcl,v 1.22 2004/02/05 04:55:47 wilkason Exp $
}
namespace eval filtering {
   variable sortModes [list "Sort by all Tracks" "Shuffle by all Tracks" "Shuffle by Folder, Sort by Track" "Shuffle by Folder, Shuffle by Track"]
}
#------------------------------------------------------------------------------
# Function   : filtering::filterHistory
# Description: postcommand callback to Display the history filter, called from
#              button next to filter box.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc filtering::filterHistory {menu} {
   global snackAmpSettings
   # Only keep up to last to filters
   set snackAmpSettings(History) [lrange $snackAmpSettings(History) 0 20]
   $menu delete 0 end
   destroy $menu.replace
   destroy $menu.append
   destroy $menu.select
   destroy $menu.insert
   $menu add cascade -label "Append To Playlist..." -menu [menu $menu.append -tearoff 0 -font menu]
   $menu add cascade -label "Insert Into Playlist..." -menu [menu $menu.insert -tearoff 0 -font menu]
   $menu add cascade -label "Replace Playlist..." -menu [menu $menu.replace -tearoff 0 -font menu]
   $menu add cascade -label "Select In Playlist..." -menu [menu $menu.select -tearoff 0 -font menu]

   $menu.append add command -label "Filtering '$::TA(filter)' on all known files" -command {
      filtering::pushFilter
      macros::filterMusic append all
   }
   $menu.append add command -label "Filtering '$::TA(filter)' on Auto-playlist files" -command {
      filtering::pushFilter
      macros::filterMusic append apl
   }

   $menu.select add command -label "Filtering '$::TA(filter)' on current Playlist" -command {
      filtering::pushFilter
      macros::filterMusic select playList
   }

   $menu.replace add command -label "Filtering '$::TA(filter)' on all known files" -command {
      filtering::pushFilter
      macros::filterMusic replace all
   }
   $menu.replace add command -label "Filtering '$::TA(filter)' on Auto-playlist files" -command {
      filtering::pushFilter
      macros::filterMusic replace apl
   }
   $menu.replace add command -label "Filtering '$::TA(filter)' on current Playlist" -command {
      filtering::pushFilter
      macros::filterMusic replace playList
   }

   $menu.insert add command -label "Filtering '$::TA(filter)' on Auto-playlist files" -command {
      filtering::pushFilter
      macros::filterMusic insert apl
   }
   $menu.insert add command -label "Filtering '$::TA(filter)' on current Playlist" -command {
      filtering::pushFilter
      macros::filterMusic insert playList
   }

   $menu add separator
   $menu add command -label "Reset Filter to Auto-playlist" -command {
      set ::TA(filter) {}
      macros::filterMusic replace $::TA(matchAgainstAll)
   }
   $menu add separator
   foreach {item} $snackAmpSettings(History) {
      $menu add command -label "$item" -command "[list set ::TA(filter) $item]"
   }
}
#------------------------------------------------------------------------------
# Function   : filtering::pushFilter
# Description: Save the latest filter entry if it is not already in the filter history
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc filtering::pushFilter {args} {
   if {$::TA(filter) != {} && [lsearch -exact $::snackAmpSettings(History) $::TA(filter)] < 0} {
      set ::snackAmpSettings(History) [linsert $::snackAmpSettings(History) 0 $::TA(filter)]
   }
}
#------------------------------------------------------------------------------#
# Function   : filtering::filterMusic
# Description: This is a common routine to filter the playlist, either by
#              text or matching against the auto-playlist settings (or both)
#              The end result is placed into the playlist, either replacing
#              or appending tracks already there.
# Author     : Tom Wilkason
# Date       : 7/1/2002
#------------------------------------------------------------------------------
proc filtering::filterMusic {how what} {
   Trace
   global TA snackAmpSettings
   ##
   # Filter on entire list or just the playlist
   #
   set useList [list]
   switch -- $what {
      "playList" {
         # Use only playlist
         set useList [LB::keyOrNamesAll 1]
         set MA 1
      }
      1 -
      "all" {
         # Use all files (last known list)
         set useList [catalog::popList]
         set MA 1
      }
      0 -
      "apl" {
         # use APL files (MA will force an APL check)
         set useList [catalog::aplFiles]
         set MA 0
      }
      def {
         # use Match against All flag
         set useList [catalog::popList]
         set MA $::TA(matchAgainstAll)
      }
   }
   ##
   # Where to put the filtered data
   #
   switch -- $how {
      "replace" {
         LB::deleteAll
         set select 0
         set from "end"
      }
      "insert" {
         set from [expr {[LB::idxActive]+1}]
         set select 0
      }
      "append" {
         set from "end"
         set select 0
      }
      "select" {
         set from "end"
         set select 1
      }
      default {
         set from "end"
         set select 0
      }
   }
   set insertList [list]
   ##
   # Support boolean searches with 'or' clause
   #
   if {[string length $::TA(filter)]==0} {
      set fast 1
      set regPattern ""
   } else {
      set fast 0
      set regPattern [list]                                ;#retain wildcards
      set pattern [split [string map {" or " "|" " OR " "|" "*" ".*" "?" "."} [catalog::globEscape $TA(filter)]] |]
      foreach part $pattern {
         lappend regPattern "($part)+"
      }
      set regPattern [join $regPattern |]
   }
   # Save off some variables for fast access
   set RF $snackAmpSettings(UseRemovableFolders)
   set rmtFilter [file join $::RMT *]
   ##
   # Now scan through each file in the list and filter it
   #
   if {[catch {
      set i -1
      foreach {File} $useList {
         incr i
         # Skip removable files if disabled
         #        || should this be or below?
         if {!$RF && [string match $rmtFilter $File]} {
            continue
         }
         # Pattern match (of if no pattern ignore regexp for speed)
         if {$fast || [regexp -nocase -- $regPattern $File]} {
            # Either create an insert list or just select the track
            if {$select} {
               LB::setSelection $i
            } else {
               lappend insertList $File
            }
         }
      }
   } result]} {
      saLog "filtering::filterMusic: '$regPattern' $result"
   }
   if {[llength $insertList] > 0} {
      LB::insertList $from $insertList
   }
   LB::tryRestorePlaying
}

#------------------------------------------------------------------------------
# Function   : toggleFilter
# Description: Toggle the filter button (checkbutton won't lie flat!)
# Author     : Tom Wilkason
# Date       : 11/11/2001
#------------------------------------------------------------------------------
proc filtering::toggleFilter {args} {
   if {$::TA(matchAgainstAll)} {
      $::TA(filterAllButton) configure -relief flat
      $::TA(filterAllButton) configure -image saImages::filterOff
      set ::TA(matchAgainstAll) 0
   } else {
      $::TA(filterAllButton) configure -relief sunken
      $::TA(filterAllButton) configure -image saImages::filterImage
      set ::TA(matchAgainstAll) 1
   }
}

#------------------------------------------------------------------------------
# Function   : filtering::createMasterList
# Description: Sort or shuffle the master playlist
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc filtering::createMasterList {args} {
   Trace
   global snackAmpSettings
   variable sortModes
   set index [lsearch -exact $sortModes $snackAmpSettings(sortModeTxt)]
   switch -- $index {
      0 {set tflist [db::sortedList]}
      1 {set tflist [shuffle [db::unSortedList]]}
      2 {set tflist [shuffleByFolder [db::sortedList] ldsort]}
      3 {set tflist [shuffleByFolder [db::sortedList] shuffle]}
      default {
         package require subSorting
         set tosort [subSort::getModes]
         set tflist [db::subSorting $tosort]
      }
   }
   # Need to save order for filtering (make this go away, a memory hog)
   catalog::pushList $tflist
   return $tflist
}
#------------------------------------------------------------------------------
# Function   : filtering::sortPlaylist
# Description: Sort or shuffle the playlist files
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc filtering::sortPlaylist {args} {
   Trace
   global TA snackAmpSettings
   soundControl::latency long
   Wait 1
   ##
   # Need to restore the current playlist, in the
   # order of the newly sorted master list
   #
   # Need to keep the master list in sync,
   # can we be more efficient about using it here instead of re-shuffling/sorting?
   # Have to handle playlist with multiple entries
   filtering::createMasterList
   update
   set playList [LB::keyOrNamesAll 1] ;# Get full file names in playlist
   if {[llength $playList] > 0} {
      LB::deleteAll
      # TBD: We can improve the speed a bit on tablelists by
      # returning the colums directly
      LB::insertList end [filtering::sortAList $playList]
   }
   LB::tryRestorePlaying
   gui::updateFileDisplay
   Wait 0
   soundControl::latency
}
#------------------------------------------------------------------------------
# Function   : filtering::sortAList
# Description: Sort a list (of files) based on the current sorting mode
#              Used by the playlist sorter
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc filtering::sortAList {Items} {
   variable sortModes
   global snackAmpSettings
   set index [lsearch -exact $sortModes $snackAmpSettings(sortModeTxt)]
   switch -- $index {
      0  {return [lsort -dictionary $Items]}
      1  {return [shuffle $Items]}
      2  {return [shuffleByFolder $Items ldsort]}
      3  {return [shuffleByFolder $Items shuffle]}
      default {
         package require subSorting
         set tosort [subSort::getModes]
         return [db::subSorting $tosort $Items]
      }
   }
}

package provide saFiltering 1.0

