
#------------------------------------------------------------------------------
# Define the procedures and continue processing at the end of the file
#------------------------------------------------------------------------------

namespace eval gui {
   variable tsScale
   variable volScale
   variable trackPosition 0
   variable songDuration 0  ;# this might be better in soundControl
}
#------------------------------------------------------------------------------
# Function   : buildGui
# Description: Build the player Gui
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc gui::buildGui {args} {
   variable tsScale
   variable volScale
   variable trackPosition
   global TA LB
   set ::songArray(Title) "SnackAmp Music Player"
   set TA(songTitle) "SnackAmp Music Player"
   set TA(songBalloon) "Track Details"

   ##
   # Hide the toplevel and move it off the visible screen
   #
   toplevel .saTop -class SnackAmp -cursor $::TA(arrow) -borderwidth 0
   set ::SA .saTop
   wm title $::SA "SnackAmp Music File Player"
   ##
   # clean up nicely
   #
   wm protocol $::SA WM_DELETE_WINDOW {taExit}
   wm protocol . WM_DELETE_WINDOW {taExit}
   wm withdraw $::SA
   # Init the volumn linkages

   gui::initVolume
   ##
   # Control Panel
   #
   pack [frame $::SA.panel -borderwidth 2 -relief groove] -side top -fill x -expand false -anchor w
   frame $::SA.panel.stat -relief sunken -borderwidth 2

   ##
   # Main area for music details
   #
   frame $::SA.panel.stat.l -background $::snackAmpSettings(bg) \
      -borderwidth 0 -width 70
   settings::regColor $::SA.panel.stat.l -background ::snackAmpSettings(bg)
   layoutPlayInfo $::snackAmpSettings(showDetail) ::songArray $::SA.panel.stat.l
   # Trace the showDetail settings to update the display detail immediatly from the settings menu
   trace variable ::snackAmpSettings(showDetail) w [list gui::reLayoutPlayInfo $::SA.panel.stat.l ::songArray]

   ##
   # Time counter
   #
   label $::SA.panel.stat.t -background $::snackAmpSettings(bg) -foreground $::snackAmpSettings(sfg) -borderwidth 0 \
   -font timeStatus -height 1 -textvar TA(timeStatus) -anchor e
   settings::regColor $::SA.panel.stat.t -background ::snackAmpSettings(bg) -foreground ::snackAmpSettings(sfg)
   balloon::define $::SA.panel.stat.t "Time Playing\nTime Remaining\nTotal Time"

   ##
   # Click in title edits ID3 tag
   #
   set TA(muteFactor) 1.0
   bind $::SA.panel.stat.l.dtTitle <ButtonPress-3> {EditPlaying}
   balloon::widgetDefine $::SA.panel.stat.l.dtTitle

   ##
   # Level Meter, on the left
   #
   frame $::SA.panel.stat.meter -background $::snackAmpSettings(bg) \
      -borderwidth 0
   settings::regColor $::SA.panel.stat.meter -background ::snackAmpSettings(bg)
   if {$::snackAmpSettings(levelMeter)} {
      after idle gui::showLevelMeter
   }

   # Place the labels next to each other
   grid $::SA.panel.stat.meter -column 0 -row 0 -rowspan 1 -sticky  news    ;# Meter
   grid $::SA.panel.stat.l -column 1 -row 0 -sticky nsew                ;# Song Title
   grid $::SA.panel.stat.t -column 2 -row 0 -rowspan 1 -sticky nsew     ;# Time Status

   grid $::SA.panel.stat -row 0 -column 1 -columnspan 2 -sticky nsew;# Inner Grid
   # Player controls
   #
   snack::createIcons
   grid [frame $::SA.panel.player] -row 1 -column 1 -sticky news

   ##
   # Button data
   #
   set buttonData {
      bitmap snackPlayPrev       Prev                           "Play previous in list"
      bitmap snackPlay           {set ::TA(STOP) 0; PlayResume} "Play Selection"
      bitmap snackPause          Pause                          "Pause/Resume playback"
      bitmap snackStop           {set ::TA(STOP) 1;Stop}        "Stop playing"
      bitmap snackPlayNext       {set ::TA(STOP) 0;Next}        "Play next in list"
      image  saImages::openFile  OpenFile                       "Add a file to play list and play it immediately"
      image  saImages::treeImage AutoPlaylist                   "Manage your music collection and define Auto-playlists"
   }
   # build each one
   set tf $::SA.panel.player
   foreach {bmType File Command Help} $buttonData {
      set iroot [namespace tail $File]
      button  $tf.$iroot -$bmType $File -borderwidth 1 -command $Command
      #bind $tf.$iroot <ButtonPress-1> $Command
      pack $tf.$iroot -side left -padx 0 -pady 0 -fill y
      balloon::button $tf.$iroot $Help
   }
   # Put this button before tree button
   pack [menubutton $tf.sortMenu -relief flat -image saImages::shuffleImage -menu $tf.sortMenu.drop] \
      -before $tf.$iroot -side left -fill both
   balloon::button $tf.sortMenu  "Select sort mode"
   set mb $tf.sortMenu.drop
   ;# Rating
   set pm $tf.rating.popup
   pack [menubutton $tf.rating -relief flat -image saImages::iconthumbsup -menu $pm] \
   -before $tf.$iroot -side left -fill both
   set m [menu $pm -tearoff 0 \
   -postcommand [list gui::ratingPopup $pm] -font menu]
   balloon::button $tf.rating "Adjust rating of the playing track, the more stars the better."
   # History menu, one for button and one for RMB
   ##
   # Sort menus
   #
   package require subSorting
   menu $mb -tearoff 0 -font menu -postcommand [list subSort::getAllModes $mb]
   set TA(treeButton) $tf.treeImage
   set TA(sortButton) $tf.sortImage
   set TA(shuffleButton) $tf.shuffleImage
   set TA(pauseButton) $tf.snackPause
   ##
   # Scale to show/control Offset within file
   #
   if {$::TA(showTicks)} {
      set ti 20
   } else {
      set ti 0
   }
   set tsScale [scale $::SA.panel.ss -show no -orient horiz -width 8 -len 200 -from 0. -to 60. \
      -var gui::trackPosition -tickinterval $ti -digits 1 -sliderlength 5 -resolution 0.5 \
      -highlightcolor $::snackAmpSettings(bg) \
      -troughcolor $::snackAmpSettings(bg) -command gui::updateTime] ;#-cursor sb_h_double_arrow
   settings::regColor $tsScale -troughcolor ::snackAmpSettings(bg) \
      -highlightcolor $::snackAmpSettings(bg)

   balloon::define $tsScale "Show/Adjust playback location within file"
   grid $tsScale -column 2 -row 1 -sticky ew

   # Allow time to be adjusted for all the buttons
   set TA(setdrag) 1
   bind $tsScale <ButtonPress-1> {set TA(setdrag) 0}
   bind $tsScale <ButtonRelease-1> {set TA(setdrag) 1; PlayOffset}
   bind $tsScale <ButtonPress-2> {set TA(setdrag) 0}
   bind $tsScale <ButtonRelease-2> {set TA(setdrag) 1; PlayOffset}
   bind $tsScale <ButtonPress-3> {set TA(setdrag) 0}
   bind $tsScale <ButtonRelease-3> {set TA(setdrag) 1; PlayOffset}
   ##
   # Scale show/adjust Sound level
   #
   set TA(gain) [macros::getPlayGain]

   set volScale [scale $::SA.panel.sv -show no -orient vert -width 10 -command {PlayGain} \
         -len 25 -var TA(gain) -from 64. -to 0. -sliderlength 15 \
         -highlightcolor $::snackAmpSettings(bg) \
         -troughcolor    $::snackAmpSettings(bg)]  ;#-cursor sb_v_double_arrow
   settings::regColor $volScale -troughcolor ::snackAmpSettings(bg) \
      -highlightcolor $::snackAmpSettings(bg)
   grid $volScale -column 0 -row 0 -rowspan 3 -sticky ns
   balloon::define $volScale "Adjust the playback level"
   ##
   # Mousewheel adjust volume from toplevel
   #
   bind $tsScale <Any-Leave> {
      focus [winfo toplevel %W]
   }
   bind $tsScale <Any-Enter> {
      focus %W
   }
   bind $volScale <Any-Enter> {
      set TA(pushFocus) [focus]
      focus %W
   }
   # Reconnect the volume
   bind $volScale <ButtonPress-1> {
      gui::initVolume
   }
   bind $volScale <Any-Leave> {
      if {[string equal %W $TA(pushFocus)]} {
         focus [winfo toplevel %W]
      } else {
         if {[catch {focus $TA(pushFocus)}]} {
            focus [winfo toplevel %W]
         }
      }
   }
   ##
   # Bind the mouse roller wheel on the scales
   #
   if {$::tcl_platform(platform) != "windows"} {
      bind $volScale <Button-5> {bumpLevel -2}
      bind $volScale <Button-4> {bumpLevel 2}
      bind $tsScale <Button-5> {macros::skipPlay -5}
      bind $tsScale <Button-4> {macros::skipPlay 5}
   } else {
      bind $tsScale <MouseWheel> {
         set amount [expr {%D/abs(%D)*5}] ;# get +/- 2
         macros::skipPlay $amount
         break
      }
      bind $volScale <MouseWheel> {
         set amount [expr {%D/abs(%D)*1}] ;# get +/- 2
         macros::bumpLevel $amount
         break
      }
   }

   # Adjust grid resize behavior and make things span properly
   grid columnconfigure $::SA.panel.stat 0 -weight 0
   grid columnconfigure $::SA.panel.stat 1 -weight 1            ;# For the label to stretch
   grid columnconfigure $::SA.panel 0 -weight 0
   grid columnconfigure $::SA.panel 1 -weight 0
   grid columnconfigure $::SA.panel 2 -weight 1

   # Collapsable frame for playlist
   set CF [cframe $::SA.hull "Tracks"]
   pack $::SA.hull -side top -fill both -expand true
   cframe.close $CF

   ##
   # Filter box
   pack [frame $CF.filter -borderwidth 2 -relief groove] -side top -expand no -fill x
   pack [button $CF.filter.label -text "Apply Filter" -pady 0 -padx 2 -font entry -command {
      filtering::pushFilter
      macros::filterMusic replace $::TA(matchAgainstAll)
   } -borderwidth 1] \
      -side left -expand no -fill both
   balloon::button $CF.filter.label "Replace files matching filter to current playlist"

   ##
   # Filter entry
   pack [entry $CF.filter.entry -textvariable TA(filter) -width 20] -side left -expand yes -fill x
   set ::FI $CF.filter.entry
   $::FI config -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)
   settings::regColor $::FI -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
   balloon::define $::FI "Enter text to restrict files to show in playlist, Buttons to left or right to apply filter. Double click to clear. You can use 'or' for boolean matches. Use * or ? for wildcards."
   ##
   # Dropdown button
   pack [menubutton $CF.filter.history -image saImages::dropdownImage -menu $CF.filter.history.list -relief flat] -side left
   balloon::button $CF.filter.history  "Options to retrieve prior filter entries or apply current entries against the current auto-playlist files (APL) or against all of the known files"
   # History menu, one for button and one for RMB
   menu $CF.filter.history.list -postcommand [list filtering::filterHistory $CF.filter.history.list]
   menu $::FI.popup -postcommand [list filtering::filterHistory $CF.filter.entry.popup]
   bind $::FI <ButtonPress-3> {
      tk_popup %W.popup %X %Y
   }

   ##
   # Filter against all or just auto playlist
   set TA(filterAllButton) $CF.filter.doall
   pack [button $CF.filter.doall -image saImages::filterOff\
      -command filtering::toggleFilter -borderwidth 1] \
      -side left -expand no -fill x
   balloon::button $CF.filter.doall "Push this to have filter applied to all files (by default), not just ones in the current auto-playlist"

   ##
   # LB::AppendFile/Replace buttons for filter
   pack [button $CF.filter.btApp -image saImages::appendImage -command {
      filtering::pushFilter
      macros::filterMusic append $::TA(matchAgainstAll)
   } -borderwidth 1] -side left
   balloon::button $CF.filter.btApp "Append files matching filter to current playlist"

   pack [button $CF.filter.btRep -image saImages::replaceImage -command {
      filtering::pushFilter
      macros::filterMusic replace $::TA(matchAgainstAll)
   } -borderwidth 1] -side left
   balloon::button $CF.filter.btRep "Replace files matching filter to current playlist"
   ##
   # Find button
   #
   pack [button $CF.filter.btFind -image saImages::findImage -command {
      findNextSong
   } -borderwidth 1] -side left
   balloon::button $CF.filter.btFind "Find the next song in the playlist that matches the Filter"

   ##
   # Scrolled Listbox
   #
   set LB [LB::create $CF.frame]
   ##
   # Default then User Key Assigments and menus
   #
   bindings::defaultHotKeys
   bindings::assignHotKeys
   gui::InstallMenu
   gui::InstallLBMenu $LB
   # The songlist packs last, so it is the one that contracts the most
   pack $CF.frame -expand yes -fill both -side top
   gui::fixMenuDetail $::snackAmpSettings(showDetail)

   ##
   # Callbacks to immediatly set the fonts when the user changes them
   #
   settings::installFontCallbacks
   ##
   # Create an icon if needed
   #
   if {$::tcl_platform(platform) == "windows"} {
      if {[catch {package require IconSupport} result]} {
         puts stderr $result
      } else {
         iconSupport
      }
   }
}
#------------------------------------------------------------------------------
# Initialize the mixer controls
#------------------------------------------------------------------------------
proc gui::initVolume {args} {
   foreach {line} [snack::mixer lines] {
      set ::TA(line,$line) [snack::mixer volume $line]
      if {[snack::mixer channels $line] == "Mono"} {
         snack::mixer volume $line ::TA(line,$line)
      } else {
         snack::mixer volume $line ::TA(line,$line) ::TA(line,$line)
      }
   }
}
#------------------------------------------------------------------------------
# Function   : setTime
# Description: Set the scale timer to some value
# Author     : Tom Wilkason
# Date       : 8/8/2003
#------------------------------------------------------------------------------
proc gui::setTime {val} {
   variable tsScale
   $tsScale set $val
}
#------------------------------------------------------------------------------
# Function   : InstallMenu
# Description: Build the menu bar at the top of the GUI
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc gui::InstallMenu {} {
   global TA LB snackAmpSettings tcl_platform
   bindings::init $::SA menuBindings
   set M $::SA.bRmenubar
   # Make menus as small as possible
   menu $M \
      -font menu \
      -activebackground [$::SA cget -background] \
      -activeforeground [$::SA cget -highlightcolor] \
      -activeborderwidth 1
   # attach it to the main window
   $::SA config -menu $M
   # Create more cascade menus
   foreach {m u} {File 0 Edit 0 Playlists 0 AutoPlaylists 0 Settings 0 Help 0} {
      switch -- $m {
         Playlists {
            set $m [menu $M.m$m -tearoff 0 -postcommand [list gui::playListFiles $M.m$m] -font menu]
            set TA(plPopUp) $M.m$m
            bindings::addCommand $M.m$m "Clear" {Clear}
            bindings::addCommand $M.m$m "Open" {macros::ReadPlaylistFile {} replace} 0
            bindings::addCommand $M.m$m "Save all" {macros::SavePlaylistFile $::LB replace all $::TA(PlayListFile)} 1
            bindings::addCommand $M.m$m "Save all as" {macros::SavePlaylistFile $::LB replace all}
            bindings::addCommand $M.m$m "Save selected as" {macros::SavePlaylistFile $::LB replace selected}
            bindings::addCommand $M.m$m "Append all as" {macros::SavePlaylistFile $::LB append all}
            $M.m$m add separator
            $M.m$m add cascade -label "Save selected to..." -menu [menu $M.m$m.uReplace -tearoff 0 -font menu]
            $M.m$m add cascade -label "Append selected to..." -menu [menu $M.m$m.uAppend -tearoff 1 -font menu]
            $M.m$m add separator
            $M.m$m add cascade -label "Append from..." -underline 0 -menu [menu $M.m$m.mAppend -tearoff 0 -font menu]
            $M.m$m add cascade -label "Insert from..." -underline 0 -menu [menu $M.m$m.mInsert -tearoff 0 -font menu]
            $M.m$m add cascade -label "Replace from..." -underline 0 -menu [menu $M.m$m.mReplace -tearoff 0 -font menu]
            $M.m$m add separator
            set plrem [menu $M.m$m.remove -tearoff 0 -postcommand "[list gui::PLFilesRemove $M.m$m.remove]"  -font menu]
            $M.m$m add cascade -label "Remove..." -menu $plrem
         }
         AutoPlaylists {
            set TA(FolderMenu) $m
            set $m [menu $M.m$m -tearoff 1 -postcommand [list gui::APLFiles $M.m$m] -font menu]
            set Removal [menu $M.m$m.remove -tearoff 0 -postcommand "[list gui::APLFilesRemove $M.m$m.remove]"  -font menu]
         }
         default {
            set $m [menu $M.m$m -tearoff 0 -font menu]
         }
      }
      $M add cascade -label $m -menu $M.m$m -underline $u
   }
   ##
   # Here is how to add a menu to the "system" menu
   # Use it to put "pinned" command
   #
   ontop::pinOnSystemMenu $M $::SA
   # FILE

   bindings::addCommand $File "Open File" {OpenFile} 0
   bindings::addCommand $File "Open URL" {OpenURL}
   bindings::addCommand $File "Save Track As ..." {catalog::saveAs}
   $File add separator
   bindings::addCommand $File "Manage Media Library..."  {catalog::manageMediaFolders} 0
   bindings::addCommand $File "Index Media Library" {catalog::IndexAskFirst} 0
   $File add separator
IFDEF RMS {
   bindings::addCommand $File "Manage Removable Media..." {
      package require saSettings
      folderManager::launch RemovableFolders
   }
   bindings::addCommand $File "Catalog Removable Media" catalog::CatalogMedia 1
   bindings::addCommand $File "Mount Removable Media"  {catalog::MountMedia "Removable"}
   $File add separator
}
   bindings::addCommand $File "Manage Media Folders to Monitor for Changes..." {
      package require saSettings
      folderManager::launch IncomingDirs
   }
   bindings::addCommand $File "Create Music Listing (csv files)" {catalog::documentCache}
   $File add separator
   bindings::addCommand $File "Track Info..." {catalog::properties [list [soundControl::fileName]]}
   $File add separator
   bindings::addCommand $File "Exit" {taExit} 1

   # EDIT
   bindings::addCommand $Edit "Select All" {LB::setSelection 0 end;focus $LB} 0
   $Edit add separator
   bindings::addCommand $Edit "Sync APL to Selected Track" {SelectFile 0}
   bindings::addCommand $Edit "Sync APL to Playing Track" {SelectFile 1}
   bindings::addCommand $Edit "Locate Playing Track" {locatePlaying}
   $Edit add separator
   bindings::addCommand $Edit "Edit ID3 for Selected Track" {EditPLID3}
   bindings::addCommand $Edit "Edit ID3 for Playing Track" {EditPlaying} 0
   $Edit add separator
   bindings::addCommand $Edit "Rename Selected Track" {RenameFileDialog [LB::nameActive]}
   bindings::addCommand $Edit "Rename Playing Track" {RenameFileDialog [soundControl::fileName]} 0
   $Edit add separator
   # Need to figure out auto-binding for events like this
   bindings::addCommand $Edit "Copy Selected to Clipboard" {clipboardCopy $::LB 0} 0
   bindings::addCommand $Edit "Cut Selected to Clipboard" {clipboardCut $::LB 0} 1
   bindings::addCommand $Edit "Paste Clipboard" {clipboardPaste $::LB 1} 0
   $Edit add separator
#   $Edit add command -label "Reset Playing Track Auto Gain"  -command {
#      gainControl::resetSome [list [soundControl::fileName]]
#      snack::audio scaling 1.0
#   }
   bindings::addCommand $Edit "Delete Selection" {Delete}


   bindings::addCommand $AutoPlaylists "Manage Auto-Playlists" {AutoPlaylist} 0
#   menu $AutoPlaylists.subsort -font menu  -tearoff 0
   # Below is also deleted by callback due to tearoff
#   gui::aplSubsortMenu $AutoPlaylists
   $AutoPlaylists add cascade -label "Remove..." -menu $Removal -underline 0
   $AutoPlaylists add separator

   # SETTINGS
   if {$snackAmpSettings(enableConsole)} {
      $Settings add check -label "Show Console" -variable TA(Console) -underline 5 -command {
         if {[info command console] == {}} {
            catch {package require saConsole}
         }
         Console
      }
   }
   $Settings add check -label "Show Mini Player" -underline 5 -variable ::snackAmpSettings(activeMiniPlayer) -command {
      package require miniPlayer
      miniPlayer::toggle
   }
#   if {$::tcl_platform(platform) == "unix"} {
      bindings::addCommand $Settings "Show Mixer" {snack::mixerDialog} 2
#   }
   $Settings add separator
   if {[llength [snack::mixer lines]] > 1} {
      $Settings add cascade -label "Volume Controls..." -underline 0 -menu [menu $Settings.volume -tearoff 0 -font menu]
         foreach {line} [snack::mixer lines] {
            $Settings.volume add radiobutton -label $line -variable snackAmpSettings(lineSelect) -value $line
         }
   }
   # Force select the first (default) output device, fixes bug
   # on Linux where write errors occur randomly
   if {! [info exists ::snackAmpSettings(outputDevice)]} {
      set snackAmpSettings(outputDevice) [lindex [snack::audio outputDevices] 0]
   }
   snack::audio selectOutput $::snackAmpSettings(outputDevice)
   $Settings add cascade -label "Output Device..." -underline 0 -menu [menu $Settings.output -tearoff 0 -font menu]
      foreach {device} [snack::audio outputDevices] {
         $Settings.output add radiobutton -label $device -variable snackAmpSettings(outputDevice) -value $device \
            -command [list snack::audio selectOutput $::snackAmpSettings(outputDevice)]
      }
   $Settings add cascade -label "Minimum Bitrate..." -menu [menu $Settings.bitrate -tearoff 0 -font menu]
      set ::TA(minRate) 0
      foreach {rate} [list 0 24 64 96 128 140 160 180 192 256 320] {
         $Settings.bitrate add radiobutton -label $rate -variable ::TA(minRate) -value $rate
      }
   $Settings add cascade -label "Minimum Rating..." -menu [menu $Settings.rating -tearoff 0 -font menu]
      $Settings.rating add radiobutton -label "--" -variable ::snackAmpSettings(minTxtRating) -value "--"
      foreach {rate} [list 1 2 3 4 5] {
         set txt [string repeat * $rate]
         $Settings.rating add radiobutton -label $txt -variable ::snackAmpSettings(minTxtRating) -value $txt
      }
   if {[string length [info command ntSetPriority]] > 0} {
      $Settings add cascade -label "Runtime Priority..." -menu [menu $Settings.priority -tearoff 0 -font menu]
      ntSetPriority $::snackAmpSettings(priority)
      foreach {priority} [list high normal low] {
         $Settings.priority add radiobutton -label $priority \
            -variable ::snackAmpSettings(priority) -value $priority \
            -command {after 0 ntSetPriority $::snackAmpSettings(priority)}
      }
   }
   $Settings add separator
   #$Settings add check -label "Continuous Play" -variable snackAmpSettings(ContinuousPlay)
   #$Settings add check -label "Shuffle Play" -variable snackAmpSettings(randomPlay) \
   #   -command filtering::sortPlaylist
   #$Settings add separator
   bindings::addCommand $Settings "Manage Media Library..." {catalog::manageMediaFolders}
   bindings::addCommand $Settings "Manage Removable Media..." {
      package require saSettings
      folderManager::launch RemovableFolders
   }
   bindings::addCommand $Settings "Manage Folders to Monitor for Changes..." {
      package require saSettings
      folderManager::launch IncomingDirs
   }
   $Settings add separator
   bindings::addCommand $Settings "Other Settings ..." {package require saSettings;settings::tabbedSettingsGui} 6

   # HELP
   if {$::snackAmpSettings(EnableWebServer)==1} {
      bindings::addCommand $Help "User Manual" {::browser::help} 0
      bindings::addCommand $Help "Internal Web Server Home" {::browser::urlOpen $snackAmpSettings(internalWebBase)/} 0
   } else {
      bindings::addCommand $Help "User Manual" {::browser::urlOpen "http://snackamp.sourceforge.net/user_manual/index.html"} 0
   }
   bindings::addCommand $Help "SnackAmp Home Page at SourceForge" {::browser::urlOpen "http://snackamp.sourceforge.net/"} 0
   $Help add separator
   bindings::addCommand $Help "Versions"  gui::cvsVersions
   bindings::addCommand $Help "About"  gui::about
}
#------------------------------------------------------------------------------
# Function   : InstallLBMenu
# Description: Install Listbox Menus and hard coded bindings
# Author     : Tom Wilkason
# Date       : 6/18/2002
#------------------------------------------------------------------------------
proc gui::InstallLBMenu {LB} {
   set m $LB.popup
   bindings::init $::SA menuBindings ;# Top Level Keys affect the playlist
   bindings::init $LB menuBindings   ;# Playlist unique keys override
   menu $m -tearoff 0 -font menu
   bindings::addCommand $m "Play Selection" {Play}
   $m add separator
   bindings::addCommand $m "Copy" {clipboardCopy $::LB 0}
   bindings::addCommand $m "Cut" {clipboardCut $::LB 0}
   bindings::addCommand $m "Paste" {clipboardPaste $::LB 1}
   bindings::addCommand $m "Stage" {linkToStaging $::LB 1}
   $m add separator
   bindings::addCommand $m "Sort Selection" {sortSelection}
   bindings::addCommand $m "Shuffle Selection" {shuffleSelection}
   $m add separator
   bindings::addCommand $m "Remove Selection" {Cut 1}
   bindings::addCommand $m "Clear Entire List" {LB::deleteAll;gui::updateFileDisplay}
   $m add separator
   bindings::addCommand $m "Delete File(s)" {Delete}
   bindings::addCommand $m "Rename File" {RenameFile [LB::nameActive]}
   $m add separator
   bindings::addCommand $m "Edit ID3 Info" {EditPLID3}
   bindings::addCommand $m "ID3 Table Editor" {PLNameToTag 0}
   bindings::addCommand $m "ID3 Tag To Name" {PLTagToName}
   $m add separator
   bindings::addCommand $m "Reset Auto Level" {gainControl::resetSome}
   menu $m.rating -tearoff 1 -font menu -title "Playlist Ratings"
   foreach {n} $db::ratingOptions {
      $m.rating add command -label $n \
      -command [list LB::setSelectedRating $n]
   }
   $m add separator
   $m add cascade -label "Track(s) Rating" -menu $m.rating
   $m add separator
   $m add cascade -label "Playlist Tracks" -menu $::TA(plPopUp)
   $m add separator
   bindings::addCommand $m "Track Info..." {catalog::properties [LB::nameSelected 1]}

   bind [LB::body] <ButtonPress-3> {
      tk_popup $LB.popup %X %Y
   }
   # Bindings
   balloon::plDefine
   bind [LB::body] <Double-ButtonPress-1> {
      set here [$LB nearest [LB::Y %y]]
      LB::activate $here
      Play active
   }

   bind [LB::body] <B1-Motion> {
      LB::Drag [LB::Y %y]
   }
   bind [LB::body] <ButtonPress-1> {
      focus %W
      LB::Select [LB::Y %y]
   }
}
#------------------------------------------------------------------------------
# Function   : updateTime
# Description: When Slider is updated, update time display, this is called
#              automatically when the time slider is updated during play.
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc gui::updateTime {args} {
   global TA
   variable trackPosition
   set elapsed   [expr {round($trackPosition)}]
   set remaining [expr {round($gui::songDuration - $elapsed)}]
   switch -- $::snackAmpSettings(showDetail) {
      1 {
         if {$gui::songDuration < 3600} {
            set TA(timeStatus) "[clock format $elapsed -format %M:%S]/[clock format $remaining -format %M:%S]"
         } else {
            set TA(timeStatus) "[clock format $elapsed -format %T -gmt 1]/[clock format $remaining -format %T -gmt 1]"
         }
      }
      2 {
         if {$gui::songDuration < 3600} {
            set TA(timeStatus) "[clock format $elapsed -format %M:%S]\n[clock format $remaining -format %M:%S]"
         } else {
            set TA(timeStatus) "[clock format $elapsed -format %T -gmt 1]\n[clock format $remaining -format %T -gmt 1]"
         }
      }
      default {
         if {$gui::songDuration < 3600} {
            set TA(timeStatus) "[clock format $elapsed -format %M:%S]\n[clock format $remaining -format %M:%S]\n[clock format [expr {round($gui::songDuration)}] -format %M:%S]"
         } else {
            set TA(timeStatus) "[clock format $elapsed -format %T -gmt 1]\n[clock format $remaining -format %T -gmt 1]\n[clock format [expr {round($gui::songDuration)}] -format %T -gmt 1]"
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : showLevelMeter
# Description: Insert the level meter at the appropriate height
# Author     : Tom Wilkason
# Date       : 3/6/2003
#------------------------------------------------------------------------------
proc gui::showLevelMeter {args} {
   update
   set height [winfo reqheight $::SA.panel.stat]
   levelmeter::create $::SA.panel.stat.meter.lm -width 8 \
      -length $height -orient vertical -oncolor lightblue
   pack $::SA.panel.stat.meter.lm -side bottom -fill both -expand true
   after 1000 [list levelmeter::cmUpdate]
}

#------------------------------------------------------------------------------
# Function   : reLayoutPlayInfo
# Description: On a write to the showDetail variable, update the amount
#              of displayed info on the main gui
# Author     : Tom Wilkason
# Date       : 12/18/2003
#------------------------------------------------------------------------------
proc gui::reLayoutPlayInfo {hull sArray _G I mode} {
   upvar #0 $_G G
   set level $G($I)
   foreach child [winfo children $hull] {
      destroy $child
   }
   gui::layoutPlayInfo $level $sArray $hull
   gui::fixMenuDetail $level
   updateTime
   wm geometry .saTop {}
}
##
# Place the top level menu in the appropriate place
#
proc gui::fixMenuDetail {level} {
   ##
   # Move menu bar to pop-up for very small layout
   #
   set tlm .saTop.bRmenubar ;# hack
   set tm $::SA.panel.stat.l.dtTitle ;# hack
   if {$level > 1} {
      if {[string length [$::SA cget -menu]]==0} {
         $::SA configure -menu $tlm
      }
      bind $tm <ButtonPress-3> {
         EditPlaying
      }
   } else {
      if {[string length [$::SA cget -menu]]!=0} {
         $::SA config -menu {}
      }
      bind $tm <ButtonPress-3> {
         tk_popup .saTop.bRmenubar %X %Y
      }
   }
}
#------------------------------------------------------------------------------
# Function   : gui::layoutPlayInfo
# Description: Layout the contents of the main window
# Author     : Tom Wilkason
# Date       : 1/12/2002
#------------------------------------------------------------------------------
proc gui::layoutPlayInfo {threshold _Var hull} {
   # Label Variable Row Col ColSpan Font
   set color(songLabel) tfg
   set color(songInfo) fg
   # Layout the data
   # 6 columns
   set gridData {
      1 Title  Title  0 5 songLabel
      2 Artist Artist 1 5 songInfo
      2 Album  Album  2 1 songInfo
      2 Rate   Rate   2 1 songInfo
      3 Track  Track  3 1 songInfo
      3 Year   Year   3 1 songInfo
      3 Genre  Genre  3 1 songInfo
      4 Date   fDate  4 1 songInfo
      4 Size   fSize  4 1 songInfo
      4 Freq   sampRate 4 1 songInfo
   }
   set uCol 0
   set uRow 0
   ##
   # There is a tk bug when a full col is skipped
   #
   foreach {Use Title Field sRow cSpan font} $gridData {
      # label
      if {$threshold >= $Use} {
         # Special handling to work around tk col bug
         if {$Title=="Album" && $threshold>= 3} {
            incr cSpan 2
         }
         if {$sRow > $uRow} {
            set uCol 0
            set uRow $sRow
         }
         #W: Todo, link the label to a variable which can be cleared
         # if there is no value.
         label $hull.lb$Title -text $Title  -font smallLabel -pady 0 \
            -background $::snackAmpSettings(bg) \
            -foreground $::snackAmpSettings(lfg) \
            -anchor e
         settings::regColor $hull.lb$Title \
            -background ::snackAmpSettings(bg) \
            -foreground ::snackAmpSettings(lfg)
         label $hull.dt$Title -textvariable ${_Var}($Field) -font $font -pady 0  -bd 0 \
            -background $::snackAmpSettings(bg) \
            -foreground $::snackAmpSettings($color($font)) \
            -anchor w -bd 0
         settings::regColor $hull.dt$Title \
            -background ::snackAmpSettings(bg) \
            -foreground ::snackAmpSettings($color($font))
         # If only title, then don't bother with label
         # No label on minimal threshold
         if {$threshold > 1} {
            grid $hull.lb$Title -row $sRow -column $uCol -sticky e -pady 0
         }
         grid $hull.dt$Title -row $sRow -column [incr uCol] -columnspan $cSpan -sticky w -pady 0
         incr uCol $cSpan
      }
   }
   grid columnconfigure $hull 1 -weight 1
}
#------------------------------------------------------------------------------
# Function   : gui::playListFiles
# Description: postcommand callback to show play list files on menu, called
#              from play list menu postcommand
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc gui::playListFiles {menu} {
   global TA LB
   set Files [catalog::getPlaylists]
   if {[llength $Files] == 0} {
      return
   }
   $menu.mAppend delete 0 end
   $menu.mReplace delete 0 end
   $menu.mInsert delete 0 end
   $menu.uAppend delete 0 end
   $menu.uReplace delete 0 end
   ##
   # Need to sort the list by decending date
   #
   foreach {File} $Files {
      set Value [file rootname [file tail $File]]
      $menu.mAppend add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::ReadPlaylistFile $File "append"]
      $menu.mInsert add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::ReadPlaylistFile $File "insert"]
      $menu.mReplace add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::ReadPlaylistFile $File "replace"]

      $menu.uAppend add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::SavePlaylistFile $::LB "append" "selected" $File]
      $menu.uReplace add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::SavePlaylistFile $::LB "replace" "selected" $File]
   }
}
#------------------------------------------------------------------------------
# Popup menu to adjust curent file stuff
#------------------------------------------------------------------------------
proc gui::ratingPopup {menu} {
   global TA
   variable trackRating
   $menu delete 0 end
   set me $TA(currentMF)
   set rating [db::getValues $me Rating]
   set trackRating $rating
   foreach {n} $db::ratingOptions {
      $menu add radiobutton -label $n -value $n \
      -variable gui::trackRating \
      -command [list db::setVal $me Rating $n]
   }
}
#------------------------------------------------------------------------------
# Function   : gui::APLFiles
# Description: post command Callback to show auto playlist setting files, called
#              from auto-play list menu postcommand
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc gui::APLFiles {menu} {
   global TA LB
   set Files [catalog::getAutoPlaylists]
   if {[llength $Files] == 0} {
      return
   }
   $menu delete 3 end
   $menu add separator
   foreach {File} $Files {
      set Value [file rootname [file tail $File]]
      $menu add radiobutton -label $Value -value $Value \
         -variable TA(AutoPlayListSelection) -command [list readAPLFile $File 1]
   }
}
#------------------------------------------------------------------------------
# Function   : gui::APLFilesRemove
# Description: post command Callback to allow removal of auto playlist setting files
#              from auto-play list menu postcommand
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc gui::APLFilesRemove {menu} {
   global TA LB
   set Files [catalog::getAutoPlaylists]
   if {[llength $Files] == 0} {
      return
   }
   $menu delete 0 end
   foreach {File} $Files {
      set Value [file rootname [file tail $File]]
      $menu add radiobutton -label $Value -value $Value \
         -command [list removeAPLFile $File]
   }
}

#------------------------------------------------------------------------------
# Function   : gui::PLFilesRemove
# Description: post command Callback to allow removal of auto playlist setting files
#              from auto-play list menu postcommand
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc gui::PLFilesRemove {menu} {
   global TA LB
   set Files [catalog::getPlaylists]
   if {[llength $Files]==0} {
      return
   }
   $menu delete 0 end
   foreach {File} $Files {
      set Value [file rootname [file tail $File]]
      $menu add radiobutton -label $Value -value $Value \
         -command [list removePLFile $File]
   }
}
#------------------------------------------------------------------------------
# Function   : about
# Description: About dialog box
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc gui::snackVersion {args} {
   set results "Snack engine version $::snack::patchLevel"
   if {[catch {package present snackogg} result]} {
      append results "\n   Snack ogg library not present"
   } else {
      append results "\n   Snack ogg version $result"
   }
   return $results
}
proc gui::about {args} {
   set Message {
   SnackAmp Music Player Version $::SAVersion
   [gui::snackVersion]
   Tcl/Tk Version [set ::tcl_patchLevel]
   Snackamp user folder located at '$::HOME'
   Using $::TA(database) database
   Current Database Statistics
[db::statsText "   "]

   Features:
   - Plays wav, mp3 and ogg files and http streams
   - Complete playlist control
   - Media manager including auto-playlists
   - Edit ID3 tags, including smart-renaming and smart-tagging.
   - Integrated web server for remote control and serving streaming music

    (c)Tom Wilkason - enjoy
       tom.wilkason@cox.net
       http://snackamp.sourceforge.net/
   }
   clipboard clear
   clipboard append [subst $Message]
   tk_messageBox -title "SnackAmp Information" -icon info -message [subst $Message]
}
##
# Dialog showing the CVS Headers
#
proc gui::cvsVersions {args} {
   append vinfo "CVS Version Info:\n"
   foreach {_module} [lsort -dictionary [info globals *_History]] {
      global $_module
      upvar $_module module
      foreach {xx file rest} [split $module ":,"] {break}
      foreach {v ver date time user xx yy} $rest {break}
      append vinfo [format "%-5s\t%-10s\t%-60s" $ver $date [file tail $file]]\n
   }
   clipboard clear
   clipboard append $vinfo
   tk_messageBox -title "OK to Continue" -icon info -message $vinfo
}
##
# Update the number of files in the playlist
# Note: Fix when we commonize sortMode
proc gui::updateFileDisplay {} {
   Trace
   global snackAmpSettings LB TA
   array set lookup {
      0  "Sorted"
      1  "Shuffled"
      2  "Shuffled Folders, Sorted Tracks"
      3  "Shuffled Folders, Shuffled Tracks"
   }
#   if {$snackAmpSettings(sortModeTxt) < 4} {
#      set mode "[commify [LB::idxEnd]] Tracks ($lookup($snackAmpSettings(sortModeTxt))) [LB::totTime]"
#   } else {
      set mode "[commify [LB::idxEnd]] Tracks ($snackAmpSettings(sortModeTxt)) [LB::totTime]"
#   }
   $::SA.hull.label configure -text $mode
}
package provide saGUI 1.0
