#------------------------------------------------------------------------------
# snackAmp - urlRedirect package 1.0 for snackAmp
# provides dynamic http applications for SnackAmp
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
#
# To create a new dynamic url create a namespace with the name of the url
# e.g. namespace eval playlist -> /playlist
# to enable it, create an init routine that returns a 1 (for security)
# you can have the url return either an existing file (e.g. an mp3 file), in
# that case create a namespace command ::process, that returns the name
# of the file.
#
# If you want to create dynamic html with the namespace, create a namespace
# command ::process, that returns content. You will need an extra parameter
# passed to it that you set to the file extension required.
#
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/urlRedirect.tcl,v 1.14 2004/02/11 00:55:48 wilkason Exp $
}
#------------------------------------------------------------------------------
#                              j u s t c o n t r o l
#------------------------------------------------------------------------------

namespace eval remote/justcontrol {
   proc init { } {
     return 1
   }
}
proc remote/justcontrol::process {sock {params ""}} {
   Trace
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         bt_Level {
            # Would be nice not to have to push data back
            # maybe hold to end to push it out if needed.
            if {[catch {::macros::bumpLevel $Cmd} result] } then {
            }
         }
         default {}
      }
   }
   return [list dhtml html 0]
}

##
# Wrapper for jukebox function, makes the browser namespace react based
# on calling function
#
#------------------------------------------------------------------------------
#                              j u k e b o x
#------------------------------------------------------------------------------
namespace eval jukebox/display {
   proc init { } {
     return 1
   }
}
#------------------------------------------------------------------------------
# Function   : display
# Description: Main function that is called by web server when /display
#              is entered.
#              e.g. /display
# Author     : Tom Wilkason
# Date       : 12/15/2002
#------------------------------------------------------------------------------
proc jukebox/display::process {sock {params ""}} {
   Trace
   append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [HTH::getRefreshScript [HTH::nextRefresh] "/jukebox/display"]
   # Wait to let player take hold
   #append Result ${search::close_window}
   append Result [HTH::playStatus 0]
   append Result [HTH::trackInfo [soundControl::fileName]]
   append Result [HTH::nextToPlay 2 $::snackAmpSettings(httpNextLength)]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}

##
# Jukebox current playing
#
namespace eval jukebox/header {
   proc init { } {
     return 1
   }
   proc [namespace current]::process {sock {params ""}} {
      Trace
      append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
      append Result "<body>\n"
      append Result [HTH::getRefreshScript [HTH::nextRefresh] "/jukebox/header"]
      append result [H2 "Now [string tolower [soundControl::playState]]:"]
      append Result [HTH::trackInfo [soundControl::fileName] 1]
      append Result "</body>\n</html>"
      return [list dhtml html $Result]
   }
}
##
# Jukebox next
#
namespace eval jukebox/next {
   proc init { } {
     return 1
   }
   proc [namespace current]::process {sock {params ""}} {
      Trace
      append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
      append Result "<body>\n"
      append Result [HTH::getRefreshScript [HTH::nextRefresh] "/jukebox/next"]
      # Need to launch hypertext to browse into a different target, how?
      append Result [HTH::nextToPlay 2 10]
      append Result "</body>\n</html>"
      return [list dhtml html $Result]
   }
}
##
# Jukebox next
#
namespace eval jukebox/last {
   proc init { } {
     return 1
   }
   proc [namespace current]::process {sock {params ""}} {
      Trace
      append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
      append Result "<body>\n"
      append Result [HTH::getRefreshScript [HTH::nextRefresh] "/jukebox/last"]
      append Result [HTH::queuedToPlay 2 20]
      append Result "</body>\n</html>"
      return [list dhtml html $Result]
   }
}
##
# Jukebox folder browser (for running in an object)
#
namespace eval jukebox/folder {
   proc init { } {
     return 1
   }
   proc [namespace current]::process {sock {params ""}} {
      Trace
      set control 2
      append result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
      append result "<body>\n"
      if {[string length $params]} {
         foreach param [split $params &]  {
            foreach {CmdType Cmd} [split $param =] {break}
            set dCmd [HttpD::Decode $Cmd]
            switch -- $CmdType {
               folder {
                  append result [H2 "Select a folder and pick your favorite track"]
                  append result [browser::folderList $Cmd $control $params]
               }
               play {
                  # Play a specific song
                  # /control?play=some+name.mp3
                  set track $dCmd
                  if {[catch {LB::InsertPlay $track} error] } then {
                     debug $result
                     append result $result
                  } else {
                     append result [H2 "Playing '$track'"]
                  }
               }
               append {
                  set track $dCmd
                  if {[HTH::isMedia $track]} {
                     set next [LB::idxFile $track]
                     if {$control==1 || $next==-1 || $next <= [LB::idxPlaying] || ($next-[LB::idxActive])>$::snackAmpSettings(minNextTrack)} {
                        # Insert after last web insert (instead of end) if possible
                        LB::userInsert $track
                        append result [H2 "Queue $track to Playlist"]
                        gui::updateFileDisplay
                     } else {
                        append result [H2 "Skipping $track, it is already in playlist"]
                     }
                  } else {
                     append result [H2 "Skipping $track, it is invalid"]
                  }
               }
            }
         }
      } else {
         append result [H2 "Select a folder and pick your favorite track"]
         append result [browser::folderList [info hostname] $control $params]
      }
      append result "</body>\n</html>\n"
      return [list dhtml html [string trim $result]]
   }
}
namespace eval jukebox/browser {
   proc init { } {
     return 1
   }
   proc process {sock {params ""}} {
      dirBrowser::doit $sock $params
   }
}
namespace eval jukebox/search {
   proc init { } {
     return 1
   }
   proc process {sock {params ""}} {
      search::doit $sock $params
   }
}
#------------------------------------------------------------------------------
#                               c o n t r o l
#------------------------------------------------------------------------------
##
# control namespace to control the SnackAmp player from
# a web browser. When user enters /control ... in to the browser
# The process function of this namespace will be called passing any
# user parameters. To add a new command then create another namespace
# and provide an init and process function.
#
##
# Wrapper for remote/search function, makes the search namespace react based
# on calling function
#
namespace eval remote/search {
   proc init { } {
     return 1
   }
   proc process {sock {params ""}} {
      search::doit $sock $params
   }
}
namespace eval remote/browser {
   proc init { } {
     return 1
   }
   proc process {sock {params ""}} {
      dirBrowser::doit $sock $params
   }
}
namespace eval remote/control {
   variable Details
     # <meta http-equiv="Content-Type" content="text/html; charset=windows-1252">
    #  <meta http-equiv="pragma"           content="no-cache">

   variable head {
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
   <head>
      <meta http-equiv="Content-Language" content="en-us">
      <meta name="GENERATOR"              content="SnackAmp Web Server">
      <meta name="ProgId"                 content="SnackAmp.Document">
      <LINK REL=StyleSheet                HREF="$css" TYPE="text/css" MEDIA=screen>
      <title>$title</title>
   }
   proc init { } {
     return 1
   }
}
#------------------------------------------------------------------------------
# Function   : remote/control::process state machine
# Description: Main function that is called by web server when /remote/control
#              is entered. Inside is a loop that iterates over each parameter
#              on the POST/GET and handles each one.
#              e.g. /remote/control?a=2&b=c  <= two params a= and b=
# Author     : Tom Wilkason
# Date       : 4/24/2002
#------------------------------------------------------------------------------
proc remote/control::process {sock {params ""}} {
   Trace
   ##
   # Replace puts with appending to string then doing as a batch if needed
   #
   append Result [HTH::leader "SnackAmp Control" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [js::scripts popupInfo]
   append Result [HTH::getRefreshScript [HTH::nextRefresh] "/remote/control?Status=Refresh+Data"]
   #append Result "$params"
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         bt_Control {
            if {[catch {::macros::$Cmd} result] } then {
               append Result "<p>$::errorInfo</p>\n"
            }
            #after 50 set ::remote/control::flag 1
            #vwait ::remote/control::flag
         }
         bt_Level {
            if {[catch {::macros::bumpLevel $Cmd} result] } then {
               append Result "<p>$::errorInfo</p>\n"
            }
         }
         play {
            # Play a specific song
            # /remote/control?play=some+name.mp3
            if {[catch {playSpecificSong $Cmd} error] } then {
               debug $result
               append Result $result
            }
            #after 50 set ::remote/control::flag 1
            #vwait ::remote/control::flag
         }
         delete {
            LB::deleteIndex $Cmd
            gui::updateFileDisplay
            # Remove from playlist
         }
         cmb_APL {
            # Select an auto-playlist file
            # TFW: If a real + is in file name, this may fail
            after 200 [list macros::readAPLFile [file join $::HOME $Cmd.tafs] 1]
         }
         cmb_Sort {
            # Sort/Shuffle the files
            set ::snackAmpSettings(sortModeTxt) $Cmd
            set subSort::sortSelection $::snackAmpSettings(sortModeTxt)
            # Hack, if a file then a custom sourt must be read in
            set fname [file join $::HOME $Cmd.ss]
            if {[file exists $fname]} {
               subSort::readFile $Cmd
            }
            after 200 filtering::sortPlaylist
         }
         filter_text {
            # stage 1
            set ::TA(filter) $Cmd
            # Note: These will happen in this order, search_replace can process
            # Button=Filter&filter_text=ab&filter_what=All+Files&search_replace=Replace
         }
         filter_what {
            # stage 2
            switch -glob -- [string tolower $Cmd] {
               "all*files" {set Search_What 1}
               "play*" {set Search_What playList}
               "auto*" {set Search_What 0}
               default {set Search_What 1}
            }
         }
         search_replace {
            # stage 3, do the filtering
            if {![info exists Search_What]} {set Search_What 1}
            set Search_Replace [string tolower $Cmd]
            filtering::pushFilter
            macros::filterMusic $Search_Replace $Search_What
            # stage 3
         }
         cmb_PL {
            # Stage 1
            # Set the playlist file
            foreach ext {m3u taf} {
               set Playlist_File [file join $::HOME $Cmd.$ext]
               if {[file exists $Playlist_File]} {
                  break
               }
            }
         }
         PL_search_replace {
            # Stage 2
            # And then apply it
            macros::ProcessPlaylistFile $Playlist_File [string tolower $Cmd] 0
         }
         default {}
      }
   }
   # Wait to let player take hold
   append Result [HTH::playStatus 1]
   append Result [HTH::trackInfo [soundControl::fileName]]
   append Result [HTH::nextToPlay 1 $::snackAmpSettings(httpNextLength)]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
#------------------------------------------------------------------------------
# Function   : ::remote/control::aplFiles
# Description: Callback from TML file to return a list of all auto-playlist files
# Author     : Tom Wilkason
# Date       : 4/28/2002
#------------------------------------------------------------------------------
proc remote/control::aplFiles {args} {
   Trace
   set result ""
   # Create option list of all auto-playlist files
   foreach {file} [catalog::getAutoPlaylists] {
      set choice [ROOTTAIL $file]
      if {[string equal $choice $::TA(AutoPlayListSelection)]} {
         append result "<option selected>$choice</option>\n"
      } else {
         append result "<option>$choice</option>\n"
      }
   }
   return [string trim $result]
}
#------------------------------------------------------------------------------
# Function   : ::remote/control::plFiles
# Description: Callback from TML file to return a list of all playlist files
# Author     : Tom Wilkason
# Date       : 4/28/2002
#------------------------------------------------------------------------------
proc remote/control::plFiles {args} {
   Trace
   set result ""
   # Create option list of all auto-playlist files
   foreach {file} [catalog::getPlaylists] {
      set choice [ROOTTAIL $file]
      if {[string equal $choice $::TA(PlayListSelection)]} {
         append result "<option selected>$choice</option>\n"
      } else {
         append result "<option>$choice</option>\n"
      }
   }
   return [string trim $result]
}
#------------------------------------------------------------------------------
#                              coverart server processing
#------------------------------------------------------------------------------
namespace eval coverart {
   proc init { } {
     return 1
   }
}
# Returns a local file that contains coverart
# For security only handles specific file names
proc coverart::process {sock {params ""}} {
   Trace
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         switch -- $CmdType {
            get {
               set folder [HttpD::Decode $Cmd]
               set img [coverart::covername $folder]
               if {[string length $img]} {
                  set ext [string trimleft [file extension $img] .]
                  return [list file $ext $img]
               }
            }
         } ;# end switch
      }
   }
   return [list dhtml html ""]
}
##
# Determine if a valid coverart file exists in this folder
#
proc coverart::covername {folder} {
   Trace
   foreach {image} $::snackAmpSettings(coverart) {
      set img [file join $folder $image]
      if {[file exists $img]} {
         return $img
      }
   }
   return ""
}

#------------------------------------------------------------------------------
#                              d i s p l a y
#------------------------------------------------------------------------------
namespace eval display {
   proc init { } {
     return 1
   }
}
#------------------------------------------------------------------------------
# Function   : display
# Description: Main function that is called by web server when /display
#              is entered.
#              e.g. /display
# Author     : Tom Wilkason
# Date       : 12/15/2002
#------------------------------------------------------------------------------
proc display::process {sock {params ""}} {
   Trace
   append Result [HTH::leader "SnackAmp Playlist" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [HTH::getRefreshScript [HTH::nextRefresh] "/display"]
   # Wait to let player take hold
   #append Result ${search::close_window}
   append Result [HTH::playStatus 0]
   append Result [HTH::trackInfo [soundControl::fileName]]
   append Result [HTH::nextToPlay 0 $::snackAmpSettings(httpNextLength)]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
#------------------------------------------------------------------------------
#                              s t r e a m i n g  f i l e  b r o w s e r
#------------------------------------------------------------------------------
namespace eval dirBrowser {
}
##
# url to show a directory listing to allow browsing
# of files by folder.
# todo:
# button to play folder
# button in insert file/folder into playlist
# todo: Add form controls to define how remote/tracks/sorting is to be used.
#------------------------------------------------------------------------------
# Function   : dirBrowser::browsePlay
# Description:
# Author     : Tom Wilkason
# Date       : 6/19/2003
#------------------------------------------------------------------------------
proc dirBrowser::doit {sock {params ""}} {
   Trace
   set myns [uplevel {string trim [namespace current] :}]
   switch -glob -- $myns {
      jukebox* {set control 2}
      remote* {set control 1}
      display* {set control 0}
      default {set control 0}
   }
   upvar #0 Httpd$sock data
   set title "SnackAmp Jukebox"
   set result [HTH::leader $title $::snackAmpSettings(cssName)]
   append result "<body>"
   append result [js::scripts popupInfo toggleAll_jb_add toggleAll_jb_search_pl textWindow]

   set Tracks [list]
   set sortMethod "shuffle"
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         set dCmd [HttpD::Decode $Cmd]
         switch -- $CmdType {
            db_add_start {
               # When processing a playlist add, this is the first param
               set playing [LB::idxActive]
               set old_end [LB::idxEnd]
               # The back reference uses the db_add_start param (which is the params from the
               # request that originated the search) which was passed in here. It is saved
               # In a hidden field from the prior form.
               append result [AREF "?[NOAMP $dCmd]" Back]
               append result [H2 "Added the following tracks to the playlist"]
               append result "<table class='songPlaylist'>\n"
            }
            db_add {
               # When processing a playlist add, these are the middle params, one per item to add
               set track [NOAMP [ROOTTAIL $dCmd]]
               set folder [NOAMP [HTH::displayFolder $dCmd]]
               append result [HTH::tableRow [list $folder $track] headerRow]
               lappend Tracks $dCmd
            }
            db_append_process {
               # When processing a playlist add, this is the final param, signals end of list
               append result "</table>\n"
               if {[llength $Tracks] > 0} {
                  foreach {track} $Tracks {
                     set next [LB::idxFile $track]
                     if {$control==1 || $next==-1 || $next <= [LB::idxPlaying] || ($next-[LB::idxActive])>$::snackAmpSettings(minNextTrack)} {
                        LB::userInsert $track
                     }
                  }
                  append result [H2 "Your requested tracks were queued to play"]
                  gui::updateFileDisplay
               } else {
                  append result [H2 "No tracks selected or tracks already in playlist"]
               }
            }
            play {
               # Play a specific song
               # /control?play=some+name.mp3
               set track $dCmd
               if {[catch {LB::InsertPlay $track} error] } then {
                  debug $result
                  append result $result
               } else {
                  append result [H2 "Playing '$track'"]
               }
            }
            append {
               set track $dCmd
               if {[HTH::isMedia $track]} {
                  set next [LB::idxFile $track]
                  if {$control==1 || $next==-1 || $next <= [LB::idxPlaying] || ($next-[LB::idxActive])>$::snackAmpSettings(minNextTrack)} {
                     # Insert after last web insert (instead of end) if possible
                     LB::userInsert $track
   #                  append result [H2 "Added $track to Playlist"]
                     gui::updateFileDisplay
                  } else {
                     append result [H2 "Skipping $track, it is already in playlist"]
                  }
               } else {
                  append result [H2 "Skipping $track, it is invalid"]
               }
            }
            insert {
#               set track $dCmd
#               append result [H2 "Queued $track to play next"]
#               set cp [LB::idxPlaying]
#               LB::InsertFile [incr cp] $track
#               gui::updateFileDisplay
            }
            delete {
               LB::deleteIndex $dCmd
               gui::updateFileDisplay
               # Remove from playlist
            }
            folder {
               # No refresh when browsing
               # append result [HTH::getRefreshScript [HTH::nextRefresh] "?folder=$Cmd"]
               # TODO: nextToPlay queuedToPlay folderList should all show metakit data
               append result [H2 "Now [string tolower [soundControl::playState]]:"]
               append result [HTH::trackInfo [soundControl::fileName] 1]
               append result {<table><tr><td valign="top">}
               append result [HTH::nextToPlay $control 10 "&folder=$Cmd"]
               append result {</td><td valign="top">}
               append result [HTH::queuedToPlay $control 10 "&folder=$Cmd"]
               append result {</td></tr></table>}
               append result [H2 "Select a folder and pick your favorite track"]
               append result [browser::folderList $Cmd $control $params]
            }
         }
      }
   } else {
      append result [HTH::getRefreshScript [HTH::nextRefresh] "?folder=[info hostname]"]
      append result [H2 "SnackAmp currently [string tolower [soundControl::playState]]"]
      append result [HTH::trackInfo [soundControl::fileName] 1]
      append result {<table><tr><td valign="top">}
      append result [HTH::nextToPlay $control 15]
      append result {</td><td valign="top">}
      append result [HTH::queuedToPlay $control 15]
      append result {</td></tr></table>}
      append result [H2 "Select a folder and pick Your favorite track"]
      append result [browser::folderList [info hostname] $control $params]
   }
   append result "</body>\n</html>\n"
   return [list dhtml html [string trim $result]]
}
#            db_insert_process {
#               ;# When processing a playlist add, this is the final param, signals end of list
#               append result "</table>\n"
#               append result [H2 "Your requested tracks are now queued to play next"]
#               set cp [LB::idxPlaying]
#               foreach {track} $Tracks {
#                  LB::InsertFile [incr cp] $track
#               }
#               gui::updateFileDisplay
#            }
##
# Return the folderList of the browser window
#
proc browser::folderList {Cmd control params} {
   Trace
   # If selected is the toplevel, assume parent is the hostname
   set folder [HttpD::Decode $Cmd]
   array set sides {1 left 2 left}
   set sideNum 1
   set files [list]
   set me  [info hostname]
   append result "<div class='folderDiv'>"
   append result [subst {$HTH::folderopen [AREF "?folder=$me" $me] }]
   if {[string equal $me $folder]} {
      set atroot 1
      set parent $me
      set folders $::TA(mediaDirs)
      append result "<br>\n"
   } else {
      set atroot 0
      set folders [list]
      set validFolder 0
      ##
      # For security, verify this folder is actual one off
      # one of the media folders.
      #
      foreach media $::TA(mediaDirs) {
         if {[string match $media* $folder ]} {
            set validFolder 1
            break
         }
      }
      if {$validFolder==0} {
         set parent $me
         set folders $::TA(mediaDirs)
         append result "<br>\n"
      }  else {
         ##
         # Display this and child folders
         #
         # Create back track in proper order
         set tlevel $folder
         set dirPath {}
         while {1} {
            set dirPath [linsert $dirPath 0 $tlevel]
            # Can't back up beyond a media folder
            if {[lsearch $::TA(mediaDirs) $tlevel] >= 0} {
               break
            }
            set under [file dirname $tlevel]
            if {[string equal $tlevel $under]} {
               # Can't backup any more
               break
            } else {
               set tlevel $under
            }
         }
         ##
         # Create a back path to the root folder from selected folder
         # Last folder had no hyperlink
         set i 0
         foreach {tlevel} $dirPath {
            incr i
            if {$i < [llength $dirPath]} {
               append result [subst {<b></b>[SP][AREF "?folder=[ENCODE $tlevel]" [FTAILIFCAN $tlevel]] }]
            }  else {
               append result "<b></b>[SP][FTAILIFCAN $tlevel]<br>\n"
            }
         }
         ##
         # Build up a list of all the folders and files in the two tables, folders then tracks
         # TFW: cache the files and folders, fire after time to unset them later
         set files [db::filesInFolder $folder 0]
         set folders [lsort -dictionary [glob -nocomplain -types d -directory $folder *]]
      }
   }
   ##
   # Show the sub-folders in a table along with coverart if possible
   #
   if {[llength $folders]} {
      incr sideNum
      set tail [file tail $folder]
      # Find minimum folder length for indent offsets
      if {($control==1 && $::snackAmpSettings(remote,brcoverart)) || ($control==2 && $::snackAmpSettings(jukebox,brcoverart))} {
         set sca 1
      } else {
         set sca 0
      }
      foreach item $folders {
         if {$sca} {
            set cov [HTH::coverArt $item]
            if {[string length $cov]} {
               set ca [AREF "?folder=[ENCODE $item]" $cov]
               append ca <br>
            } else {
              set ca ""
            }
         } else {
              set ca ""
         }
         if {$atroot} {
            set link [subst {[SP]$HTH::folderclosed [AREF "?folder=[ENCODE $item]" $item]<br>\n}]
         } else {
            set link [subst {[SP]$HTH::folderclosed [AREF "?folder=[ENCODE $item]" [FTAILIFCAN $item]]<br>\n}]
         }
         append result [subst {<tr><td>$ca$link</td></tr>}]
      }
   }
   append result "</div>"
   ##
   # Show the files in a table
   # TFW: This should be in a form with each item with a checkbox to add multiple
   # ones to the playlist (like search results)
   # TFW: The table needs alternating rows of color, with no border. Can use styles
   if {[llength $files]} {
      if {($control==1 && $::snackAmpSettings(remote,rettofolder)) || ($control==2 && $::snackAmpSettings(jukebox,rettofolder))} {
         set rtf 1
      } else {
         set rtf 0
      }
      append result "<div class='fileDiv''>"
      if {$control==1} {
         append result [subst {<FORM METHOD="POST" ACTION="" NAME="jb_add_form">}]
         append result "<input type=hidden name=db_add_start value=\"$params\">"
      }
      append result [HTH::coverArt $folder]
      append result "<font size='5'>[file tail $folder]</font>"
      append result "<table class='songPlaylist'>"
      foreach file $files {
         set bgc [HTH::rowStyle i]
         # Need to strip out prior append=params (no longer needed with popup)
         # regsub {(append=.*?)&} $params  {} params
         set cb "<INPUT TYPE=checkbox NAME='db_add' VALUE=\"$file\">"
         # This is for dag to not show the same folder over
         if {$rtf} {
            set ftail [AREF "?append=[ENCODE $file]&folder=$Cmd" [DBINFO $file] "Queue Track to Play"]
         } else {
            set ftail [AREF "?append=[ENCODE $file]&folder=[info hostname]" [DBINFO $file] "Queue Track to Play"]
         }
         if {$control==1} {
            set buttonLink [HTH::popupInfo $file]
            set trImage [HTH::popUpProcess play "" $file] ;# This is not working! get java error
            append result [HTH::tableRow [list $cb $buttonLink "$trImage $ftail"] $bgc]
         } else {
            set trImage $HTH::trackimage
            append result [HTH::tableRow [list "$trImage $ftail"] $bgc]
         }
      }
      append result "</table>\n"
      if {$control==1} {
         append result "<INPUT TYPE=submit VALUE='Queue Selected' NAME=db_append_process>\n"
         append result "</FORM>\n"
         append result [subst {[AREF "javascript:toggleAll_jb_add(true)" "Select All"] | [AREF "javascript:toggleAll_jb_add(false)" "Unselect All"]<br>}]
      }
      append result "</div>"
   } else {
   #                  append result "<hr>\n"
   #                  append result [H2 "No tracks within [file tail $folder]"]
   }
   return $result
}

##################################################################################
#                             m i n i c o n t r o l
##################################################################################
##
# mini remote control (for P80x phones :-)
# returns a mini playlist
#
namespace eval remote/mini {
   proc init { } {
     return 1
   }
}

proc remote/mini::execute {} {
   upvar data data
   set args [split $data(query) &]
   eval [concat [list HttpD::GetParams {playId "" cmd "" volume "" panel.x "" panel.y "" filter "" gofilter ""}] $args]
   set rc ""
   if {$playId!=""} {
      set n [LB::keyFromIndex $playId]
      if {$n!=""} {
	 LB::activate $playId; Play active
      }
   }
   # Lay on commands for the buttons.gif image
   set px ${panel.x}; set py ${panel.y}; if {$px!=""} {
      if {$py<16} {
         if {$px<80} {set cmd [lindex {prev stop pause play next} [expr {$px/16}]]}
         if {$px>=106} {
	    set volume [expr {$px-106}]
	 }
      }  else  {
	 set pos [expr {round(double([soundControl::length])*double($px)/170.0)}]
	 PlayOffset $pos
      }
   }
   if {($gofilter!="")||($filter!="")} {
       set ::TA(filter) $filter
       filterMusic replace
       return
   }
   if {$cmd!=""} {
      switch -- $cmd {
	 prev {
	    Prev; update
	 }
	 next {
	    Next; update
	 }
	 play {
	    Play
	 }
	 pause {
	    Pause
	 }
	 stop {
	    Stop
	 }
      }
   }
   if {$volume!=""} {
      set ::TA(gain) $volume
      ::PlayGain $volume
   }
   return ""
}

proc remote/mini::_calctime {t} {
   set t [expr {round($t)}]
   return [format "%d:%02d" [expr {$t/60}] [expr {$t%60}]]
}

proc remote/mini::status {} {
   if {[snack::audio active]} {
      if {$soundControl::playState!="Stopped"} {
	 if {$soundControl::paused==1} {
	    set pre "Paused"
         }  else  {
	    set pre "Playing"
         }
	 set vol [expr {round(($::TA(gain)*100.0)/64.0)}]
         return "$pre [_calctime [soundControl::position]] / [_calctime [soundControl::duration]] ($vol %)"
      }  else  {
	 return "Stopped"
      }
   }
   return "Not playing anything"
}
proc remote/mini::buttons {} {
   set rc ""
   return {<center><form action="/remote/mini/" method="get"><input type="image" name="panel" src="/remote/mini/buttons.gif" width="170" height="28"></form></center>}
}
proc remote/mini::playlist {styleN styleA cnt} {
   set cnt0 [expr {$cnt/2}]
   set cnt1 [expr {$cnt-$cnt0-1}]
   set cur [LB::idxPlaying]
   if {$cur<0} {set cur 0}
   set end [LB::idxEnd]
   if {$cur<$cnt0} {
      incr cnt1 [expr {$cnt0-$cur}]
      set cnt0 $cur
   }
   if {($cur+$cnt1)>=$end} {
      set c [expr {($cur+$cnt1)-$end+1}]
      incr cnt1 -$c
      incr cnt0 $c
      if {$cur<$cnt0} {
	 set cnt0 $cur
      }
   }
   set rc "<table width=\"100%\" border=\"0\" cellspacing=\"0\" cellpadding=\"1\">\n"
   set pos [expr {$cur-$cnt0}]
   set cur [LB::idxPlaying]
   for {set i 0} {$i<($cnt0+$cnt1+1)} {incr i} {
      if {$pos==$cur} {
	 set cls $styleA
	 set a0 "<i>"
	 set a1 "</i>"
      }  else  {
	 set cls [lindex $styleN [expr {$i%[llength $styleN]}]]
	 set a0 "<a href=\"/remote/mini/?playId=$pos\">"
	 set a1 "</a>"
      }
      append rc "<tr><td height=\"14\" class=\"$cls\" onclick=\"location.href='/remote/mini/?playId=$pos'\">$a0[string replace [LB::keyFromIndex $pos] 27 end ...]$a1<br></td></tr>\n"
      incr pos
   }
   append rc "</table>"
   return $rc
}


#------------------------------------------------------------------------------
#                             g e t p l a y l i s t
#------------------------------------------------------------------------------
##
# playlist namespace, returns a nicly formatted list of playlist and other
# to build up streaming playlists.
#
namespace eval streaming/getplaylist.pls {
   proc init { } {
     return 1
   }
}
##
# Playlist code, from a web browser
# Processes a request to play a track, playlist, active playlist or auto-playlist
# Creates a winamp type pls dynamic file that when processed by the /streaming/live command
# will create a queue of songs then play them.
# Note: Add ability to send the entire stream length down as icy data. Good for tracks
# The PlayFile command will be called to process the files (via the registry setting)
##
# Note: TFW: Add another option to sort, shuffle or shufflesort
#            pass it on to the /streaming/live parameter, the sort param
#            should be specified first, otherwise it will be
#            ignored.
proc streaming/getplaylist.pls::process {sock {params ""}} {
   Trace
   upvar #0 Httpd$sock data
   set result ""
   set sortMethod "shuffle"
   set how "stream"
   set param ""
   set method "unknown"
   set max 50
   if {$::snackAmpSettings(useIP)} {
      set refer "http://$::snackAmpSettings(ipAddress):$::snackAmpSettings(usedPort)"
   } elseif {[string length $::snackAmpSettings(altStSrc)]>0} {
      set refer "http://$::snackAmpSettings(altStSrc):$::snackAmpSettings(usedPort)"
   } else {
      set refer $::snackAmpSettings(internalWebBase)
   }
   if {[string length $params]} {
      foreach pr [split $params &]  {
         foreach {CmdType Cmd} [split $pr =] {break}
         # Create a dynamic playlist
         switch -- $CmdType {
            active -
            playlist -
            track -
            folder -
            apl {
               set method $CmdType
               set param $Cmd
            }
            tree {
               switch -- $Cmd {
                  no {set fmethod folder}
                  yes {set fmethod treefolder}
               }
            }
            sort {set sortMethod [HttpD::Decode $Cmd]}
            how  {set how [HttpD::Decode $Cmd]}
            submit  {}
            default {saLog "streaming/getplaylist.pls::process: Invalid subcommand - '$CmdType'"}
         }
      }
      ##
      # We have all the params, now process the options
      #
      switch -- [string tolower $how] {
         stream {
            # Create a stream of multiple tracks (except track option)
            switch -- $method {
               active {
                  set    result "File1=$refer/streaming/live?sort=[ENCODE $sortMethod]&active=$param\n"
                  append result "Title1=Active Playlist"
               }
               apl {
                  set    result "File1=$refer/streaming/live?sort=[ENCODE $sortMethod]&apl=$param\n"
                  append result "Title1=Auto Playlist [HttpD::Decode $param]"
               }
               playlist {
                  set    result "File1=$refer/streaming/live?sort=[ENCODE $sortMethod]&playlist=$param\n"
                  append result "Title1=Playlist [HttpD::Decode $param]"
               }
               treefolder -
               folder {
                  set    result "File1=$refer/streaming/live?sort=[ENCODE $sortMethod]&$fmethod=$param\n"
                  append result "Title1=Folder [nameToShort [HttpD::Decode $param]]"
               }
               track {
                  set    result "File1=$refer/streaming/live?track=$param\n"
                  append result "Title1=Track [nameToShort [HttpD::Decode $param]]"
               }
               default {saLog "streaming/getplaylist.pls::process: Invalid subcommand method=$method"}
            }
         }
         tracks {
            # Create individual tracks
            switch -- $method {
               active {
                  if {[catch {streaming/live::activeplaylist $sortMethod} tracks]} then {
                     saLog $tracks
                  } else {
                     set i 1
                     set result ""
                     foreach index $tracks {
                        set track [db::TrackFromDb $index]
                        append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                        append result "Title$i=[nameToShort $track]\n"
                        incr i;if {$i > $max} {break}
                     }
                  }
               }
               apl {
                  if {[catch {streaming/live::apl [HttpD::Decode $param] $sortMethod} tracks] } then {
                     saLog $tracks
                  } else {
                     set i 1
                     set result ""
                     foreach index $tracks {
                        set track [db::TrackFromDb $index]
                        append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                        append result "Title$i=[nameToShort $track]\n"
                        incr i;if {$i > $max} {break}
                     }
                  }
               }
               playlist {
                  if {[catch {streaming/live::playlist [HttpD::Decode $param] $sortMethod} tracks]} then {
                     saLog $tracks
                  } else {
                     set i 1
                     set result ""
                     # shuffle
                     foreach index $tracks {
                        set track [db::TrackFromDb $index]
                        append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                        append result "Title$i=[nameToShort $track]\n"
                        incr i;if {$i > $max} {break}
                     }
                  }
               }
               folder {
                  set i 1
                  set result ""
                  set path [HttpD::Decode $param]
                  set files [db::filesInFolder $path 0]
                  foreach track $files {
                     append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                     append result "Title$i=[nameToShort $track]\n"
                     incr i;if {$i > $max} {break}
                  }
               }
               treefolder {
                  set i 1
                  set result ""
                  set path [HttpD::Decode $param]
                  set tracks [db::filesMatchingPattern [file join $path *]]
                  foreach track $tracks {
                     append result "File$i=$refer/streaming/live?track=[ENCODE $track]\n"
                     append result "Title$i=[nameToShort $track]\n"
                     incr i
                     if {$i > $max} {break}
                  }
               }
               track {
                  set    result "File1=$refer/streaming/live?track=$param\n"
                  append result "Title1=Track [nameToShort [HttpD::Decode $param]]"
               }
               default {saLog "streaming/getplaylist.pls::process: Invalid subcommand method=$method"}
            }
         }
         default {saLog "streaming/getplaylist.pls::process: Invalid subcommand how=$how"}
      }
      return [list dhtml pls [string trim $result]]
   } else {
      set title "SnackAmp Jukebox"
      set result [HTH::leader $title $::snackAmpSettings(cssName)]
      append result [H1 "Please specify a parameter"]
      append result "</body>\n</html>"
      if {$::TA(useputs)} {
         set mypath [HTH::WriteTempFile pls $result]
         return [list file pls $mypath]
      } else {
         return [list dhtml html $mypath]
      }
   }
}
#------------------------------------------------------------------------------
#                              l i v e
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
# Function   : streaming/live namespace
# Description: Process streaming files
# Author     : Tom Wilkason
# Date       : 3/13/2003
#------------------------------------------------------------------------------
namespace eval streaming/live {
   proc init {args} {
      return 1
   }
}
##
# the /streaming/live process method just retuns a file name from an existing file.
# used to initiate a stream to a client. This is run from shortcuts
# AFTER the playlist has been built and selected.
# Note: TFW: Add another option to sort, shuffle or sort/shuffle
#
proc streaming/live::process {sock {params ""}} {
   Trace
   upvar #0 Httpd$sock data
   set Tracks [list]
   set sortMethod "shuffle"
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         set Cmd [HttpD::Decode $Cmd]
         switch -- $CmdType {
            apl {
               # Get some apl file list of indicies
               if {[catch {streaming/live::apl $Cmd $sortMethod} result] } then {
                  return -code error $result
               } else {
                  if {[llength $result] > 0} {
                     streamBuffer::setList $sock $result
                     set data(streamingLength) 44000000
                     set data(Title) "AutoPlaylist From $data(mime,host)"
                     set data(Album) "$Cmd ([llength $result] trks)"
                     return [list file html [streamBuffer::popFile $sock]]
                  } else {
                     return -code error "No apl files for '$Cmd'"
                  }
               }
            }
            playlist {
               # Get some tracks from a playlist
               if {[catch {streaming/live::playlist $Cmd $sortMethod} result]} then {
                  return -code error $result
               } else {
                  ##
                  # Don't return track, return a m3u file
                  #
                  if {[llength $result] > 0} {
                     streamBuffer::setList $sock $result
                     set data(streamingLength) 44000000
                     set data(Title) "Playlist From $data(mime,host)"
                     set data(Album) "$Cmd ([llength $result] trks)"
                     return [list file html [streamBuffer::popFile $sock]]
                  } else {
                     return -code error "No playlist files for '$Cmd'"
                  }
               }
            }
            active {
               # Get some tracks from a playlist
               if {[catch {streaming/live::activeplaylist $sortMethod} result]} then {
                  return -code error $result
               } else {
                  ##
                  # Don't return track, return a m3u file
                  #
                  if {[llength $result] > 0} {
                     streamBuffer::setList $sock $result
                     set data(streamingLength) 44000000
                     set data(Title) "Active Playlist From $data(mime,host)"
                     set data(Album) "$Cmd ([llength $result] trks)"
                     return [list file html [streamBuffer::popFile $sock]]
                  } else {
                     return -code error "No active playlist files for '$Cmd'"
                  }
               }
            }
            folder {
               set result [list]
               set files [sortByMethod [db::filesInFolder $Cmd 0] $sortMethod]
               foreach file $files {
                  set loc [db::whereInDb $file]
                  if {$loc >= 0} {
                     lappend result $loc
                  }
               }

               if {[llength $result] > 0} {
                  streamBuffer::setList $sock $result
                  set data(streamingLength) 44000000
                  set data(Title) "Folder $Cmd From $data(mime,host)"
                  set data(Album) "$Cmd ([llength $result] trks)"
                  return [list file html [streamBuffer::popFile $sock]]
               } else {
                  return -code error "No folder files found for '$Cmd'"
               }
            }
            treefolder {
               set files [sortByMethod [db::filesMatchingPattern [file join $Cmd *]] $sortMethod]
               set result [list]
               foreach file $files {
                  set loc [db::whereInDb $file]
                  if {$loc >= 0} {
                     lappend result $loc
                  }
               }

               ##
               # Don't return track, return a m3u file
               #
               if {[llength $result] > 0} {
                  streamBuffer::setList $sock $result
                  set data(streamingLength) 44000000
                  set data(Title) "Treefolder $Cmd From $data(mime,host)"
                  set data(Album) "$Cmd ([llength $result] trks)"
                  return [list file html [streamBuffer::popFile $sock]]
               } else {
                  return -code error "No folder files found for '$Cmd'"
               }
            }
            track {
               # Process a single track in a playlist
               set loc [db::whereInDb $Cmd]
               # for DB case can improve by querying all at once
               if {$loc >= 0} {
                  set file [db::TrackFromDb $loc]
                  streamBuffer::setList $sock $loc
                  set data(streaming) $loc
                  set data(streamingLength) [file size $file]
                  if {[catch {catalog::trackFullData $file Details} error] } then {
                     saLog $error
                  } else {
                     # Fill in data that will go into icy-meta tags
                     foreach {index} {Title Album Artist Genre Rate} {
                        if {[info exists Details($index)]} {
                           set data($index) $Details($index)
                        } else {
                           set data($index) ""
                        }
                     }
                  }
                  return [list file html [streamBuffer::popFile $sock]]
               }
            }
            sort        {set sortMethod $Cmd}
            default {
               return -code error "streaming/live::process: Invalid subcommand - '$CmdType'"
            }
         }
      }
   } else {
      return -code error "streaming/live::process: No Parameter '$params'"
   }
}
##
#  Read in an autoplaylist file and return the tracks,
#  properly sorted, within it to the stream
#
proc streaming/live::apl {name sortMethod} {
   Trace
   set fName [file join $::HOME $name.tafs]
   if {![file exists $fName]} {
      return -code error "$fName does not exist"
   } else {
      if {[catch {open $fName r} fid]} then {
         return -code error $fid
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]
      close $fid
      # Get folder list
      set biglist [list]
      # Grab copy of folders
      array set myte [aplFolder::folderArray]
      foreach line $full {
         foreach {Index Value} $line {
            set myte($Index) $Value
         }
      }
      # Grab locations in an array and shuffle/sort it
      # TODO: Can this can be optimized?
      set tracks [sortByMethod [db::unSortedList] $sortMethod]
      foreach track $tracks {
         set trackDir [file dirname $track]
         if {[info exists myte($trackDir)]} {
            if {$myte($trackDir)==1} {
               lappend biglist [db::whereInDb $track]
            }
         } else {
            #saLog "Couldn't find '$trackDir' as an index"
         }
      }
   }
   return $biglist
}
##
# Return a list of indicies for a playlist
#
proc streaming/live::playlist {name sortMethod} {
   Trace
   set name [string trim $name \"] ;#"
   foreach ext {m3u taf} {
      set fName [file join $::HOME $name.$ext]
      if {[file exists $fName]} {
         break
      }
   }
   if {![file exists $fName]} {
      saLog "streaming/live::playlist: $fName does not exist"
      return -code error "$fName does not exist"
   } else {

      # Here we have a file case issue
      if {[catch {open $fName r} fid]} then {
         return -code error $fid
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]

      close $fid
      set tracks [sortByMethod $full $sortMethod]
      # Get folder list
      set biglist [list]
      foreach line $tracks {
         set line [file join $line] ;# make tcl name
         set loc [db::whereInDb $line]
         if {$loc >= 0} {
            lappend biglist $loc
         } else {
 #           saLog "$line not in database"
         }
      }
   }
   return $biglist
}
##
# Return a list of indicies to current playlist on GUI
#
proc streaming/live::activeplaylist {sortMethod} {
   set biglist [list]
   ##
   # Note: May want to get from cursor foward.
   #
   set tracks [sortByMethod [LB::keyOrNamesAll 1] $sortMethod]
   foreach file $tracks {
      set loc [db::whereInDb $file]
      if {$loc >= 0} {
         lappend biglist $loc
      }
   }
   return $biglist
}
proc streaming/live::folder {name} {
}
proc streaming/live::recurfolder {name} {
}
#------------------------------------------------------------------------------
# Function   : sortByMethod
# Description: Sort a list by the appropriate method
# Author     : Tom Wilkason
# Date       : 3/24/2003
#------------------------------------------------------------------------------
proc streaming/live::sortByMethod {Items how} {
   saLog "[me] [llength $Items] $how"
   set index [lsearch -exact $::filtering::sortModes $how]
   switch -- $index {
      0  {return [lsort -dictionary $Items]}
      1  {return [shuffle $Items]}
      2  {return [shuffleByFolder $Items ldsort]}
      3  {return [shuffleByFolder $Items shuffle]}
      default {
         package require subSorting
         set tosort [subSort::getModes]
         return [db::subSorting $tosort $Items]
      }
   }
}

#------------------------------------------------------------------------------
#                              s t r e a m i n g  f i l e  b r o w s e r
#------------------------------------------------------------------------------
namespace eval streaming/browser {
   proc init { } {
     return $::snackAmpSettings(EnableStreaming)
   }
   variable brform {
  <H2>Enter Play Folder Option</H2>
  <form name=option_form>
  <TABLE>
    <TR>
      <TD>Sorting</TD>
      <TD>Streaming</TD>
      <TD>Subfolders</TD>
    </TR>
    <TR>
     <TD>
       <select size="1" name="cmb_Sort" class="formText">
         [HTH::sortModes]
       </select>
     </TD>
     <TD>
       <select size="1" name="cmb_How" class="formText">
         <option selected>stream</option>
         <option>tracks</option>
       </select>
     </TD>
     <TD>
       <select size="1" name="cmb_Sub" class="formText">
         <option selected>no</option>
         <option>yes</option>
       </select>
     </TD>
    </TR>
  </TABLE>
  </form>
   }
}
##
# url to show a directory listing to allow browsing
# of files by folder.
# todo:
# button to play folder
# button in insert file/folder into playlist
# todo: Add form controls to define how streaming/tracks/sorting is to be used.
proc streaming/browser::process {sock {params ""}} {
   Trace
   variable brform
   upvar #0 Httpd$sock data
   set title "SnackAmp Jukebox"
   set result [HTH::leader $title $::snackAmpSettings(cssName)]
   append result <body>
   append result [js::scripts popupInfo foldersel]

   set sortMethod "shuffle"
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         switch -- $CmdType {
            sort {
               set sortMethod $Cmd
            }
            folder {
               append result [H1 "SnackAmp media folders on [info hostname]"]
               set folders [list]
               set files [list]
               set folder [HttpD::Decode $Cmd]
               set validFolder 0
               foreach media $::TA(mediaDirs) {
                  if {[string match $media* $folder ]} {
                     set validFolder 1
                     break
                  }
               }
               if {$validFolder==0} {
                  set folder [lindex $::TA(mediaDirs) 0]
               }
               set files [db::filesInFolder $folder 0]
               set folders [lsort -dictionary [glob -nocomplain -types d -directory $folder *]]
               #----------------------------
               ##
               # Create a list of all folders, includling toplevel ones
               #
               set fldr $folder
               while {1} {
                  # Get list of all parent folders
                  set under [file dirname $fldr]
                  if {[string equal $under $fldr]} {
                     break
                  } else {
                     lappend folders $fldr ;########
                     set fldr $under
                  }
               }
               # List of toplevel folder
               foreach item $::TA(mediaDirs) {
                  lappend folders $item
                  if {[string match "$item*" $folder]} {
                     set icon $HTH::folderopen
                  } else {
                     set icon $HTH::folderclosed
                  }
                  append result [subst {$icon <a href="?folder=[ENCODE $item]">$item</a><br>\n}]
               }
               #----------------------------
               set folders [lsort -unique $folders]
               # Show the folders
               if {[llength $folders]} {
                  append result "<hr>\n"
                  append result [subst $brform]
                  append result [H2 "Current folder $folder"]
                  append result "<table class='songPlaylist'>"
                  # Find minimum folder length for indent offsets
                  foreach item $folders {
                     lappend reps [llength [file split $item]]
                  }
                  set minrep [lindex [lsort -integer $reps] 0]
                  foreach item $folders {
                     set icon $HTH::folderclosed
                     if {[string equal $item $folder]} {
                        set icon $HTH::folderopen
                     }
                     set rep [expr {[llength [file split $item]]-$minrep}]
                     if {[string length [file tail $item]]} {
                        set titum [file tail $item]
                     } else {
                        set titum $item
                     }
                     set link "[string repeat [SP] $rep]$icon <a href=\"?sort=$sortMethod&folder=[ENCODE $item]\">$titum</a><br>\n"
                     # S = Stream folder contents
                     # T = Tracks for folder contents
                     set base "/streaming/getplaylist.pls"
                     set F [subst {<a href="javascript:foldersel('$base','[ESCAPE $item]')" title="Play Track Immediately">$HTH::playbutton</A>}]
                     append result [HTH::tableRow [list $F $link] darkRows]
                  }
                  append result "</table>\n"
               }
               # Show the files
               if {[llength $files]} {
                  append result "<hr>\n"
                  append result [HTH::coverArt $folder]
                  append result "<table class='songPlaylist'>"
                  foreach file $files {
                     set I [HTH::popupInfo $file]
                     set link [AREF "/streaming/getplaylist.pls?sort=$sortMethod&track=[ENCODE $file]" $HTH::playbutton "Play Track Immediately"]<br>
                     set size "[commify [expr {[file size $file]/1024}]] KB"
                     append result [HTH::tableRow [list $I $link [DBINFO $file] $size] [HTH::rowStyle i]]
                  }
                  append result "</table>\n"
               } else {
                  append result "<hr>\n"
                  append result [H1 "No tracks within $folder"]
               }
            }
         }
      }
   } else {
      append result [H1 "SnackAmp folder browser"]
      append result [H2 "Please select a folder below to browse"]
      foreach folder $::TA(mediaDirs) {
         append result $HTH::folderclosed[AREF "?folder=[ENCODE $folder]" $folder]<br>
      }
   }
   append result "</body>\n</html>\n"
   return [list dhtml html [string trim $result]]
}
#------------------------------------------------------------------------------
#                 s t r e a m i n g   p l a y l i s t  e x a m p l e
#------------------------------------------------------------------------------
##
# Example code to display a selection of sources on this server
#
namespace eval streaming/playlists {
   proc init { } {
     return 1
   }
}
proc streaming/playlists::process {sock {params ""}} {
   Trace
   upvar #0 Httpd$sock data
   ##
   # This can be prettied up to select the various
   # playlists that are created.
   # Can add options to use sorting options
   #
   set title "SnackAmp Jukebox"
   set result [HTH::leader $title $::snackAmpSettings(cssName)]
   append result <body>
   ##
   # Need to add code to show folder trees
   #
   append result {[H1 <a href='/streaming/getplaylist.pls?active=active'>Current playlist</a>]}
   set Files [catalog::getAutoPlaylists]
   # Create option list of all auto-playlist files
   if {[llength $Files] > 0} {
      append result [H1 "Auto-Playlist files"]
      foreach {file} $Files {
         set choice [ROOTTAIL $file]
         append result [AREF "/streaming/getplaylist.pls?apl=[ENCODE $choice]" $choice]<br>
      }
   }
   # Create option list of all playlist files
   set Files [catalog::getPlaylists]
   if {[llength $Files] > 0} {
      append result [H1 "Playlist files"]
      foreach {file} $Files {
         set choice [ROOTTAIL $file]
         append result [AREF "/streaming/getplaylist.pls?playlist=[ENCODE $choice]" $choice]<br>
      }
   }
   append result "</body>\n</html>\n"
   return [list dhtml html [string trim $result]]
}
#------------------------------------------------------------------------------
# Function   : search namespace
# Description: Handles interface to request songs to be added to playlist
#              Can be called from different URL types and reacts accordingly
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
namespace eval search {
   variable myns
   variable control
   variable search_form {
    <!-- Search Database -->
    <table border="1" width="100%" bordercolor="darkblue" cellspacing="0" cellpadding="0">
      <tr>
        <td width="100%" bgcolor="darkblue" align="center"><font color="white" face="Helvetica">
        <b>Search All Tracks for Matching Text</b></font></td>
      </tr>
      <tr>
        <td width="100%" align="left">
          <form method="POST" action="/search" TITLE="Enter text then hit Search to find all songs that contain the text">
            <input type="submit" value="Search" name="Button">
            <font face="Helvetica">
              <input type="text" name="jb_search" size="30">
            </font>
          </form>
        </td>
      </tr>
    </table>
    <p>
   }
   variable close_window {<a href="" onClick="window.close()">Click this to close the window</A>}
   ##
   # Javascript helper routines
   #
   proc init { } {
     return 1
   }
}
#------------------------------------------------------------------------------
# Function   : search::doit state machine
# Description: Main function that is called by web server when /search
#              is entered.
#              e.g. /search?jb_search_pl=*abc*
# Author     : Tom Wilkason
# Date       : 12/15/2002
#------------------------------------------------------------------------------

# TFW: /search is used now but search needs to be called as /remote/search or /jukebox/search
proc search::doit {sock {params ""}} {
   Trace
   variable search_form
   variable javascript
   upvar #0 Httpd$sock data
   variable myns [uplevel {string trim [namespace current] :}]
   append Result [HTH::leader "SnackAmp Search Page" $::snackAmpSettings(cssName)]
   append Result "<body>\n"
   append Result [js::scripts popupInfo toggleAll_jb_add toggleAll_jb_search_pl foldersel getselected]
   append Result [subst {<a href="/$myns?$params">Refresh</a>|}]
   append Result ${search::close_window}
   switch -glob -- $myns {
      jukebox* {
         set cont 0
      }
      remote* {
         set cont 1
      }
      default {}
   }
   variable control $cont
   # Appropriate link to browser url e.g. /jukebox/browser
   set myBrowser [file join [file dirname $myns] browser]
   ##
   # Need to have search_form optional
   #
   #append Result [subst $search_form]
   set Search_What pl
   set Tracks [list]
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         jb_search {
            # stage 1
            set filter $Cmd
         }
         search_what {
            # stage 2
            switch -glob -- [string tolower $Cmd] {
               "all*files" {
                  # User specified text to search catalog
                  append Result [H2 "The following tracks matched your search criteria '$filter'"]
                  append Result [H2 "Please check desired tracks and press <i>Queue Selected</i> at the bottom to add them to the playlist"]
                  # Pass params for putting in back button on next form
                  append Result [search::searchCatalog $myBrowser $filter [expr {$::snackAmpSettings(httpNextLength)*5}] $params]
               }
               "play*" {
                  # User specifed some text to search within the playlist
                  # This goes in a pop-up
                  append Result [H2 "The following tracks from the Play List matched your search criteria '$filter'"]
                  append Result [search::searchPlaylist $myBrowser $filter 0]
               }
               "auto*" {
                  # User specifed some text to search within the playlist
                  # This goes in a pop-up
                  append Result [H2 "The following tracks from the Auto-Playlist matched your search criteria '$filter'"]
                  append Result [search::searchPlaylist $myBrowser $filter 1]
               }
            }
         }
         play {
            # Play a specific song
            # /control?play=some+name.mp3
            if {[catch {LB::InsertPlay $Cmd} error] } then {
               debug $result
               append Result $result
            }
         }
         delete {
            LB::deleteIndex $Cmd
            gui::updateFileDisplay
            # Remove from playlist
         }
         question {
            # This goes in a pop-up
            # javascript pop-up track details (like playinfo)
            if {![file exists $Cmd]} {
               set Cmd [LB::nameFromKey $Cmd 1]
            }
            # TFW: would be nice to make the fonts smaller
            append Result [HTH::trackInfo $Cmd]
         }
         jb_add_start {
            # When processing a playlist add, this is the first param
            set playing [LB::idxActive]
            set old_end [LB::idxEnd]
            # The back reference uses the jb_add_start param (which is the params from the
            # request that originated the search) which was passed in here. It is saved
            # In a hidden field from the prior form.
            append Result |[AREF "/$myns?$Cmd" Back]\n
            append Result [H2 "Added the following tracks to the playlist"]
            append Result "<table class='songPlaylist'>\n"
         }
         jb_add {
            # When processing a playlist add, these are the middle params, one per item to add
            set next [LB::idxFile $Cmd]
            if {$control==1 || $next==-1 || $next<=$playing} {
               set track [NOAMP [file rootname [file tail $Cmd]]]
               set folder [NOAMP [HTH::displayFolder $Cmd]]
               append Result [HTH::tableRow [list $folder $track] headerRow]
               lappend Tracks $Cmd
            }
         }
         jb_append_process {
            # When processing a playlist append, this is the final param, signals end of list
            append Result "</table>\n"
            if {[llength $Tracks] > 0} {
               append Result [H2 "Your requested tracks were queued to play"]
               LB::userInsertList $Tracks
            } else {
               append Result [H2 "No tracks selected or tracks already in playlist"]
            }
         }
      }
   }
   # Insert Display Playlist link
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
#  jb_insert_process {
#     ;# When processing a playlist insert, this is the final param, signals end of list
#     append Result "</table>\n"
#     if {[llength $Tracks] > 0} {
#        append Result [H2 "Your requested tracks are now queued to play next"]
#        set cp [LB::idxPlaying]
#        foreach {track} $Tracks {
#           LB::InsertFile [incr cp] $track
#        }
#        gui::updateFileDisplay
#     } else {
#        append Result [H2 "No tracks selected or tracks already in playlist"]
#     }
#  }
#------------------------------------------------------------------------------
# Function   : getAPLFileList
# Description: Return a list of all APL files
# Author     : Tom Wilkason
# Date       : 12/24/2002
#------------------------------------------------------------------------------
proc search::getAPLFileList {args} {
   variable control
   variable myns
   #set RF $snackAmpSettings(UseRemovableFolders)
   #set rmtFilter [file join $::RMT *]
   # TFW: Add proper support for removable media here
   ##     Move this to aplFolder namespace
   set result [list]
   foreach {File} [db::sortedList] {
      if {[aplFolder::isSet [file dirname $File]]} {
         lappend result $File
      }
   }
   return $result
}
#------------------------------------------------------------------------------
# Function   : searchPlaylist
# Description: Based on a pattern from the web entry box, filter the current
#              playlist and return the matching results
#
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc search::searchPlaylist {parNs filter {aplOnly}} {
   variable control
   variable myns
   set i 0
   set result {<table class='songPlaylist'>}
   if {$aplOnly} {
      set files [search::getAPLFileList]
   } else {
      set files [LB::keyOrNamesAll 1]
   }
   if {$control==1} {
      append result [HTH::tableRow [list Num ? P X Folder "Track Name"] headerRow]
   } else {
      append result [HTH::tableRow [list Num ? Folder "Track Name"] headerRow]
   }
   foreach {file} $files {
      if {[string match -nocase *$filter* $file]} {
         set bgc [HTH::rowStyle j 5]
         set track [ROOTTAIL $file]
         set folder [HTH::displayFolder $file]
         if {$folder==""} {
            set flink [SP]
         } else {
            set flink [AREF "/$parNs?folder=[ENCODE [file dirname $file]]" $folder]
         }
         set Info [HTH::popupInfo $file]
         # The second param just performs another search using the same filter
         if {$control==1} {
            set Play [subst {<a href="/$myns?play=[ENCODE $file]&jb_search=$filter&search_what=Playlist" title="Play Track Immediately">$HTH::playbutton</a>}]
            set Delete [subst {<a href="/$myns?delete=$i&jb_search=$filter&search_what=Playlist">$HTH::deleteimage</a>}]
            append result [HTH::tableRow [list $i $Info $Play $Delete $flink $track] $bgc]
         }  else {
            append result [HTH::tableRow [list $i $Info $flink $track] $bgc]
         }
      }
      incr i
   }
   append result "</table>"
   #append result "<a href='javascript:toggleAll_jb_search_pl(true)'>Select All</a> | <a href='javascript:toggleAll_jb_search_pl(false)'>Unselect All"
   return $result
}

#------------------------------------------------------------------------------
# Function   : search::searchCatalog
# Description: Search all songs (up to N) for results
#              Returns HTML shown all the tracks that matched the filter
#              Info and checkboxes are included
# Author     : Tom Wilkason
# Date       : 12/22/2002
#------------------------------------------------------------------------------
proc search::searchCatalog {parNs filter max parms} {
   variable control
   variable myns
   append result "<FORM METHOD='POST' ACTION='/$myns' NAME='jb_add_form'>"
   append result "<input type='hidden' name='jb_add_start' value='$parms'>"
   # Initial param will be jb_add_start, followed by each checked item as jb_add, ending with jb_append_process=Submit
   append result {<table class='songPlaylist'>}
   if {$control==1} {
      append result [HTH::tableRow [list Chk ? P Folder "Track Name"] headerRow]
   } else {
      append result [HTH::tableRow [list Chk ? Folder "Track Name"] headerRow]
   }
   ##
   # Search all files for match
   #
   set i 0
   foreach {file} [db::filesMatchingPattern *$filter* 1] {
      set bgc [HTH::rowStyle i]
      set track [ROOTTAIL $file]
      set folder [HTH::displayFolder $file]
      if {$folder==""} {
         set flink [SP]
      } else {
         set flink [AREF "/$parNs?folder=[ENCODE [file dirname $file]]" $folder]
      }
      set cb [subst {<INPUT TYPE=checkbox NAME="jb_add" VALUE="$file">}]
      set infoBtn [HTH::popupInfo $file]
      # The second param just performs another search using the same filter
      if {$control} {
         set playBtn [subst {<a href="/$myns?play=[ENCODE $file]&jb_search=$filter&search_what=AllFiles"  title="Play Track Immediately">$HTH::playbutton</a>}]
         append result [HTH::tableRow [list $cb $infoBtn $playBtn $flink $track] $bgc]
      }  else {
         append result [HTH::tableRow [list $cb $infoBtn $flink $track] $bgc]
      }
      if {$i>$max} {
         break
      }
   }
   append result "</table>\n"
   if {$i>$max} {
      append result "<p>Results limited to $max matches</p>"
   }
   append result "<INPUT TYPE=submit VALUE='Queue Selected' NAME=jb_append_process>\n"
   append result "</FORM>\n"
   append result [subst {<a href="javascript:toggleAll_jb_add(true)">Select All</a> | <a href="javascript:toggleAll_jb_add(false)">Unselect All</a><br>}]
   return $result
}
##
# Misc routines not in a namespace
#
proc AREF {ref txt {title ""}} {
   if {[string length $title]} {
      return "<a href='$ref' title='$title'>$txt</a>"
   } else {
      return "<a href='$ref'>$txt</a>"
   }
}
proc H2 {ref} {
   return "<h2>$ref</h2>\n"
}
proc H1 {ref} {
   return "<h1>$ref</h1>\n"
}
proc SP {args} {
   return {&nbsp;}
}
proc NOAMP {data} {
   return [string map {& &amp\;} $data]
#   return [string map {& %26} $data]
}
proc ENCODE {data} {
   return [HttpD::Encode $data]
}
proc ESCAPE {data} {
   return [string map {' \\' & %26} $data]
}
proc ROOTTAIL {file} {
   return [file rootname [file tail $file]]
}
##
# Show the tail of a file without the folder
# or file extension.
#
proc DBINFO {file} {
   if {0} {
      return [LB::nameToKey $file]
   }  else {
      return [TAILIFCAN $file]
   }
}
proc TAILIFCAN {folder} {
   set tail [file tail $folder]
   if {$::snackAmpSettings(cleanListings)} {
      if {[string length $tail]} {
         return [fixplus::cleanName [file rootname $tail]]
      } else {
         return [fixplus::cleanName $folder]
      }
   }  else {
      if {[string length $tail]} {
         return [file rootname $tail]
      } else {
         return $folder
      }
   }
}
##
# Like TAILIFCAN but intended for folders
# where the file extension is not stripped
# off
#
proc FTAILIFCAN {folder} {
   set tail [file tail $folder]
   if {$::snackAmpSettings(cleanListings)} {
      if {[string length $tail]} {
         return [fixplus::cleanName $tail]
      } else {
         return [fixplus::cleanName $folder]
      }
   } else {
      if {[string length $tail]} {
         return $tail
      } else {
         return $folder
      }
   }
}
##
# Javascript routines, embed js::start, the variable then js::end
#
namespace eval js {
   proc start {args} {
     return {
     <script language="JavaScript">
     //<!--
     }
   }
   proc end {args} {
      return  {
     //-->
     </script>
     }
   }
   proc scripts {args} {
      set result [start]
      foreach {script} $args {
         variable $script
         append result [set [set script]]
      }
      append result [end]
      return $result
   }
   variable popupInfo {
        function popupInfo(url,height,width,tag) {
           var prefs = "scrollbars=yes,menubars=no,toolbars=no,"
           + "resizable=yes,left=0,top=0,height="+height+",width="+width;
           window.open(url, tag, prefs);
        }
   }
   variable textWindow {
        function textWindow(url) {
           var prefs = "scrollbars=no,menubars=no,toolbars=no,"
           + "resizable=yes,left=100,top=100,height=20,width=500;"
           tWin=window.open(url, "Text", prefs);
           setTimeout("tWin.close()",1000)
        }
   }
   variable toggleAll_jb_add {
        function toggleAll_jb_add(checked) {
           len = document.jb_add_form.elements.length;
           var i;
           for (i = 0; i < len; i++) {
              document.jb_add_form.elements[i].checked = checked;
           }
        }
   }
   variable toggleAll_jb_search_pl {
        function toggleAll_jb_search_pl(checked) {
           len = document.jb_search_pl_form.elements.length;
           var i;
           for (i = 0; i < len; i++) {
              document.jb_search_pl_form.elements[i].checked = checked;
           }
        }
   }
   variable countDown {
     <script language="JavaScript">
     //<!--
        //configure refresh interval (in seconds)
         var countDownInterval=$period;
         var countDownTime=countDownInterval+1;
         function countDown(){
            countDownTime--;
            if (countDownTime <=0) {
               countDownTime=countDownInterval;
               clearTimeout(counter);
               window.location.href = '$url'
            } else {
               // Check again in one second
               counter=setTimeout("countDown()", 1000);
            }
         }
         // Start if off
         if (document.all||document.getElementById)
            countDown();
         else
            window.onload=countDown;
     //-->
     </script>
   }
   variable foldersel {
      function foldersel(url,folder) {
        var useurl='none';
        var selsort='no sort';
        var selhow='no stream';
        var selsub='no';
        var i=0;
        //stream
        //selhow = getselected(document.option_form.cmb_How,2);
        for (i==0;i<2;i++) {
           if (document.option_form.cmb_How[i].selected) {
              selhow=document.option_form.cmb_How[i].text;
              break;
           }
        }

        //subfolders
        for (i==0;i<2;i++) {
           if (document.option_form.cmb_Sub[i].selected) {
              selsub=document.option_form.cmb_Sub[i].text;
              break;
           }
        }
        //sort
        for (i==0;i<4;i++) {
           if (document.option_form.cmb_Sort[i].selected) {
              selsort=document.option_form.cmb_Sort[i].text;
              break;
           }
        }
        useurl = url+
          '/?sort='  + selsort +
          '&tree='   + selsub  +
          '&how='    + selhow  +
          '&folder=' + folder;
        window.open(useurl,"tinyWindow", 'width=100,height=100');
      }
   }
   variable getselected {
      // getselected: Return the selected item from a combo box
      function getselected(element,len) {
         var i;
         for (i==0; i<len; i++) {
            alert(element[i].selected + ' ' + element[i].text);
            if (element[i].selected) {
               return element[i].text;
            }
         }
         return 'none';
      }
   }
}
#------------------------------------------------------------------------------
# Http helper routines
#------------------------------------------------------------------------------
namespace eval HTH {
# TFW: Can these references be computed from root as relative?
   variable id3Cache
   variable appendbutton   {<img border=0 src="/images/append.png">}
   variable folderclosed   {<img border=0 src="/images/folderclosed.gif">}
   variable folderopen     {<img border=0 src="/images/folderopen.gif">}
   variable infobutton     {<img border=0 src="/images/info.png">}
   variable insertbutton   {<img border=0 src="/images/insert.png">}
   variable playbutton     {<img border=0 src="/images/play.png">}
   variable trackimage     {<img border=0 src="/images/track.gif">}
   variable trackcropimage {<img border=0 src="/images/trackcrop.gif">}
   variable deleteimage    {<img border=0 src="/images/delete.png">}
   variable InfoCache

   variable TagIDs {
      Active        Active
      "File Name"   Name
      Folder        Folder
      Artist        Artist
      Title         Title
      Album         Album
      Track         Track
      Media         Media
      Year          Year
      Genre         Genre
      Note          Note
      Comp          Comp
      Encoding      Enc
      Desc          Desc
      Size          fSize
      Date          fDate
      Duration      duration
      Position      Position
      "Bit Rate"    Rate
      Gain          levelFactor
   }
   array set row_styles {
      0 {lightRows}
      1 {darkRows}
   }
   variable songTemplate {
      <table class="detailsTable">
        <tr>
          <td rowspan="7" valign="center">$Details(Cover)</td>
        </tr>
        <tr>
          <td colspan="6" class="songTemplateTitle"><b>[browser::allMusicRef $Details(Title) Title]</b></td>
        </tr>
        <tr>
          <td class="songTemplateLabel" >By</td>
          <td class="songTemplateInfo"   colspan="5">[browser::allMusicRef $Details(Artist) Artist]</td>
        </tr>
        <tr>
          <td class="songTemplateLabel" >From</td>
          <td class="songTemplateInfo"   colspan="5">[browser::allMusicRef $Details(Album) Album]</td>
        </tr>
        <tr>
          <td class="songTemplateLabel">File</td>
          <td class="songTemplateInfo"  colspan="5">$Details(Name)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel">Year</td>
          <td class="songTemplateInfo">$Details(Year)</td>
          <td class="songTemplateLabel">Genre</td>
          <td class="songTemplateInfo">$Details(Genre)</td>
          <td class="songTemplateLabel">Rate</td>
          <td class="songTemplateInfo">$Details(Rate)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel">Track</td>
          <td class="songTemplateInfo">$Details(Track)</td>
          <td class="songTemplateLabel">Duration</td>
          <td class="songTemplateInfo">$Details(duration)</td>
          <td class="songTemplateLabel">Position</td>
          <td class="songTemplateInfo">$Details(Position)</td>
        </tr>
      </table>
   }
   variable songInfoTemplate {
      <table class="infoTable">
        <tr>
          <td rowspan="4" valign="center">$Details(Cover)</td>
        </tr>
        <tr>
          <td class="songTitle" colspan="2" ><b>$Details(Title)</b></td>
        </tr>
        <tr>
          <td class="songLabel" >By</td>
          <td class="songInfo">$Details(Artist)</td>
        </tr>
        <tr>
          <td class="songLabel">From</td>
          <td class="songInfo">$Details(Album)</td>
        </tr>
      </table>
   }
   variable head {
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
   <head>
      <meta http-equiv="Content-Language" content="en-us">
      <meta name="GENERATOR"              content="SnackAmp Web Server">
      <meta name="ProgId"                 content="SnackAmp.Document">
      <LINK REL=StyleSheet                HREF="$css" TYPE="text/css" MEDIA=screen>
      <title>$title</title>
   }
   proc head {title css} {
      variable head
      set result [subst $head]
   }
   ##
   # Return the shorter of time to next track change or scheduled refresh
   #
   proc nextRefresh {args} {
      set rem [soundControl::remaining]
      if {[soundControl::isPlaying] && ($rem < $::snackAmpSettings(refreshRate))} {
         return [expr {$rem+1}]
      } else {
         return $::snackAmpSettings(refreshRate)
      }
   }
   proc getRefreshScript {period url} {
      Trace
      variable refreshScript
      return [subst $js::countDown]
   }
   proc coverArt {folder} {
      Trace
      if {[string length [coverart::covername $folder]]} {
         return [subst {<img border=0 width="$::snackAmpSettings(coversize)" height="$::snackAmpSettings(coversize)" src="/coverart?get=[ENCODE $folder]" alt="[file tail $folder]">}]
      } else {
         return ""
      }
   }
   proc rowStyle {_counter {mod 1}} {
      variable row_styles
      upvar $_counter counter
      if {[catch {incr counter} result] } then {
         set counter 0
      }
      return $row_styles([expr {($counter/$mod)&0x1}])
   }
   proc tableRow {cols {class ""}} {
      if {$class==""} {
         set result <tr>
      } else {
         set result [subst {<tr class="$class">}]
      }
      foreach {element} $cols {
         append result "<td>$element</td>"
      }
      append result "</tr>\n"
      return $result
   }
   proc inLinePlay {song} {
      return [subst {<a href="/remote/control?play=[ENCODE $song]" title="Play Track Immediately">$HTH::trackimage</a>}]
   }
   proc popUpProcess {what url song} {
      switch -- $what {
         play {set im $HTH::trackimage;set title "Play Track Immediately"}
         append {set im $HTH::appendbutton;set title "Append Track To Playlist"}
         insert {set im $HTH::insertbutton;set title "Queue Track to Play Next"}
         default {}
      }
      return [subst {<a href="javascript:textWindow('$url?$what=[ENCODE [ESCAPE $song]]')" title="$title">$im</A>}]
   }
   proc noamp {data} {
      return [string map {& %26} $data]
   }
   proc streamInfo {{class ""}} {
      set result ""
      foreach {file ip} [streamBuffer::allStreamInfo] {
         append result "[tableRow [list $file $ip] $class]\n"
      }
      return $result
   }
   proc displayFolder {File} {
      global snackAmpSettings
      if {$snackAmpSettings(ListDepth) > 0} {
         # File join doesnt work due to single list
         set File [string map {// /} $File]      ;# remove UNC stuff
         set name [file dirname [join [lrange [file split $File] end-$snackAmpSettings(ListDepth) end] /]]
      } else {
         set name [file tail [file dirname $File]]
      }
      return $name
   }
   proc popupInfo {ref} {
      # Escape single quotes for javascript popup window,
      # Encode & to prevent confusion with parameter delimiter
      set ref "/remote/search?question=[ENCODE [ESCAPE $ref]]"
      return [subst {<a href="javascript:popupInfo('$ref',200,600,'Info')" title="Show Track Details in Pop-up">$HTH::infobutton</a>}]
   }
}
##
# Schedule a variable to be unset some time in the future
# TODO: Could improve to not litter with a ton of afters,
# but have one that fires every 10 seconds to iterate
# over an internal list.
proc HTH::cleanLater {_var min} {
   after cancel [list HTH::clean $_var]
   after [expr {$min*60000}] [list HTH::clean $_var]
}
#----------------------------------------------------------------------------
# Return track data into an array, use a cache if it exists
# With metakit we prob. don't need a cache
#----------------------------------------------------------------------------
proc HTH::getTrackData {file _Info} {
   upvar $_Info Info
   variable id3Cache
   if {[info exists id3Cache($file)]} {
      array set Info $id3Cache($file)
   } else {
      array unset info
      ;# Get from database if possible
      set hasTag [db::lookup $file Info 1]
      # Insert the tag data if available
      foreach item {Title Artist Album} {
         if {$Info($item)==""} {
            set Info($item) [id3Tag::guess${item} $file]
         }
      }
      # Save for next time
      set HTH::id3Cache($file) [array get Info]
   }
   HTH::cleanLater HTH::id3Cache($file) 3
}
proc HTH::leader {title css {url "xx"}} {
   Trace
   # title is used in subst
   append result [head $title $css]
   if {![string equal $url "xx"]} {
     append result \n
     append result [subst {    <META HTTP-EQUIV="Refresh" CONTENT="$::snackAmpSettings(refreshRate);URL=$url">}]
     append result \n
   }
   append result "  </head>\n"
}
#------------------------------------------------------------------------------
# Function   : ::HTH::songTable
# Description: Return HTML for the song details
# Author     : Tom Wilkason
# Date       : 4/29/2002
#------------------------------------------------------------------------------
proc HTH::songTable {_Details} {
   upvar $_Details Details
   variable songTemplate
   set result [subst $songTemplate]
   return $result
}
proc HTH::songInfo {_Details} {
   upvar $_Details Details
   variable songInfoTemplate
   set result [subst $songInfoTemplate]
   return $result
}

#------------------------------------------------------------------------------
# Function   : trackInfo
# Description: Return, in a table, the details on some file
#
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc HTH::trackInfo {fileName {short 0}} {
   variable TagIDs
   variable InfoCache
   if {[info exists InfoCache($fileName)]} {
      array set Details $InfoCache($fileName)
   } else {
      foreach {name tag} $TagIDs {
         set Details($tag) ""
      }
      if {[catch {catalog::trackFullData $fileName Details} error] } then {
         HttpD::Log "trackInfo:$error"
      }
      set Details(Folder) [file dirname $fileName]
      set Details(Cover) [HTH::coverArt $Details(Folder)]
      set Details(Name) [file tail $fileName]
      set Details(Position) "N/A"
      # Fill in non tagged data based on file name
      foreach item {Title Artist Album} {
         if {$Details($item)==""} {
            set Details($item) [id3Tag::guess${item} $fileName]
         }
      }
      ##
      # File info
      #
      if {![catch {file stat $fileName Stat}]} {
         set Details(fSize) "[commify $Stat(size)] Bytes"
         set Details(fDate) [clock format $Stat(mtime) -format $::snackAmpSettings(timeString)]
      }
      # TFW:This forces table borders on blank cells, there may be a way
      # to do this in the table definition.
      foreach field [array names Details] {
         if {[string length $Details($field)]==0} {
            set Details($field) [SP]
         }
      }
      set InfoCache($fileName) [array get Details]
   }
   HTH::cleanLater HTH::InfoCache($fileName) 5
   if {$short} {
      set result [HTH::songInfo Details]
   } else {
      set Details(Position) [clock format [expr {round([soundControl::position])}]  -format %M:%S]
      set result [HTH::songTable Details]
   }
   return $result
}
##
# Show play status
#
proc HTH::playStatus {{showWhere 1}} {
   set Active [string tolower [soundControl::playState]]
   if {$showWhere} {
      append result [H2 "SnackAmp currently $Active on [info hostname]:$::HttpD::Httpd(port)"]
   } else {
      append result [H2 "SnackAmp currently $Active"]
   }

}
#------------------------------------------------------------------------------
# Function   : Return the current playlist, with appropriate controls
# Description:
# Author     : Tom Wilkason
# Date       : 7/2/2003
#------------------------------------------------------------------------------
proc HTH::trackList {control top newEnd {parms ""}} {
   set result ""
   if {$::snackAmpSettings(listbytag)} {
      set fHeader "Artist"
   } else {
      set fHeader "Folder"
   }
   switch -- $control {
      0 {append result [HTH::tableRow [list $fHeader "Track Name"] headerRow]}
      1 {append result [HTH::tableRow [list ? X $fHeader "Track Name"] headerRow]}
      2 {
         if {$::snackAmpSettings(jukebox,popup)} {
            append result [HTH::tableRow [list ? $fHeader "Track Name"] headerRow]
         } else {
            append result [HTH::tableRow [list $fHeader "Track Name"] headerRow]
         }
      }
      default {}
   }
   set ctr 0
   for {set i $top} {$i < $newEnd} {incr i} {
      set track ""
      if {$i == [LB::idxPlaying]} {
         set bgc activeRow
      } else {
         set bgc [HTH::rowStyle ctr 1]
      }
      set song [LB::keyFromIndex $i]
      set file [LB::nameFromIdx $i 0]
      # Show either tags or folder/file names
      if {$::snackAmpSettings(listbytag)} {
         HTH::getTrackData $file Info
         set track $Info(Title)
         set folder $Info(Artist)
      } else {
         set folder [HTH::displayFolder $file]
         set track [ROOTTAIL $file]
      }
      ##
      # Show the appropriate row with appropriate controls
      #
      switch -- $control {
         0 {
            # No Control
            if {$folder==""} {
               set flink [SP]
            } else {
# Note: hyperlink for /display only
               set flink $folder
            }
            append result [HTH::tableRow [list $flink $track] $bgc]
         }
         1 {
            # Full Control
            if {$folder==""} {
               set flink [SP]
            } else {
               # Hyperlink only if remote controling (i.e. control= 1)
               # /remote is current (need to fix to use remote/browser as current)
               set flink [AREF "browser?folder=[ENCODE [file dirname $file]]" $folder "Browse This Folder"]
            }
            ##
            # Replace this so ?play= will work from any calling routine, not just /remote/control
            #
            set track "[HTH::inLinePlay $song][NOAMP $track]"
            set imInfo [HTH::popupInfo $song]
            set imDel [AREF "?delete=$i$parms" $HTH::deleteimage "Remove Track From Playlist"]
            append result [HTH::tableRow [list $imInfo $imDel $flink $track] $bgc]
         }
         2 {
            # Just hyperlink (jukebox)
            if {$folder==""} {
               set flink [SP]
            } else {
               if {$::snackAmpSettings(jukebox,linkartist)} {
                  set flink [AREF "?folder=[ENCODE [file dirname $file]]" $folder "Browse This folder"]
               } else {
                  set flink $folder
               }
               # /jukebox/browser is current
            }
            if {$::snackAmpSettings(jukebox,popup)} {
               set imInfo [HTH::popupInfo $song]
               append result [HTH::tableRow [list $imInfo $flink $track] $bgc]
            } else {
               set imInfo [HTH::popupInfo $song]
               append result [HTH::tableRow [list $flink $track] $bgc]
            }
         }
         default {}
      }
   }
   return $result
}
#------------------------------------------------------------------------------
# Function   : HttpHelpers
# Description: Create a temp file, remove it 30 seconds later
# Author     : Tom Wilkason
# Date       : 7/8/2003
#------------------------------------------------------------------------------
proc HTH::WriteTempFile {ext content} {
   set mypath [file join $::TA(TEMP) [clock clicks].$ext]
   # temp file
   set fid [open $mypath w]
   fconfigure $fid -translation binary
   puts -nonewline $fid $content
   close $fid
   after 30000 [list file delete -force $mypath]
   return $mypath
}

#------------------------------------------------------------------------------
# Return a option list of sorting modes
#------------------------------------------------------------------------------
proc HTH::sortModes {args} {
   foreach {mode} [subSort::allSortModes] {
      append result "<option>$mode</option>\n"
   }
   return $result
}
#------------------------------------------------------------------------------
# Function   : nextToPlay
# Description: Print out the next N tracks in the playlist
#
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc HTH::nextToPlay {playButton {N 20} {parms ""}} {
   Trace
   global LB
   set offset 2
   set result {}
   if {[catch {
      set top [expr {[LB::idxPlaying]-$offset}]
      set end [LB::index end]
      set newEnd [expr {$top+$N+$offset}]
      if {$newEnd > $end} {set newEnd $end}
      if {$top < 0} {set top 0}
      set current [file tail $::TA(currentMF)]
      append result [H2 "Coming up next (of [commify $end] tracks)"]
      append result "<table class='songPlaylist'>"
      append result [HTH::trackList $playButton $top $newEnd $parms]
      append result </table>
#      append result </div>
   } error]} {
      return $error
   } else {
      return $result
   }
}
##
# Return list of tracks that are queued by the user to play
#
proc HTH::queuedToPlay {playButton {N 20} {parms ""}} {
   Trace
   global LB
   set offset $N
   set result {}
   if {[catch {
      set top [expr {[LB::idxPlaying]+1}]
      set end [LB::userInsert]
      if {$top < 0} {set top 0}
      if {$end-$top <= 0} {
         append result [H2 "No tracks queued"]
      } else {
         # Limit the number of queued tracks
         if {$end-$top > $::snackAmpSettings(httpNextLength)} {
            set end [expr {$top+$::snackAmpSettings(httpNextLength)}]
         }
         #set current [file tail $::TA(currentMF)]
         append result [H2 "User queued tracks"]
         append result "<table class='songPlaylist'>"
         append result [HTH::trackList $playButton $top $end $parms]
         append result </table>
      }
#      append result </div>
   } error]} {
      return $error
   } else {
      return $result
   }
}
##
# Unset a variable
#
proc HTH::clean {_var} {
   upvar $_var var
   if {[info exists var]} {
      unset var
   }
}
##
# Verify a file is within one of the media folders
#
proc HTH::isMedia {track} {
   foreach media $::TA(mediaDirs) {
      if {[string match $media* $track]} {
         return 1
      }
   }
   return 0
}
package provide urlRedirect 1.0
##
# Utility routines
#
proc dumpSocket {} {
   foreach glb [info globals Httpdsock*] {
      global $glb
      puts "Array $glb--------------"
      catch {parray $glb}
   }
}
proc afters {args} {
   foreach after [after info] {puts [after info $after]}
}

