#!/bin/sh
# the next line restarts using wish \
exec wish "$0" ${1+"$@"}
set SAVersion "3.0a1"
#------------------------------------------------------------------------------
# snackAmp 2.2.1
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
# Originally derived from Snack snamp.tcl demo player by Tom Wilkason
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/snackAmp.tcl,v 1.107 2004/02/11 00:55:21 wilkason Exp $
}
package require Tk
wm withdraw .
#----------------------------------------------------------------------------
# Check if this is the only invocation; If on windows
# have the other instances raise itself (needed if databases being used)
#----------------------------------------------------------------------------
if {$::tcl_platform(platform) == "windows"} {
   package require dde
   ;# DDE will only launch a new instance if needed, so no need to check argv
   if {[llength [dde services TclEval snackAmp]]>0} {
      dde eval snackAmp {raise .saTop}
      exit
   }
} else {
   #----------------------------------------------------------------------------
   # Check if this is the only invocation; According to the docs, a second
   # invocation will contain #2 after SnackAmp (on unix)
   #----------------------------------------------------------------------------
   if {![string equal SnackAmp [tk appname SnackAmp]]} {
      if {[llength $argv]==0} {
         tk_messageBox -message "Sorry! You can't run two versions of SnackAmp simultaneously!"
         send SnackAmp {raise .saTop}
      } else {
         send SnackAmp PlayFile $argv
      }
      exit
   }
}
#------------------------------------------------------------------------------
# Function: scriptDir
# Return the name of the folder from which this script is playing
#------------------------------------------------------------------------------
proc scriptDir {args} {
   global src_dir auto_path argv0
   set script $argv0
   while {[file type $script] == "link"} {
      set link [file readlink $script]
      if {[file pathtype $link] == "relative"} {
         set link [file join [file dirname $script] $link]
      }
      set script $link
   }
   set src_dir [file dirname $script]
   set here [pwd]
   cd $src_dir
   set src_dir [pwd]
   cd $here
   return $src_dir
}
#------------------------------------------------------------------------------
# Function   : setTKOptions
# Description: Define default settings then read in user defined settings
# Author     : Tom Wilkason
# Date       : 8/8/2003
#------------------------------------------------------------------------------
proc setTKOptions {} {
   ##
   # Set some widget defaults before we build any guis
   #
   #set tk_strictMotif 1
   # Buttons with icons in them
#    foreach wid {mb saTop.hull saTop.panel} {
#       option add *$wid*Button.height     18 startupFile
#       option add *$wid*Button.width      18 startupFile
#       option add *$wid*Menubutton.height 18 startupFile
#       option add *$wid*Menubutton.width  18 startupFile
#    }
   option add *highlightThickness              0         startupFile
   option add *SnackAmp*Button.relief          flat      startupFile
   option add *SnackAmp*Scale.activeBackground lightblue      startupFile
   option add *SnackAmp*Scale.font             timeScale startupFile
   option add *SnackAmp*Scale.borderWidth      1 startupFile
   ##option add *Button.borderWidth              1         startupFile
   option add *Button.font                     menu      startupFile
   option add *Button.takeFocus                0         startupFile
   option add *Button.highlightThickness       1         startupFile
   #option add *Button.activeBackground         lightblue         startupFile ;# works
   #option add *Button.highlightColor           lightblue         startupFile ;# doesn't work
   option add *Button.borderWidth              1         startupFile ;# doesn't work
   option add *Checkbutton.borderWidth         1         startupFile
   option add *Checkbutton.font                menu      startupFile
   option add *Dialog.msg.font                 menu      startupFile
   option add *Dialog.msg.wrapLength           9i        startupFile
   option add *Entry.borderWidth               1     startupFile
   option add *Entry.font                      entry     startupFile
   option add *Entry.relief                    sunken    startupFile
   option add *FolderTree*Button.relief        flat      startupFile
   option add *FolderTree*Menubutton.relief    flat      startupFile
   option add *Label.takeFocus                 0         startupFile
   option add *Label.borderWidth               1         startupFile
   option add *Label.font                      entry     startupFile
   option add *Label.relief                    flat      startupFile
   option add *Listbox.borderWidth             1         startupFile
   option add *Listbox.font                    listbox   startupFile
   option add *Menu.font                       menu      startupFile
   option add *Menu.padY                       0         startupFile
   option add *Menu.tearoff                    0         startupFile
   option add *Menubutton.borderWidth          1         startupFile
   option add *Menubutton.font                 menu      startupFile
   option add *Menubutton.highlightThickness   1         startupFile
   option add *Menubutton.padY                 0         startupFile
   option add *Menubutton.relief               raised    startupFile
   option add *Radiobutton.font                menu      startupFile
   ##option add *Scale.borderWidth               1         startupFile
   if {$::tcl_platform(platform)=="unix"} {
      option add *Entry.selectBorderWidth          0       startupFile
      option add *Listbox.selectBorderWidth        0       startupFile
      option add *Text.selectBorderWidth           0       startupFile
      # Used for modern L&F of menus
      option add *Menu.activeBackground            darkblue  startupFile
      option add *Menu.activeForeground            white   startupFile
      option add *Menu.activeBorderWidth           0       startupFile
      ##option add *Menu.borderWidth                 1       startupFile
      option add *Menu.highlightThickness          0       startupFile
   }

}
#------------------------------------------------------------------------------
# Function   : postGUIsettings
# Description: Fire up the remaining stuff
# Author     : Tom Wilkason
# Date       : 8/8/2003
#------------------------------------------------------------------------------
proc postGUIsettings {} {
   global snackAmpSettings TA
   fixScaleBinding
   ##
   # Bring up the GUI now and set the width to the prior saved width
   # Also restore the prior position if the user wants.
   if {[info exists snackAmpSettings(playerPos)] && $snackAmpSettings(restorePlayerPos)} {
      set pp "+[join $snackAmpSettings(playerPos) +]"
   } else {
      set pp ""
   }
   update idletasks
   wm geometry $::SA "$::snackAmpSettings(minWidth)x[winfo reqheight $::SA]$pp"
   wm deiconify $::SA

   # Set debugging mode
   snack::debug $::snackAmpSettings(Trace)
   [soundControl::sound] config -debug $::snackAmpSettings(Trace)
   soundControl::latency normal

   # Read the various cache files and restore playlist if needed.
   # Monitor stuff if required
   after 5000 catalog::MonitorScanFolders
   # quietly mount any removable media
#   if {$snackAmpSettings(UseRemovableFolders)} {
#      after 1000 catalog::MountMedia "xx" 1
#   }
   ##
   # Restore mini-player if required
   #
   if {$::snackAmpSettings(restoreMiniPlayer)} {
      if {$::snackAmpSettings(activeMiniPlayer)} {
         foreach {x y} $::snackAmpSettings(positionMiniPlayer) {break}
         package require miniPlayer
         miniPlayer::toggle $x $y
      }
   } else {
      set ::snackAmpSettings(activeMiniPlayer) 0
   }
}
##
# If this is not wrapped, add libs off snackAmp to auto_path
# The wrapped version uses virtual folders which are inserted
# by the wrap program. This method works with all tclkit
# versions.
#
proc checkType {args} {
   global TA
   if {[hasNamespace ::starkit]} {
      # Note: Dual dirname since SnackAmp/wrap is location
      # of info script
      set TA(appname) "\"[file nativename [info nameofexecutable]]\""
      set TA(isWrapped) 1
      set TA(snackRoot) $::starkit::topdir
      set TA(realRoot) [file dirname $::starkit::topdir]
      #------------------------------------------------------------------------------
      # Function   : load
      # Description: If files can be loaded locally, load them if possible
      # Author     : Tom Wilkason
      # Date       : 1/15/2003
      #------------------------------------------------------------------------------
      rename load _load
      proc load {args} {
         set packageName {}
         foreach {filename packageName} $args {break}
         set newfilename [file join $::TA(realRoot) [file tail $filename]]
         if {[file exists $newfilename] && [string length $filename]} {
            eval [list _load $newfilename $packageName]
         } else {
            eval [list _load $filename $packageName]
         }
      }
   } else {
      set TA(appname) "\"[file nativename [info nameofexecutable]]\" \"[info script]\""
      # Not wrapped in any way
      set TA(realRoot) [scriptDir]
   #   puts "Running as scripts"
      set TA(isWrapped) 0
      set TA(snackRoot) [scriptDir]
      # Need to build our own auto_path
      lappend ::auto_path $TA(snackRoot) [file join $TA(snackRoot) lib]
   }
}

# compatability: namespace exists is part of 8.4
proc hasNamespace {namespace} {
   if {[lsearch [namespace children ::] $namespace] >= 0} {
      return 1
   } else {
      return 0
   }
}

#------------------------------------------------------------------------------
# Return the appropirate packge to use for the database based on arguments
#------------------------------------------------------------------------------
proc processCommandOptions {args} {
   ##
   # Setup other home based info
   #
   global argv snackAmpSettings TA HOME
   set HOME [findMyHome]
   set ::HASDB 0
   set newArgs [list]
   set TA(database) default
   foreach {opt param} $argv {
      switch -- $opt {
         "-database" {
            switch -- $param {
               "metakit" {
                  set TA(database) snackMetakit
                  set ::HASDB 1
               }
               "sqllite" {
                  set ::HASDB 1
                  set TA(database) snackSqllite
               }
               "filedb" {
                  set TA(database) fileDB
               }
               default {
                  set TA(database) fileDB
               }
            }
         }
         "-home" {
            set TA(home) $param
            set HOME $param
            puts "Using alternate home folder of '$param'"
         }
         default {}
         default {
            lappend newArgs $opt $param
         }
      }
   }
   # Stripped out arguments used here
   set argv $newArgs
#tk_messageBox -message $HOME
   # TBD: for default catalog location use $HOME/My Documents/My Music
   if {![file exists $HOME]} {
      if {[catch {
         file mkdir $HOME
      } result] \
      } then {
         tk_messageBox -message "Sorry! I could not create the folder '$HOME'
   to save your settings file. You can do one of three things:
   1. Set the environment variable SNACKAMPHOME to some folder you can write to
   2. Run SnackAmp with a '-home location' option where you can write to location
   3. Gain write access to $::HOME

   Restart SnackAmp when you are ready" -type ok -title "Press OK when done" -icon warning
      exit
         #error
      }
   }
   set TA(settingsFile)        [file join $HOME .snackAmpDefaults]
   set TA(playCacheFile)       [file join $HOME .snackAmpPlayCache]
   set TA(defaultFolderFile)   [file join $HOME default.tafs]
   set TA(logFile)             [file join $HOME snackAmp.log]
   set TA(scriptHome)          [scriptDir]
   set TA(defaultSettingsFile) [file join $TA(scriptHome) ".snackAmpDefaults"]
   set TA(cacheFile)           ".snackAmpCache2"
   set TA(pushFile)            [file join $HOME ".snackAmpPush"]
   set TA(STOP) 0
   if {[file exists $TA(logFile)]} {
      catch {file delete $TA(logFile)}
   }
   set TA(webRoot) [file join $TA(snackRoot) docs]
   switch -- $::tcl_platform(platform) {
      "unix" {
         set TA(arrow) "arrow"
         set TA(TEMP) /tmp
      }
      default {
         set TA(arrow) "arrow"
         set TA(TEMP) $::env(TEMP)
      }
   }

   ##
   # Some initial values
   #
   set TA(PlayListSelection) {}
   set TA(PlayListFile) {}
   set TA(Console) 0
   set TA(matchAgainstAll) 0     ;# don't match against all
   set TA(currentMF) {}
   set TA(useSAFilter) 0
   ##
   # Removable media tag
   #
   set ::RMT "/RMC"
   return 1
}
#------------------------------------------------------------------------------
# Function   : findMyHome
# Description: Location the appropirate location of the HOME folder
# Author     : Tom Wilkason
# Date       : 1/23/2002
#------------------------------------------------------------------------------
proc findMyHome {args} {
   global env HOME TA
   ##
   # Default home folder names
   #
   # First check the command line, use it if specified
   if {[info exists TA(home)] && [file exists $TA(home)]} {
      return $TA(home)
   }
   # No argument, check other stuff
   set sHome ".snackAmp"
   if {[info exists env(SNACKAMPHOME)]} {
      # Then check the SNACKAMPHOME env var for home
      # and use it if it exists
      return [string trim $env(SNACKAMPHOME) \"]  ;#"
   } else {
      # Determine appropriate home location
      switch -- $::tcl_platform(platform) {
         "unix" {
            return [file join ~ $sHome]
         }
         "windows" {
            # Only use HOME if setup by a prior SnackAmp
            if {[info exists env(HOME)] && [file exists [file join $env(HOME) $sHome]]} {
               return [file join [string trim $env(HOME) \"] $sHome]  ;#"
            }
            # Otherwise try to use USERPROFILE
            if {[info exists env(USERPROFILE)] && [file exists [file join $env(USERPROFILE) $sHome]]} {
               return [file join $env(USERPROFILE) $sHome]
            }
            ;# Revert to more windows friendly name
            set sHome "snackamp"
            if {[info exists env(USERPROFILE)]} {
               return [file join $env(USERPROFILE) $sHome]
            }
            # Some older windows with profiles, just use folder
            # where snackamp is stored.
            if {[file exists [file join [scriptDir] $sHome]]} {
               return [file join [scriptDir] $sHome]
            }
            # By default just use a /settings folder
            # .snackAmp causes problems on older win9x machines
            return [file join [scriptDir] "settings"]
         }
         default {
            return [file join ~ $sHome]
         }
      }
   }
}
##
# Trick to create a menu font that matches
# the native windows
#
proc createMenuFont {args} {
   menu .temp
   if {[string match "*-*-*-*" [.temp cget -font]] || [.temp cget -font]=="system"} {
      # Handle X style fonts as defaults
      .temp configure -font {Helvetica -16}
   }
   foreach {font size} [.temp cget -font] {break}
   font create menu -size $size -family $font
   destroy .temp
}

# Routine to handle multiple arguments on linux (windows uses DDE)
proc handleArgs {argv} {
   set last [LB::idxEnd]
   foreach {file} $argv {
      if {[file extension $file]=="" || [regexp -nocase -- {\.m3u|\.pls} $file]} {
         PlayFile $file
         return
      }
      if {[file exists $file]} {
         # puts "Appending $file"
         LB::userInsert $file
      }
   }
   # If some valid files were added, then play the first one
   set here [LB::idxEnd]
   if {$here > $last} {
      incr last
      macros::Highlight $last
      after 200 Play
   }
}

#  ----------------------------------------------------------------------------
#                T H I S  I S   W H E R E   W E   S T A R T
#  Windows platforms get taskbar icons
#  and dde support
#  ----------------------------------------------------------------------------
if {$::tcl_platform(platform) == "windows"} {
   package require dde
   dde servername snackAmp
   ##
   # Set up some linked vars for direct sound focus control
   #
   set DSSCL_NORMAL                0x00000001
   set DSSCL_PRIORITY              0x00000002 ;#default
   set DSSCL_EXCLUSIVE             0x00000003
   set DSSCL_WRITEPRIMARY          0x00000004

   set DSBCAPS_STICKYFOCUS         0x00004000 ;#default
   set DSBCAPS_GLOBALFOCUS         0x00008000

   set DSSCL_FLAG                  $DSSCL_NORMAL
   set DSBCAPS_FLAG                $DSBCAPS_GLOBALFOCUS
}
#  ----------------------------------------------------------------------------
#  Do some prelimiary work checking where we are
#  and how we were launched.
#  ----------------------------------------------------------------------------
checkType
###
createMenuFont
package require saImages
createSAImages

package require cframe
package require gpWidgets
waitWindow::init
waitWindow::show "SnackAmp $::SAVersion Initializing...Please Wait"
update
package require saGUI
package require soundControl
package require macros
package require gpUtils
package require saFiltering
package require catalog
package require saSettings
package require playurl
;# updateEvents allows timer callbacks to happen
if {[info command updateEvents]=="updateEvents"} {
   proc Update {args} {updateEvents timers}
} else {
   proc Update {args} {}
}
if {[info command console] == {}} {
   catch {package require saConsole}
}
if {[catch {package require snack} result]} {
   tk_messageBox -icon warning -type ok -title "Snack Load Error\$result" \
      -message "$result"
   exit
}
#  ----------------------------------------------------------------------------
#  Now, start everything up here
#  ----------------------------------------------------------------------------
# process command line options
# Set up the gui options and default folders
setTKOptions
# Process the command line options and setup the home folder
processCommandOptions
namespace import macros::*
# Read the user settings
settings::readSettings

package require playList
# Some of these may have changed based on user settings
option add *Entry.background         $::snackAmpSettings(ebg) startupFile
option add *Entry.foreground         $::snackAmpSettings(efg) startupFile
option add *Scale.troughColor        $::snackAmpSettings(bg)  startupFile
option add *Listbox.background       $::snackAmpSettings(bg)  startupFile
option add *Listbox.foreground       $::snackAmpSettings(fg)  startupFile
option add *Listbox.selectForeground $::snackAmpSettings(fg)  startupFile

# Setup the required Database Package
if {[catch {package require $TA(database)} result] } then {
   tk_messageBox -type ok -icon warning \
      -message "Metakit is required for the SnackAmp database\nhttp://www.equi4.com/metakit/tcl.html"
   exit
}
;# Flag it the database is new or already exists
set exists [db::exists]
#------------------------------------------------------------------------------
# Open the database
#------------------------------------------------------------------------------
db::openDB
#after 5000 db::bgCacheFiles
db::looper 10000
if {$exists==0} {
   catalog::convertCacheFiles
}
unset exists


# Read the previously saved gain control settings if possible
#gainControl::restoreSettings
# Build the GUI
# Fire up the web server
package require HttpD
package require urlRedirect
# Build the user interface
waitWindow::grab
#safety measure
after 5000 waitWindow::release
gui::buildGui
# Fire up the after loops and place the gui's
postGUIsettings
# Restore cache files
catalog::playlistStartup
# Install an IRC remote control interface server
if {$snackAmpSettings(useIRC)} {
   package require ircserver
}
##
# otherwise, handle command line arguments
if {$::tcl_platform(platform) != "windows"} {
   ##
   # Play any files that are passed in (on Windows
   # these are handled by the dde server)
   #
   myIcon $::SA
   if {[info exists argv] && [llength $argv] > 0} {
      handleArgs $argv
   }
}
# Ah, finally done
waitWindow::release
waitWindow::close
